/*****************************************************************************/
/****************** (C) 1991, 1992 by Miroslav Kolar *************************/
/*									     */
/*      Classical & quantum Monte-Carlo simulation for a single		     */
/*       particle in a one-dimensional external potential V(x)               */
/*									     */
/*  Internally, potential is assumed to be scaled to certain atomic-like     */
/*  units, so that the hamiltonian is  -(1/2)(d^2/dx^2) + V(x)		     */
/*									     */
/*  To obtain the quantum version of the program, the name QUANTUM must	     */
/*  be defined, e.g., in the compiler flag -DQUANTUM. Otherwise, the	     */
/*  classical version is obtained.					     */
/*									     */
/*****************************************************************************/

/*
External files required: 1D.h
			 1Dpotential.c
			 RanGen.c
			 LinePlots.c
			 DateAndTime.c

 List of functions in this file:
	void printChoices(N, descr) int N; char *descr;
	void usage(mode) int mode;
	void defaults(mode) int mode;
	void help1()
	void help()
	void helpR()
	main(argc, argv)  int argc; char *argv[];
	void checkCL()
	void printAll(mode, msg)  int mode; char *msg;
	void err_fseek(proc)  char *proc;
	void err_fread(proc)  char *proc;
	void write_offsets()
	void read_offsets(p) char *p;
	void read_head(p) char *p;
	void write_N();
	void read_N(p) char *p;
	void write_P1()
	void read_P1(p) char *p;
	void write_P2()
	void read_P2(p) char *p;
	void write_C()
	void read_C(p) char *p;
	void write_H_short()
	void write_H()
	void read_H(p) char *p;
	void write_A()
	void read_A(p) char *p;
	void write_Lm(p) char *p;
	void OutData()
	void dynDefs()
	void dynDefs2()
	void betaTout(bp,Nline) double *bp; int Nline;
	void InData()
	void ConfigHead()
	void printParams()
	void Initialize()
	void updateDens()
	void MCcycleP1s()
	void MCcycleP1()
	void MCcycleP2()
	void MCcyclePC()
	void MCcycleR1s()
	void MCcycleR1()
	void MCcycleR2()
	void MCcycleR3()
	void MCcycleR4()
	double Pot(shift) double shift;
	double MovePolymer()
	void Mix()
	void Sample()
	void OutDensity()
        void Final()
	void mult_Final()
	void displayFile(Listing)  FILE *Listing;
	void repairFile()

External functions:
	char *DateAndTime();			(in  ~DateAndTime.c)
	void initPlot(), plotFun1(), xRange(), plotPoints(), Histogram();
						(in  ~LinePlots.c)
        Currently available potentials are defined in 1Dpotential.c:
		double (*V)();
		double (*epsVir)();  = V(x) + 0.5*(x-Shift)* dV(x)/dx
		void   (*inV)();     input of the parameters of the potential
		long   (*outV)();    output of the parameters of the potential
		void choosePotential();

        Currently available random number generators are defined in RanGen.c.
*/

#include <errno.h>
#include <sys/times.h>
#include "1D.h"

/* #define DEBUG */
/* #define DEBG1 */

#define VERSION    4    /* Change VERSION whenever any changes are made to
			   the organization of the Master File, etc.	   */

#define SC         (void)scanf(
#define SI         (void)fscanf(Input,
#define FC         (void)fclose(
#define SQR(x)     ((x)*(x))
#define ILLEGAL    { PE" Illegal option '%s' ignored!\n",argv[i]); illeg++; }
#define FIERROR(x) { PR" File '%s': %s\n", x, sys_errlist[errno]); exit(-1); }
#define max(x, y)  ((x) < (y) ? (y) : (x))

#define MAXLETT    101   /* max length of file names, text lines, etc.
			    Must be at least 10  */
/* Some additional #defines can be found in 1D.h and before the OutData()
function. */

#define INIT 	   0	/* INIT must remain 0 as the proper function of several
			   segments of the code depends on the fact that
			   actionFlag = INIT = 0 for the Initialization process.
			*/
#define MIX 	   1
#define SAMPLE	   2
#define FINAL	   3
#define EPS	   4
#define DISPLAY	   5
#define REPAIR	   6
#define MAXaction  7	/* Current maximum number of the actions
			   performed by the program. */

#ifdef QUANTUM
#define FOLDED		0
#define LINEAR		1
#define MAXinitPoints	2	/* Current maximum number of different
				   starting configurations.   */
#define S1		   0
#define S2		   1
#define S3		   2
#define S4		   3
#define S5		   4
#define S6		   5
#define S7		   6
#define MAXselectionMode   7	/* Current maximum number of different ways of
				   selecting the particle to be moved next  */
#endif
		/* If adding actions, selection modes or starting
		   configurations, don't forget to add corresponding
		   command line syntax at the beginning of main().	*/

/* Defaults:  */
#define DEFactionFlag		INIT
#define DEFpotFlag		0      /* see 1Dpotential.c   */
#define DEFPlotInt		0
#define DEFPlotIntMix		100    /* = DEFPlotInt for mixing */
/* #define DEFPlotWidth		115 */
#define DEFPlotWidth		65
#define DEFMaxCycles		1000000L
		/* Default maximum number of cycles during sampling.  */
#define DEFMaxMix		5000L
		/* Default maximum number of cycles during mixing.  */
#define DEFSamplingInt		1L
#define DEFSamplingBlk		20000L
#define DEFN_x			10
#ifdef QUANTUM
#define DEFinitPointFlag	FOLDED
#define DEFN_fict		6
#define DEFselectionMode	S1
#define DEFintT			10
#define DEFMaxShftT		0
#endif

/* ------ Global declarations: ------- */
#if defined(ultrix) || defined(__osf__)
extern int  errno;
extern void *malloc();
extern void setbuf();
#else
int  errno;
extern char *malloc();
extern int 
setbuf();
#endif
extern char *sys_errlist[];
extern double exp();
extern clock_t times();
extern void exit();

extern char *DateAndTime();			/* in DateAndTime.c  */
extern void initPlot(), plotFun1(), xRange(), plotPoints(),
	    Histogram();			/* in LinePlots.c    */

static clock_t usLast, sysLast;	/* used for timing */
static struct tms cpuT;		/*    - " -        */

static char *author = "                  Written by M. Kolar (C) 1991, 1992\n",
	    *space = "\n                                    ",
	    *space43 = "                                           ",
	    *progName;	/*  to be set equal to  argv[0]  */

 /* All important data enabling to continue/restart a MC simulation are stored
    in what is referred below as a Master File. Depending on the action
    performed in a given run of the program (determined by  actionFlag), only
    an Input Master File, or only an Output Master File, or both are required.
    At the same time, Input and Output Master Files can be identical (only one
    of them is opened at any given time).
 */
static int
   actionFlag = DEFactionFlag,
	/* = INIT - create new configuration, do some initial mixing, and write
		    the final configuration into the Output Master File.
		    [Output only]
	   = MIX - read a configuration from the Input Master File, do some
		   more mixing, and write the final configuration into the
		   Output Master File. [no samplings done; Input/Output]
		(During any mixing, every PlotInt-th cycle, the configuration
		is rendered on the Listing file.)
	   = SAMPLE - the proper Monte-Carlo (MC) calculation with
		      subsequent samplings being appended to the Output Master
		      File. After MaxCycles cycles, calculate the errors
		      (on a second pass through Output MF).  [Input/Output]
	   = FINAL - no more samplings, only calculate the averages and errors
		     using a previously generated Master File.
	   = EPS - similar to FINAL. It is possible to calculate the contri-
		   bution of varying portions of the input file. [Input only]
	   = DISPLAY - display the contents of a Master File.  [Input only]
	   = REPAIR - when a system failure occurs during the output,
		      number of blocks and cycles in the C-block of the
		      Master File may not correspond to the actual number
		      of blocks present in the file (to be found by the
		      DISPLAY function). REPAIR enables to rewrite
		      the C-block according to the user specifications,
		      reculculate the averages. After that the file still
		      need not be fully consistent (RanGen state, number
		      of accepted moves). If averages do actually change after
		      REPAIR, be aware!
	*/

   qFlag = 0,  /* if =1, do calculation of Q following Salzburg et al. [23],
		i.e., using A_p(x_i) = exp(+beta*U_p) in our Eq. (12), which
		gives Q_p = C^{P/2}*L^P / <exp(beta*U_p)> .		*/
#ifdef QUANTUM
   nQ = -1,    /* if >0, do calculation of Q, using Eq. (35), at nQ	*/
	       /*     diff. values of beta to be entered interactively	*/
#endif

   PlotInt = DEFPlotInt,
		/*  Number of cycles between consecutive renderings
		    of configurations during mixing or sampling.	*/
   PlotWidth = DEFPlotWidth;
		/*  Width (in characters) of the line-printer plots	*/
static long
 cycle = 0L,	/* MC cycle counter */
 totalCycles,   /* total # of MC cycles from the initialization (very first
		   mixing); transferred across files; can be used (together
		   with the initial configuration - determined by xMin and
		   xMax and initPointFlag) to reconstruct a calculation
		   from the beginning if necessary.			*/
 MaxCycles = DEFMaxCycles,
		/* Maximum number of cycles during mixing or sampling.  */
 SamplingInt = DEFSamplingInt, /* # of cycles between consecutive samplings */
 SamplingBlk = DEFSamplingBlk, /* # of samplings over which preliminary 
		averaging is done. These averages over SamplingBlk samplings
		are accumulated in the accumulators for the calculation of the
		ensemble averages and written into the stream Output. They are
		used on the second pass for the calculation of quadratic
		standard deviations.  */
 blocks = 0L,   /* counts sampling blocks = number of quantities in the
		  accumulators.  */
 Lm_offset = 0L, /* offset for Last modification date in file  Output. */
 V_offset = 0L,  /* offset for possible use by (*inV)() and (*outV)(). */
 N_offset = 0L,  /* offset for N_fict.				       */
 P1_offset = 0L, /* offset for "constant" parameters.		       */
 P2_offset = 0L, /* offset for re-writeable parameters.		       */
 C_offset = 0L,  /* offset for current cycle and blocks values.	       */
 H_offset = 0L,  /* offset for density Histogram in the file  Output.  */
 A_offset = 0L,  /* offset of the block averages section in the
		    file  Output.   */
 offset_Incr = 0L; /* diff of offsets of two consecutive samplings in the
		      file  Output.   */

#ifdef QUANTUM
static double *betaTilde, /* nQ values of beta~ - in external units	*/
    /* Accumulators:				  see also Sample()	*/     
	      *QbetaT,    /* block contribution to the ratio of Q's; must
			     be multiplied by (beta/beta~)^{N_fict/2}	*/
	      *QbetaTS;   /* sum */
	      /* All defined in  dyndefs */
#endif

FILE *Input = NULL, *Output = NULL, *Listing = stdout;
static FILE *Density = NULL;
static char InputFileName[MAXLETT], OutputFileName[MAXLETT],
	/* Both names can be identical. Thus a Master File from a previous run
	   (mixing) can be updated: All parameters, current configuration
	   and the current content of the accumulators will be (over)written
	   at the beginning of the file, new block averages will be
	   appended at the end.
        */
	    ListingFileName[MAXLETT], /* To print out parameters, final results,
				     configurations during mixing, etc.
				     If no file is given, stdout is used.   */
	    DensityFileName[MAXLETT]; /* To output the final density histogram
				     for plotting. If no file is given, append
				     to Listing.   */

char    text[MAXLETT];  /* Auxilliary string storage */
double  beta = -1.,  /* beta = 1/(k*T);
			k - Boltzman constant, T - temperature  */
	m = -1.,     /* mass of the real particle */
	h = 1.0545887e-34,  /* Planck constant / 2pi; in Js */
#ifdef QUANTUM
	C1,          /* = N_fict * mAt / (2*pi*hAt*hAt*betaAt)      */
	C2,          /* = N_fict * mAt / (2*hAt*hAt*betaAt*betaAt)  */
	C3,          /* = 2 * C2                                    */
	NormFactor,  /* = (C1'external)^(N_fict/2) = (C1/Cx^2)^(N_fict/2)  */
#else
	NormFactor = 1.,
#endif
	C4;          /* = N_fict / (2*betaAt)			    */

#ifdef QUANTUM

int	    N_fict = DEFN_fict;    /* number of the fictitious particles */
static int  N_fict_1,		   /*   = N_fict - 1			 */
	    t_N_fict,		   /*   = 2 * N_fict			 */
	    oP_N_fict,		   /*   = N_fict + (N_fict + 1) / 2	 */
	    oM_N_fict;		   /*   = N_fict + (N_fict - 1) / 2	 */
				   /*		see dyndefs2		 */
static double *x;   /*  size defined in   dynDefs()
	    coordinates of the fictitious particles: care will be
	    taken to guarantee at all times that x[N_fict] = x[0]     */
static int initPointFlag = DEFinitPointFlag;
	 /* = FOLDED  - initial positions of the fictitious particles
			were obtained by folded on itself a linear
			row of points, so that the distance between
			the first and last particles is <= other distances
			(for even N_fict all distances are exactly the same)
	    = LINEAR  - no folding: the distance between the first and last
			particles is the sum of all other distances
	 */
static int selectionMode = DEFselectionMode;  /* Sets the way (order) of
		selecting the fictitious particle to be moved next as
		described in  help()   */
#else

int N_fict = 1;		/* this is the single classical particle         */
static double x[1];	/* coordinate of the single classical particle;
			   will be referred to as x[0]  */
#endif

/* Density of (fictitious) points */
static int    N_x = DEFN_x;	  /*  number of histogram bins		    */
static long   *dens;   /*  N_x histogram bins	| defined dynanically	*/
static float  *densA;  /*  N_x final averages	| in dynDefs()		*/
float  xMin = 1e+30, xMax = -1e+30; /*  range of coordinates to be sampled  */

/* Accumulators: */
#ifdef QUANTUM
static double
   alfa = 0., lambda = 0., gama = 0., eps_vir = 0., /* eps = 0, */
	  /* See Herman, Bruskin, Berne, J.Chem.Phys. 76, 5150 (1982) */
   alfaS = 0., lambdaS = 0., gamaS = 0., eps_virS = 0.,
	  /*  sums  */
   alfa2 = 0., lambda2 = 0., gama2 = 0., eps_vir2 = 0.,
	  /*  block averages of squares of individual samplings */
   alfaErr2 = 0., lambdaErr2 = 0., gamaErr2 = 0., eps_virErr2 = 0.;
  	  /*  variations  */
#else
static double lambda = 0., lambdaS = 0., lambda2 = 0., lambdaErr2 = 0.;
#endif
static double qInv = 0., qInvS = 0., qInv2 = 0., qInvErr2 = 0.;
	  /* inverse of config. partition function  */

/* Acceptance ratio:    */
static long   Accepted = 0;
static double MaxShft = -1.;     /* Max change in the position of a
				    particle during one MC cycle       */

/* Atomic-like values used consistently in the internal representation:     */
static double	betaAt, MaxShftAt = -1.;		/* mAt = 1, hAt = 1 */
static float	xMinAt, xMaxAt,
		xStepAt;	 /*  = (xMaxAt - xMinAt) / N_x		    */
double	Cx = 0.,  /* coord. transformation: x = Cx * xAt	   */
	CE = 0.;  /* energy transformation: E = CE * E_atomic,
					    beta = betaAt / CE	   */
	/* For harmonic oscillator: Cx = CE = 1			   */

#ifdef QUANTUM
static double MaxShftT = DEFMaxShftT, /* For moves of the whole polymer to  */
	      MaxShftTAt = -1.;	/* be done after every intT cycles. A       */
static long   AcceptedT = 0;	/* separate acceptance ratio is calculated. */
static int    intT = DEFintT;	/* When MaxShftT = 0., these moves not done.*/

#endif

/* Command Line values:  */
static int clPlotInt = -1,
	   clPlotWidth = -1,
	   clN_x = -1,
	   clpotFlag = -1;
static long clMaxCycles = -1L,
	    clSamplingInt = -1L,
	    clSamplingBlk = -1L;
static double clbeta = -1.,
	      clMaxShft = -1.;
#define P1237 (float)1.273e29
#define M1237 (float)-1.273e29
static float  p1273 = P1237, clxMin = P1237, 
	      m1273 = M1237, clxMax = M1237;
#ifdef QUANTUM
static int clN_fict = -1,
	   clselectionMode = -1,
	   clintT = -1;
static double clMaxShftT = -1.;
#endif


#ifdef DEBG1
static short int *bins;  
/*  Used in   MCcycleR1(), MCcycleR2(), MCcycleR3()  */
#endif


/*-----------------------------------------------------------------------*/
/*     Currently available potentials are defined in 1Dpotential.c       */

static int potFlag = DEFpotFlag;
	 /* equals to the number of the selected potential */

double (*V)();		/* pointer to the potential function		*/
#ifdef QUANTUM
double (*dV)();		/* pointer to dV(x)/dx				*/
double (*epsVir)();	/* pointer to the function returning
			V(x) + 0.5*(x-Shift)* dV(x)/dx  for the selected V(x),
			where Shift is nonzero for a potential defined in
			finite interval only, and is selected so that
			the virial estimator gives correct averages	  */
#endif
void   (*inV)();	/* pointers to functions for input and output   */
long   (*outV)();	/* of the parameters of the potential		*/
float  xRet1, xRet2;  /* (approximate) classical return points of the
                          potential for energy = Vminimum + 2 / beta,
                          or some other suitable limits for the generation
                          of the initial configuration and of the
                          histogram range (in atomic units)                */

extern char *potBlurb,   /* potential description for printParams()      */
	    *potLabel[], /* potential name abbreviations to be used on
			    command line [and in help() and short help]  */
	    *potHelp[];  /* more information on potentials for help()    */
extern int  numPot;      /* total number of currectly defined potentials */
extern void choosePotential();  /* and perform potential initialization  */

/*	End of definitions for the file 1Dpotential.c			 */
/*-----------------------------------------------------------------------*/
/*-----------------------------------------------------------------------*/
/* Currently available random number generators are defined in RanGen.c. */
/* They require the following declarations:				 */

int rg1Flag, rg2Flag;  /* indicate the two selected generators for
			 particle movement and configuration acceptance	*/
char *rg1 = "\0", *rg2 = "\0"; /* pointers to command line arguments
				  (for RNG's) */

#ifdef QUANTUM
static double *randNum1, *randNum2;  /*  size defined in   dynDefs2()  */
#else
static double randNum1[1], randNum2[1];
#endif
	/* buffers to store the random numbers generated by the 
	   two generators; these numbers are just enough for one MC cycle */

extern char *DEFrg1, *DEFrg2; /* default selection */

extern void
	(*refresh_rand)(),  /* refreshes the contents of randNum1[] and
			       randNum2[] using the selected generators   */
	describe_rand_gen(),  /* for help(), usage() */
	out_RnGn_defaults(),  /* for defaults	     */
	write_RnGn(), read_RnGn(),   /* binary I/O of the state vectors to
					Master File  */
	out_RnGn_state();  /* formated output of the state vectors  */

extern int init_rand_gen();  /* selects and initializes the two generators */

extern double (*RnGn1)(),  /* pointer to the first RNG to be used in the
			      single-stream situation */
	      (*RnGn2)();  /* pointer to the 2nd RNG to be used for the
			      whole polymers moves; in the single-stream
			      case, RnGn1 and RnGn2 point to the same
			      generator */

/*	End of definitions for the file RanGen.c			 */
/*-----------------------------------------------------------------------*/

/* For printing of usage & defaults, etc.:  */
static char *action[MAXaction];
	    /*  for *potLabel[] and *potHelp[]  see  1Dpotential.c  */
#ifdef QUANTUM
static char *initPoints[MAXinitPoints],
	    *SelectMode[MAXselectionMode];
#endif


/* Some functions defined below the Main:  */
 void   Initialize(), InData(), Mix(), Sample(), Final(), mult_Final(),
	displayFile(), MCcycleP1s(), MCcycleP1(),  MCcyclePC(), repairFile(),
#ifdef QUANTUM
	MCcycleP2(), MCcycleR1s(), MCcycleR1(), MCcycleR2(), MCcycleR3(),
	MCcycleR4(),
#endif
	(*MCcycle)();  /* pointer to switch between the case of a single
			  random number generator and that of two
			  independent generators, and between different
			  ways of selecting particles in a cycle */
#ifdef DEBUG
 void printAll();
#endif


void printChoices(N, descr) int N; char *descr[];
/*=============================================*/
{
int n;

PE"{"); PE *descr);
for(n = 1; n < N; n++) { PE" "); PE descr[n]); }
PE"}");
}

void usage(mode) int mode;
/*======================*/
{
PE"\7\nUsage:"); if(mode) PE"\n");
PE"  %s options [ > file]\n", progName);
if(mode)
 PE"\n  options: Action pPotential iInFile oOutFile lListFile dDensFile\n");
 else
 { PE"  options: ");
   printChoices(MAXaction, action);
   PE" iInFile oOutFile lListFile dDensFile\n");
   PE"           p");
   printChoices(numPot, potLabel);
   PE"\n");
 }
#ifdef QUANTUM
PE"           msMaxShft fNfict hNhist xXmin XXmax mcMaxCycles piPlotInt\n");
PE"           bBeta MSMaxShftT TiIntT q Q\n");
if(mode)
 PE"           pwPlotWidth siSamplInt sbSamplBlk InitConfig SelectMode\n");
 else
 { PE"           pwPlotWidth siSamplInt sbSamplBlk\n           ");
   printChoices(MAXselectionMode, SelectMode);
   PE" ");
   printChoices(MAXinitPoints, initPoints);
   PE"\n");
 }
#else
PE"           msMaxShft hNhist xXmin XXmax bBeta mcMaxCycles piPlotInt\n");
PE"           q pwPlotWidth siSamplInt sbSamplBlk\n");
#endif
PE"           r1RandGen  r2RandGen              ( -h  ?  -help -hR)\n");
}

void defaults(mode) int mode;
/*=========================*/
{
PE"Defaults:\n ");
#ifdef QUANTUM
PE"  %s  p%s  f%d  h%d  pi%d  mc%ld  si%ld  sb%ld  %s  pw0  %s\n",
        action[DEFactionFlag], potLabel[DEFpotFlag], DEFN_fict, DEFN_x,
	DEFPlotInt, DEFMaxCycles, DEFSamplingInt,
	DEFSamplingBlk, initPoints[DEFinitPointFlag],
	SelectMode[DEFselectionMode]);
#else
PE"  %s  p%s  h%d  pi%d  mc%ld  si%ld  sb%ld  pw0\n",
        action[DEFactionFlag], potLabel[DEFpotFlag], DEFN_x,
	DEFPlotInt, DEFMaxCycles, DEFSamplingInt, DEFSamplingBlk);
#endif
PE"   r1%s  r2%s", DEFrg1, DEFrg2);
#ifdef QUANTUM
PE"   MS%d  Ti%d", DEFMaxShftT, DEFintT);
#endif
PE"\n        (for mixing [%s,%s]:  mc%ld  pi%d  pw%d)\n", action[INIT],
	action[MIX], DEFMaxMix, DEFPlotIntMix, DEFPlotWidth);
PE"        (for final [%s]:  si0)\n\n", action[FINAL]);
out_RnGn_defaults(N_fict);
if(mode) { PE"Example: %s  b0.4 o93_1 h12 x0.23 X15.0 ", progName);
#ifdef QUANTUM
	   PE"%s %s ", SelectMode[MAXselectionMode - 1],
				initPoints[MAXinitPoints - 1]);
#endif
	   PE"r2halton(3)(3,5,7)(1,4)(9)\n\n");
	 }
}

void help1()
/*========*/
{
usage(0);
defaults(0);
PE"####### For more help, enter command: %s -help  and/or  %s -hR #######\n\n",
	progName, progName);
}


void help()
/*=======*/
{
			int n;
#ifdef QUANTUM
PE"\n             Quantum Monte-Carlo Simulation for a Single\n");
#else
PE"\n            Classical Monte-Carlo Simulation for a Single\n");
#endif
PE"           Particle in a One-Dimensional External Potential\n");
PE author);
usage(1);
PE"\n  Only some of these options need be given in any order. All essential\n");
PE"  parameters not supplied (as defaults, from InFile) will be prompted for.\n");
PE"  Priority: 1. Command line values, 2. InFile values, 3. Defaults.\n");
PE"      Exception: q ");
#ifdef QUANTUM
PE"and Q options");
#else
PE"option");
#endif
PE" relevant only during the initialization %s.\n",
		action[INIT]);
PE"  Tip: start the program interactively, it will ask you for all necessary\n");
PE"  parameters ommitted on the command line, then put it in the background.\n\n");
PE"  Action: %s - create new configuration, do the initial mixing, and write\n",
		action[INIT]);
PE"              the final configuration into the OutFile (\"Master File\").\n");
PE"              Every PlotInt-th cycle, plot the config'n on the ListFile.\n");
PE"          %s - read a config'n from InFile Master File, do more mixing\n",
		action[MIX]);
PE"              (MaxCycles cycles), then record the final config'n to OutFile.\n");
PE"              Every PlotInt-th cycle, plot the config'n on the ListFile.\n");
PE"          %s - read a config'n from InFile, do the proper Monte-Carlo calcu-\n",
		action[SAMPLE]);
PE"              lation with subsequent samplings appended to OutFile. After\n");
PE"              MaxCycles cycles, the density of points and the averages\n");
PE"              are calculated and written to DensFile and ListFile, respec-\n");
PE"              tively. Then go once more through OutFile and calculate\n");
PE"              the errors.\n");
PE"          %s - no more samplings, only the calculation of the final results\n",
		action[FINAL]);
PE"              from previous samplings stored in InFile. All blocks from\n");
PE"              InFile are used, and the results can be appended to files\n");
#ifdef QUANTUM
PE"              epsN, eps_virN, ..., d-epsN, ... together with Nfict\n");
PE"              (to plot the convergence with Nfict after several calculations).\n");
#else
PE"              eps, eps_vir, ..., d-eps, ...\n");
#endif
PE"          %s - a series of final calculations is done such that an initial\n",
		action[EPS]);
PE"              portion of InFile can be skipped, and the number of cycles\n");
PE"              to average over is successively increased.\n");
PE"              The results are stored in the files eps, eps_vir, ..., \n");
PE"              d-eps, ... as a function of the number of samplings.\n");
PE"          %s - display the contents of InFile Master File and exit.\n",
		action[DISPLAY]);
PE"          %s - when a system failure occurs during the output, number\n",
		action[REPAIR]);
PE"               of blocks and cycles in the C-block of the Master File may\n");
PE"               not correspond to the actual number of blocks present in\n");
PE"               in the file (to be found by the  %s  function).\n",
		action[DISPLAY]);
PE"               %s  enables to rewrite the C-block according to the user\n",
		action[REPAIR]);
PE"               specifications, reculculate the averages. After that the\n");
PE"               file still need not be fully consistent (RanGen state\n");
PE"               state, number of accepted moves). If averages do\n");
PE"               actually change after %s, something may be wrong!\n\n",
		action[REPAIR]);
PE"  Potential:  "); PE potLabel[0]); PE potHelp[0]);
for(n = 1; n < numPot; n++) {
PE"\n              "); PE potLabel[n]); PE potHelp[n]);
			    }
PE"\n\n  Beta:       1/(kT)\n\n");
#ifdef QUANTUM
PE"  MaxShft:    max displacement of fictitious particles during one move\n\n");
PE"  Nfict:      # of fictitious particles for quantum MC\n\n");
PE"  InitConfig: %s  - initially, the fictitious particles are\n",
		initPoints[LINEAR]);
PE"                     equidistant on a single line\n");
PE"              %s  - this line is folded onto itself\n\n",initPoints[FOLDED]);
PE"  MaxShftT:   max displacement for the moves of the whole polymer to be\n");
PE"  IntT:       attempted done after every IntT MC cycles. If MaxShftT = 0.,\n");
PE"              no total moves will be done.\n\n");
PE"  Nhist, Xmin, Xmax:  histogram of the density of fictitious particles\n");
#else
PE"  MaxShft:    max displacement of the particle during one move\n\n");
PE"  Nhist, Xmin, Xmax:  histogram of the particle density\n");
#endif
PE"                      with Nhist bins between Xmin and  Xmax\n\n");
PE"  PlotInt: Interval (# of cycles) between plotting of configurations.\n");
PE"           (PlotInt = 0 means no plotting of configurations)\n\n");
PE"  PlotWidth: Width (in characters) of the line-printer type plots\n");
PE"             (PlotWidth = 0 means no plotting at all)\n\n");
PE"  MaxCycles: Max number of cycles for mixing or sampling\n\n");
PE"  SamplInt: Interval (# of cycles) between samplings (if Action = %s);\n",
		action[SAMPLE]);
PE"            If Action = %s, see above\n\n", action[FINAL]);
PE"  SamplBlk: Averages over SamplBlk samplings are appended to OutFile\n");
PE"            to be used for the calculation of errors\n\n");
PE"  q:  Calculate also the absolute value of the partition function Q_P\n");
PE"      following Salzburg et al., i.e., using A_p(x_i) = exp(+beta*U_p) .\n");
PE"      Must still be multiplied by the volume of the N_fict-D space\n");
PE"      available. If action = %s or %s, these results are put into the file\n",
                action[FINAL], action[EPS]);
PE"      ");
#ifdef QUANTUM
PE"qN  or  ");
#endif
PE"qq, and the variance of their inverse into  d-qInv\n\n");
#ifdef QUANTUM
PE"  Q:  Calculate also the ratios  Q_P(beta\~)/Q_P(beta)  for a series\n");
PE"      of  beta\~  to be entered interactively. If action = %s, all these\n",
		action[EPS]);
PE"      results are put into the multicolumn file QQ\n\n");
#endif
PE"  If no DensFile is given, histogram is appended to ListFile. If no ListFile\n");
PE"  is given, standard output is used which can be redirected to a file.\n");
PE"  OutFile can be identical with InFile (if only one of the two is given,\n");
PE"  they are assumed identical). If certain command line parameters (Beta,\n");
PE"  histogram limits or # of bins, etc.) differ from those in InFile,\n");
PE"  new OutFile is required.\n\n");
defaults(1);
PE"######## For help on `RandGen` and `SelectMode`, enter: %s -hR ########\n\n",
	progName);
exit(0);
}
/* End of help() */


void helpR()
/*========*/
{
describe_rand_gen();
PE"            %s  ... used for r2 only: then a single stream\n",
		SINGLE_STREAM_STRING);
PE"                 generated by r1 is used for everything, and a call to r1 is\n");
PE"                 made only when a random number is actually needed. If both\n");
PE"                 generators are the same, they always generate two\n");
#ifdef QUANTUM
PE"                 independent streams.\n\n");
PE"  In the two-stream case, depending on the value of SelectMode (see below),\n");
PE"  Nfict/2*Nfict/Nfict/1.5*Nfict (r1) and Nfict/2*Nfict/2*Nfict/1.5*Nfict (r2)\n");
PE"  random numbers are precalculated, and put into the arrays randNum1[] (r1)\n");
PE"  and randNum2[] (r2) before every MC cycle.\n\n");
PE"                   Two Streams                          |One Stream\n");
PE"              ------------------------------------------|------------------\n");
PE"  SelectMode: %s - periodic selection of the ficticious |periodic selection\n",
		SelectMode[S1]);
PE"                   particle to be moved: r1 used for    |of the fictitious\n");
PE"                   particle movement, r2 for acceptance |particle to be\n");
PE"                   decisions. Some randNum2[i] are not  |moved\n");
PE"                   used at all if energy change dU'i<0  |\n");
PE"                   during the acceptance decision.      |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - periodic selection of ficticious     |the same as %s\n",
		SelectMode[S2], SelectMode[S1]);
PE"                   particles: random numbers are taken  |\n");
PE"                   alternately from randNum1[] and      |\n");
PE"                   randNum2[] only when actually needed |\n");
PE"                   either for particle moves or accept- |\n");
PE"                   ance decisions. Continuous blocks    |\n");
PE"                   of unused random numbers may remain  |\n");
PE"                   at the end of both randNum1[] and    |\n");
PE"                   randNum2[] as dU'i<0.                |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - periodic selection. Simulation of a  |N/A\n",
		SelectMode[S7]);
PE"                   one-stream combined generator; see   |(but will have the\n");
PE"                   P. L'Ecuyer, Comm.ACM 31, 742 (1988):|same effect as %s)\n",
		SelectMode[S1]);
PE"                   (randNum1[i]+randNum2[i]) mod 1 used |\n");
PE"                   when actually needed either for      |\n");
PE"                   particle moves or acceptance decision|\n");
PE"                   Some i may remain unused at the end. |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - random selection of particles to be  |random selection\n",
		SelectMode[S3]);
PE"                   moved: r1 for the selection, r2      |of the particle\n");
PE"                   (randNum2[2*Nfict]) alternately      |to be moved\n");
PE"                   for generating the moves and         |\n");
PE"                   acceptance decisions. Some           |\n");
PE"                   randNum2[i] not used at all.         |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - almost the same as %s, except that   |the same as %s\n",
		SelectMode[S4], SelectMode[S3], SelectMode[S3]);
PE"                   numbers are taken from randNum2[]    |\n");
PE"                   consecutively as needed (any unused  |\n");
PE"                   numbers are blocked at the end of    |\n");
PE"                   randNum2[]).                         |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - similar to %s, but r1 used for       |the same as %s\n",
		SelectMode[S5], SelectMode[S3], SelectMode[S3]);
PE"                   generating the moves and r2 both     |\n");
PE"                   for selection and acceptance.        |\n");
PE"              ------------------------------------------|------------------\n");
PE"              %s - random selection of particles:       |the same as %s\n",
		SelectMode[S6], SelectMode[S3]);
PE"                   random numbers are taken alter-      |\n");
PE"                   nately from randNum1[1.5*Nfict] and  |\n");
PE"                   randNum2[1.5*Nfict] only when actu-  |\n");
PE"                   ally needed either for selection, or |\n");
PE"                   moves, or acceptance. Any unused     |\n");
PE"                   random numbers are thus blocked at   |\n");
PE"                   the end of both randNum1[] and       |\n");
PE"                   randNum2[]  ~  to  dU'i < 0.         |\n\n");
#else
PE"                 independent streams. In the two-stream case, r1 is used for\n");
PE"                 for particle movement, r2 for configuration acceptance.\n\n");
#endif
PE"############### For more help, enter command: %s -help ###############\n\n",
	progName);
exit(0);
}
/* End of helpR() */



/*=====================================*/
/*    MAIN program starts here!!!!     */
/*=====================================*/
main(argc, argv)  int argc; char *argv[];
/*=====================================*/
{
  int  i, atoi(), n,
       illeg = 0; /* see #define ILLEGAL  */
  int  strcmp(), strncmp();
  long atol();
  char *ARG, *strcpy();
  double atof(), aux;

/*  Define the wording of the command line commands/parameters:  */
 action[INIT] = "I";
 action[MIX] = "M";
 action[SAMPLE] = "S";
 action[FINAL] = "F";
 action[EPS] = "E";
 action[DISPLAY] = "D";
 action[REPAIR] = "R";
#ifdef QUANTUM
 initPoints[FOLDED] = "Fld";
 initPoints[LINEAR] = "Lnr";
 SelectMode[S1] = "P1";
 SelectMode[S2] = "P2";
 SelectMode[S3] = "R1";
 SelectMode[S4] = "R2";
 SelectMode[S5] = "R3";
 SelectMode[S6] = "R4";
 SelectMode[S7] = "PC";
#endif

#ifdef DEBUG
 PE"\nAvailable potentials:  ");
 for(n = 0; n < numPot; n++) { PE potLabel[n]); PE"   "); }
 PE"\n");
#endif

progName = argv[0];
for (i=1; i<argc; i++)
  {
   ARG = argv[i];
#ifdef DEBUG
   PE"\n..... processing '%s'\n",ARG);
#endif
   if ( ! strcmp(ARG,"-help") || ! strcmp(ARG,"-HELP") )  help();
   if ( ! strcmp(ARG,"-hR") || ! strcmp(ARG,"-HR") )  helpR();
   if ( ! strcmp(ARG,"-h") || ! strcmp(ARG,"-H") ||
	! strcmp(ARG,"?") || ! strcmp(ARG,"-?") )  { help1(); exit(0); }

     else if(! strcmp(ARG,action[INIT]))    actionFlag = INIT;
     else if(! strcmp(ARG,action[MIX]))	    actionFlag = MIX;
     else if(! strcmp(ARG,action[SAMPLE]))  actionFlag = SAMPLE;
     else if(! strcmp(ARG,action[FINAL]))   actionFlag = FINAL;
     else if(! strcmp(ARG,action[EPS]))     actionFlag = EPS;
     else if(! strcmp(ARG,action[DISPLAY])) actionFlag = DISPLAY;
     else if(! strcmp(ARG,action[REPAIR]))  actionFlag = REPAIR;
     else if(! strncmp(ARG,"r1",2)) rg1 = ARG + 2;
     else if(! strncmp(ARG,"r2",2)) rg2 = ARG + 2;
     else if(! strncmp(ARG,"pi",2))
       { ARG += 2; if(isdigit(*ARG)) clPlotInt = atoi(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"pw",2))
       { ARG += 2; if(isdigit(*ARG)) clPlotWidth = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'p')
           { ++ARG; for(n = 0; n < numPot; n++)
			{ if(! strncmp(ARG,potLabel[n],strlen(potLabel[n])) )
						{ clpotFlag = n; break; }
			}
             if( clpotFlag == -1) ILLEGAL
           }
     else if(*ARG == 'i') strcpy(InputFileName,++ARG);
     else if(*ARG == 'o') strcpy(OutputFileName,++ARG);
     else if(*ARG == 'l') strcpy(ListingFileName,++ARG);
     else if(*ARG == 'd') strcpy(DensityFileName,++ARG);
     else if(! strcmp(ARG,"q")) qFlag = 1;
#ifdef QUANTUM
     else if(! strcmp(ARG,"Q")) nQ = 1;
     else if(! strcmp(ARG,SelectMode[S1])) clselectionMode = S1;
     else if(! strcmp(ARG,SelectMode[S2])) clselectionMode = S2;
     else if(! strcmp(ARG,SelectMode[S3])) clselectionMode = S3;
     else if(! strcmp(ARG,SelectMode[S4])) clselectionMode = S4;
     else if(! strcmp(ARG,SelectMode[S5])) clselectionMode = S5;
     else if(! strcmp(ARG,SelectMode[S6])) clselectionMode = S6;
     else if(! strcmp(ARG,SelectMode[S7])) clselectionMode = S7;
     else if(! strcmp(ARG,initPoints[LINEAR])) initPointFlag = LINEAR;
     else if(! strcmp(ARG,initPoints[FOLDED])) initPointFlag = FOLDED;
     else if(*ARG == 'f')
       { ++ARG; if(isdigit(*ARG)) clN_fict = atoi(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"Ti",2))
           { ARG += 2; if(isdigit(*ARG)) clintT = atoi(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"MS",2))
           { aux = atof(ARG + 2); if(aux >= 0.) clMaxShftT = aux; else ILLEGAL }
#endif
     else if(*ARG == 'h')
       { ++ARG; if(isdigit(*ARG)) clN_x = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'x') clxMin = (float)atof(++ARG);
     else if(*ARG == 'X') clxMax = (float)atof(++ARG);
     else if(*ARG == 'b')
           { aux = atof(++ARG); if(aux > 0.) clbeta = aux; else ILLEGAL }
     else if(! strncmp(ARG,"ms",2))
           { aux = atof(ARG + 2); if(aux > 0.) clMaxShft = aux; else ILLEGAL }
     else if(! strncmp(ARG,"mc",2))
       { ARG += 2; if(isdigit(*ARG)) clMaxCycles = atol(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"si",2))
       { ARG += 2; if(isdigit(*ARG)) clSamplingInt = atol(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"sb",2))
       { ARG += 2; if(isdigit(*ARG)) clSamplingBlk = atol(ARG); else ILLEGAL }
     else ILLEGAL
  }

if(actionFlag < SAMPLE) { PlotInt = DEFPlotIntMix;
			  MaxCycles = DEFMaxMix;
			}

if (illeg || 1 == argc) help1(); 

/* Initialize Files:  */
if(ListingFileName[0] != '\0')
 { if( (Listing = fopen(ListingFileName,"w")) == NULL)
				  FIERROR(ListingFileName)
 }
if(actionFlag > INIT)
 { if(actionFlag < FINAL || actionFlag == REPAIR)
	{ if(InputFileName[0] != '\0' && OutputFileName[0] == '\0')
			strcpy(OutputFileName,InputFileName);
	  if(InputFileName[0] == '\0' && OutputFileName[0] != '\0')
			strcpy(InputFileName,OutputFileName);
	}
   while(InputFileName[0] == '\0')
	{ PR"\nEnter input file name: "); SC"%s", InputFileName); }
 }
if(actionFlag == SAMPLE)
 { while(OutputFileName[0] == '\0')
	{ PR"\nEnter output file name: "); SC"%s", OutputFileName); }
 }
/* For actionFlag < SAMPLE, OutputFileName[] is checked in Mix() only
   when actually needed.
   Density file is opened in OutDensity().
*/

PR space); PR DateAndTime());
#ifdef QUANTUM
PR"\n\n             Quantum Monte-Carlo Simulation for a Single\n");
#else
PR"\n\n             Classical Monte-Carlo Simulation for a Single\n");
#endif
PR"           Particle in a One-Dimensional External Potential\n");
PR author); PR"\n");

#ifdef DEBUG
   printAll(0, "main() - after argument processing");
#endif

switch (actionFlag)
 {
   case INIT:    Initialize();
		 Mix();
		 break;
   case	MIX:     InData();
		 Mix();
		 break;
   case	SAMPLE:  InData();
		 Sample();
		 Final();
		 PR
		"\nSample: CPU time in seconds used for Monte Carlo proper:\n");
		 PR"     user %g, system %g, total %g\n",
			(float)(cpuT.tms_utime - usLast)/ 60.,
			(float)(cpuT.tms_stime - sysLast)/ 60.,
			(float)(cpuT.tms_utime + cpuT.tms_stime - usLast
			 - sysLast) / 60.);
		 break;
   case	FINAL:   InData();
		 Final();
		 break;
   case	EPS:     InData();
		 mult_Final();
		 break;
   case	DISPLAY: displayFile(Listing);
		 break;
   case	REPAIR:  repairFile();
		 PlotWidth = 0;
		 Final();
 }
(void)times(&cpuT);
PR"CPU time in seconds used by the whole job:\n    user  %g, system  %g,\
 total  %g\n", (float)cpuT.tms_utime / 60., (float)cpuT.tms_stime / 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime) / 60.);
PR DateAndTime()); PR"\n\n");
FC Listing);
exit(0);
}  /* ------------------------ end of Main --------------------------  */


void checkCL()
/*============         Activate all command line values that make sense  */
{
int chng = 0,
    chngH = 0,
    chngS = 0,
    i,
    same_I_O = ! strcmp(InputFileName, OutputFileName);
void printParams();

if(clPlotWidth != -1) PlotWidth = clPlotWidth;
	else if(actionFlag == SAMPLE || actionFlag == FINAL
				 || actionFlag == EPS) PlotWidth = 0;

if(actionFlag >= FINAL)
{

if(clPlotInt != -1) chng = 1;
if(clN_x != -1) chng = 1;
if(clpotFlag != -1) chng = 1;
if(clMaxCycles != -1L) chng = 1;
if(clSamplingBlk != -1L) chng = 1;
if(clbeta != -1.) chng = 1;
if(clxMin != p1273) chng = 1;
if(clxMax != m1273) chng = 1;
if(clMaxShft != -1.) chng = 1;
#ifdef QUANTUM
if(clN_fict != -1) chng = 1;
if(clselectionMode != -1) chng = 1;
if(clintT != -1) chng = 1;
if(clMaxShftT != -1.) chng = 1;
#endif
if(actionFlag == DISPLAY || actionFlag == REPAIR)
   { if(clSamplingInt != -1L) chng = 1;
     if(clPlotWidth != -1) chng = 1;
     if(chng)
	{ PR"\n When %s is specified, all other command line parameters,\n",
		action[DISPLAY]);
	  PR" except for i... l... have no effect, of course!\n");
	}
     return;
   }
if(chng)
	{ PR"\n When %s is specified, all other command line parameters,\n",
		action[FINAL]);
	  PR" except for i... l... d... pw... si... have no effect!\n");
	}

}
else
{

if(clN_x != -1) { if(actionFlag > INIT && clN_x != N_x) chngH = 1;
		  N_x = clN_x;
		}
if(clpotFlag != -1) { if(actionFlag > INIT && clpotFlag != potFlag)
				{ PR"\nPotential required on the command line\
 differs from that in input file!\n");
				  exit(-1);
				}
		      potFlag = clpotFlag;
		    }
if(clMaxCycles != -1) MaxCycles = clMaxCycles;
if(clPlotInt != -1) PlotInt = clPlotInt;
	else if(actionFlag == SAMPLE) PlotInt = 0;
if(clSamplingInt != -1)
	{ if((lambdaS != 0. || lambdaErr2 != 0.) &&
			 clSamplingInt != SamplingInt) chngS = 1;
	  SamplingInt = clSamplingInt;
	}
if(clSamplingBlk != -1)
	{ if((lambdaS != 0. || lambdaErr2 != 0.) &&
			 clSamplingBlk != SamplingBlk) chngS = 1;
	  SamplingBlk = clSamplingBlk;
	}
if(clbeta != -1.)  { if(actionFlag > INIT && clbeta != beta) chng = 1;
		     beta = clbeta;
		   }
if(clxMin != p1273 && clxMax != m1273 && clxMin < clxMax)
	{ if(actionFlag > INIT && (xMin != clxMin || xMax != clxMax)) chngH = 1;
	  xMin = clxMin; xMax = clxMax;
	}
if(clxMin == p1273 && clxMax != m1273 && xMin < clxMax)
	{ if(actionFlag > INIT && xMax != clxMax) chngH = 1;
	  xMax = clxMax;
	}
if(clxMin != p1273 && clxMax == m1273 && clxMin < xMax)
	{ if(actionFlag > INIT && xMin != clxMin) chngH = 1;
	  xMin = clxMin;
	}
if(clMaxShft != -1.) MaxShft = clMaxShft;
#ifdef QUANTUM
if(clintT != -1)  intT = clintT;
if(clMaxShftT != -1.) MaxShftT = clMaxShftT;
if(clN_fict != -1) { if(actionFlag > INIT && clN_fict != N_fict)
				{ PR"\nNumber of fictitious particles\
 required on the command line differs\n");
				  PR"from that in input file!\n");
				  exit(-1);
				}
		     N_fict = clN_fict;
		   }

if(clselectionMode != -1)
   { char dec[2];

     if(actionFlag > INIT && clselectionMode != selectionMode && same_I_O)
      { PE"Original Selection Mode in the input Master File was %s\n",
			SelectMode[selectionMode]);
	PE"Selection Mode required on the command line is %s\n",
			SelectMode[clselectionMode]);
	PE"Do you want to add new samplings to the input Master File using\
 the new\n");
	PE"Selection Mode? ... (y/n) ");
	SC"%s", dec);
	if(dec[0] != 'y') exit(-1);
	PR"\nOriginal Selection Mode in the input Master File was %s\n",
			SelectMode[selectionMode]);
      }
     selectionMode = clselectionMode;
   }
#endif

if(actionFlag > INIT && (*rg1 != '\0' || *rg2 != '\0'))
 { if( same_I_O )
	{ PR"\nYou can't change/reinitialize random number generator(s)\n");
	  PR"unless you give a new output Master File!\n");
	  PR"Old generators used!\n\n");
	  /* exit(-1); */
	  *rg1 = '\0';
	  *rg2 = '\0';
	}
   else { PR"\n");
	  if(*rg1 != '\0')
		PR"1st random number generator changed/reinitialized!\n");
	  if(*rg2 != '\0')
	      { if(rg2Flag >= 0)
		  if(strcmp(rg2,SINGLE_STREAM_STRING))
		     PR"2nd random number generator changed/reinitialized!\n");
		  else
		     PR"Switch from two streams to a single stream!\n");
		if(rg2Flag < 0 && strcmp(rg2,SINGLE_STREAM_STRING))
		  PR"Switch from single stream to two streams!\n");
	      }
	  PR"\n");
	}
 }

if( same_I_O )
 {
   if(actionFlag == MIX && (lambdaS != 0. || lambdaErr2 != 0.))
	{ PR"\n Input file contains a completed calculation.\n");
	  PR" You can't corrupt it by further mixing!\n");
	  PR" New output Master File must be given to store the output from\
 mixing.\n\n");
	  exit(-1);
	}
   if(chng)
	{ PR"\n \'beta\' from the command line differs from that in input\
 file.\n");
	  PR" New output file must be given!\n\n");
	  exit(-1);
	}
   if(chngH)
	{ PR"\n Histogram parameters from the command line differ\
 from those in input file.\n");
	  PR" New output file must be given!\n\n");
	  exit(-1);
	}
  if(chngS && actionFlag == SAMPLE)
	{ PR"\n \'SamplingBlk\' or \'SamplingInt\' from the command line\
 differs from that\n");
	  PR" in input file. New output file must be given!\n\n");
	  exit(-1);
	}
 }

if(actionFlag == SAMPLE)
 {
  i = SamplingBlk * SamplingInt;
  MaxCycles = max(i, ((MaxCycles + i / 2) / i) * i);
 }

} /* ----  end of:  if(actionFlag >= FINAL)  ... else ...  ----  */

if(actionFlag == INIT)
  { while(beta <= 0.0)
	{ PE"\7Enter positive value of beta: "); SC"%lf", &beta); }

    choosePotential(potFlag, 0);  /* among others, it calculates
		      Cx and CE, and C1, C2 and C3 in the quantum case */

    if(MaxShft < 0.0) { MaxShftAt = 0.25 * (xRet2 - xRet1);
			MaxShft = MaxShftAt * Cx;
		      }
  }
if(MaxShftAt < 0.) MaxShftAt = MaxShft / Cx;
#ifdef QUANTUM
if(MaxShftTAt < 0.) MaxShftTAt = MaxShftT / Cx;
#endif

while(xMax <= xMin)
	{ PR"\nFor density histogram, enter xMin < xMax: ");
	  SC"%f%f", &xMin, &xMax); }

if(actionFlag == INIT || chngH)
 { xMaxAt = xMax / Cx;
   xMinAt = xMin / Cx;
   xStepAt = (xMaxAt - xMinAt) / N_x;
   betaAt = beta * CE;
 }

if(actionFlag < FINAL)
      { if( init_rand_gen(&rg1, &rg2, actionFlag, N_fict) ) exit(-1);
#ifdef QUANTUM
	if(strcmp(rg2, SINGLE_STREAM_STRING))
	    switch (selectionMode)
		 { case S1: MCcycle = MCcycleP1;
                            break;
		   case S2: MCcycle = MCcycleP2;
			    break;
		   case S3: MCcycle = MCcycleR1;
                            break;
		   case S4: MCcycle = MCcycleR2;
                            break;
		   case S5: MCcycle = MCcycleR3;
                            break;
		   case S6: MCcycle = MCcycleR4;
                            break;
		   case S7: MCcycle = MCcyclePC;
                            break;
		 }
	else
	    switch (selectionMode)
		 { case S7:
		   case S1:
		   case S2: MCcycle = MCcycleP1s;
			    break;
		   case S3: 
		   case S4: 
		   case S5: 
		   case S6: MCcycle = MCcycleR1s;
                            break;
		 }
#else
	  MCcycle = strcmp(rg2, SINGLE_STREAM_STRING) ? MCcycleP1 : MCcycleP1s;
#endif
	}

#ifdef QUANTUM
NormFactor = pow(C1 / (Cx * Cx), 0.5 * N_fict);
#endif

}  /* ------------------- end of checkCL() -----------------------  */


#ifdef DEBUG
void printAll(mode, msg)  int mode; char *msg;
/*==========================================*/
{ 
 /*  Current state of the selected RNG is printed only if mode != 0  */

 int i;

 PE"\nGLOBAL PARAMETERS ("); PE msg); 
 PE"):\nactionFlag = %d, potFlag = %d\n", actionFlag, potFlag);
 PE"beta = %g, betaAt = %g\n", beta, betaAt);
 PE"xMin = %g, xMinAt = %g, xMax = %g, xMaxAt = %g\n",
	 xMin, xMinAt, xMax, xMaxAt);
 PE"N_x = %d, xStepAt = %g\n", N_x, xStepAt);
 PE"Cx = %g, CE = %g, m = %g\n", Cx, CE, m);
#ifdef QUANTUM
 PE"Initial configuration: %s\n", initPoints[initPointFlag]);
 PE"C1 = %g, C2 = %g, C3 = %g, ", C1, C2, C3);
#endif
 PE"C4 = %g\n", C4);
 PE"MaxShft = %g, MaxShftAt = %g\n", MaxShft, MaxShftAt);
 PE"MaxCycles = %ld, SamplingInt = %ld, SamplingBlk = %ld, PlotInt = %ld\n",
	 MaxCycles, SamplingInt, SamplingBlk, PlotInt);
 PE"cycle = %ld, blocks = %ld, Accepted = %ld\n", cycle, blocks, Accepted);
 PE"totalCycles = %ld\n", totalCycles);

 PE"N_fict = %d\n", N_fict);
 PE"qFlag = %d\n", qFlag);
#ifdef QUANTUM
 PE"nQ = %d\n", nQ);
 if(MaxShftT > 0.) { PE"Whole-polymer moves are ON:\n");
		     PE"	MaxShftT = %g, MaxShftTAt = %g\n",
						MaxShftT, MaxShftTAt);
		     PE"	intT = %d\n", intT);
		     PE"	AcceptedT = %ld\n", AcceptedT);
		   }
 PR"Selection mode = %s\n", SelectMode[selectionMode]);
#endif
 if(mode) { PE"x:"); for(i = 0; i < N_fict; i++) PE" %g", x[i]); PE"\n"); }
 if(actionFlag > MIX)
  { PE"lambda = %g, lambda2 = %g\n", lambda, lambda2);
#ifdef QUANTUM
    PE"alfa = %g, gama = %g, eps_vir = %g\n", alfa, gama, eps_vir);
    PE"alfa2 = %g, gama2 = %g, eps_vir2 = %g\n", alfa2, gama2, eps_vir2);
#endif
    PE"qInv = %g, qInv2 = %g\n", qInv, qInv2);
  }

 if(mode) { PE"dens:"); for(i = 0; i < N_x; i++) PE" %ld", dens[i]); PE"\n"); }
 if(actionFlag == FINAL || actionFlag == EPS
			|| lambdaErr2 != 0. || lambdaS != 0.)
  {  PE"lambdaS = %g, lambdaErr2 = %g\n", lambdaS, lambdaErr2);
#ifdef QUANTUM
     PE"alfaS = %g, gamaS = %g, eps_virS = %g\n", alfaS, gamaS, eps_virS);
     PE"alfaErr2 = %g, gamaErr2 = %g, eps_virErr2 = %g\n",
		 alfaErr2, gamaErr2, eps_virErr2);
#endif
    PE"qInvS = %g, qInvErr2 = %g\n", qInvS, qInvErr2);
  }

 PE"rg1Flag = %d,  rg2Flag = %d\n", rg1Flag, rg2Flag);
 out_RnGn_state(mode, stderr);

 if( actionFlag == INIT || strcmp(OutputFileName,InputFileName) )
  { PE"\nLm_offset = %ld, V_offset = %ld, N_offset = %ld\n",
	 Lm_offset, V_offset, N_offset);
    PE"P1_offset = %ld, P2_offset = %ld, C_offset = %ld\n",
	 P1_offset, P2_offset, C_offset);
    PE"H_offset = %ld, A_offset = %ld, offset_Incr = %ld",
	 H_offset, A_offset, offset_Incr);
  }

 PE"\n\n");
}
#endif


void err_fseek(proc)  char *proc;
/*=============================*/
{  PR"\7\n"); PR proc);  PR": fseek error\n");  exit(-1);  }

void err_fread(proc)  char *proc;
/*=============================*/
{  PR"\7\n");
   PR proc);
   if(feof(Input))  PR": End of Input file\n");
	else	    PR": fread error\n");
   exit(-1);
}

#define SEEKBEG(F,o) if( fseek(F, o, SEEK_SET) )  err_fseek(p);

void write_offsets()
/*================*/
{
 WL(&Lm_offset, 1)  WL(&V_offset, 1)  WL(&N_offset, 1)  WL(&P1_offset, 1) 
 WL(&P2_offset, 1)  WL(&C_offset, 1)  WL(&H_offset, 1)  WL(&A_offset, 1) 
 WL(&offset_Incr, 1) 
}

void read_offsets(p) char *p;
/*=========================*/
{
 RL(&Lm_offset, 1)  RL(&V_offset, 1)  RL(&N_offset, 1)  RL(&P1_offset, 1) 
 RL(&P2_offset, 1)  RL(&C_offset, 1)  RL(&H_offset, 1)  RL(&A_offset, 1) 
 RL(&offset_Incr, 1) 
}

void read_head(p) char *p;
/*======================*/
{
int version;

#ifdef QUANTUM
char *cl = "CLASSICAL";
#else
char *qu = "QUANTUM";
#endif

 SEEKBEG(Input, Lm_offset)
 while(getc(Input) != '\n') ;
 SI"%s %*s %*s %*s Version # %d\n\n", text, &version);
#ifdef DEBUG
 PE"%s    Version # %d\n", text, version);
#endif
#ifdef QUANTUM
 if( ! strcmp(text,cl) )
  { PR" Quantum version of the program cannot handle a CLASSICAL input file\n");
    exit(-1);
  }
#else
 if( ! strcmp(text,qu) )
  { PR" Classical version of the program cannot handle a QUANTUM input file\n");
    exit(-1);
  }
#endif

 if(version != VERSION)
	{ PE"\nUse %s version %d to handle this file!!!\n\n",
						 progName, version);
	  exit(0);
	}

 SI"Potential # %d", &potFlag);
 choosePotential(potFlag, 1);
 (*inV)(V_offset);
} /* read_head() */

void write_N()
/*==========*/
{
#ifdef QUANTUM
 PO"N_fict = %d\n", N_fict);
#endif
 PO"xMin = %.8g, xMax = %.8g, N_x = %d\n", xMin, xMax, N_x);
 PO"rg1Flag = %d,  rg2Flag = %d\n", rg1Flag, rg2Flag);
 PO"qFlag = %d", qFlag);
#ifdef QUANTUM
 PO", nQ = %d\n", nQ);
 PO"%d:  %s", initPointFlag, initPoints[initPointFlag]);
#endif
 PO"\n");
}

void read_N(p) char *p;
/*===================*/
{
 SEEKBEG(Input, N_offset)
#ifdef QUANTUM
 SI"N_fict = %d\n", &N_fict);
#endif
 SI"xMin = %*f, xMax = %*f, N_x = %d\n", &N_x);
 SI"rg1Flag = %d,  rg2Flag = %d\n", &rg1Flag, &rg2Flag);
#ifdef QUANTUM
 SI"qFlag = %d, nQ = %d\n", &qFlag, &nQ);
 SI"%d:  %*s\n", &initPointFlag);
#else
 SI"qFlag = %d\n", &qFlag);
#endif
}

void write_P1()
/*===========*/
{
 WD(&beta, 1)  WD(&betaAt, 1)  WD(&Cx, 1)  WD(&CE, 1)  
 WF(&xMin, 1)  WF(&xMax, 1)  WF(&xMinAt, 1)  WF(&xMaxAt, 1) 
 WF(&xStepAt, 1)  WD(&C4, 1) 
#ifdef QUANTUM
 WD(&C1, 1)  WD(&C2, 1)  WD(&C3, 1)  
 if(nQ > 0) { WD(betaTilde, nQ) }
#endif
}

void read_P1(p) char *p;
/*====================*/
{
 SEEKBEG(Input, P1_offset)
 RD(&beta, 1)  RD(&betaAt, 1)  RD(&Cx, 1)  RD(&CE, 1) 
 RF(&xMin, 1)  RF(&xMax, 1)  RF(&xMinAt, 1)  RF(&xMaxAt, 1) 
 RF(&xStepAt, 1)  RD(&C4, 1) 
#ifdef QUANTUM
 RD(&C1, 1)  RD(&C2, 1)  RD(&C3, 1)  
 if(nQ > 0) { RD(betaTilde, nQ) }
#endif
}

void write_P2()
/*===========*/
{
 WD(&MaxShft, 1)  WD(&MaxShftAt, 1)  WL(&MaxCycles, 1) 
 WL(&SamplingInt, 1)  WL(&SamplingBlk, 1)  WI(&PlotInt, 1) 
#ifdef QUANTUM
 WD(&MaxShftT, 1)  WD(&MaxShftTAt, 1)
 WI(&intT, 1)
#endif
}

void read_P2(p) char *p;
/*====================*/
{
 SEEKBEG(Input, P2_offset)
 RD(&MaxShft, 1)  RD(&MaxShftAt, 1)  RL(&MaxCycles, 1) 
 RL(&SamplingInt, 1)  RL(&SamplingBlk, 1)  RI(&PlotInt, 1) 
#ifdef QUANTUM
 RD(&MaxShftT, 1)  RD(&MaxShftTAt, 1)
 RI(&intT, 1)
#endif
}

void write_C()
/*==========*/
{
#define MCYCLE   2140000000L
#define FORMAT   "cycle = %11ld\nblocks = %11ld\n"
#define TFORMAT  "totalCycles = %11ld"
#define WARNING  { PR"\ncycle > MCYCLE --> cycle approaches allowable \
maximum!\n");\
 PR" For larger values, switch to double cycle, etc., and adjust (T)FORMAT\n");\
 PR" and all other formats involving cycle, etc., accordingly!\n");\
 PR" Don't update old files using a wider new format!\n\n"); exit(-1); }

 if(cycle > MCYCLE)  WARNING
 PO FORMAT, cycle, blocks);
#ifdef QUANTUM
 PO"selectionMode = %2d\n", selectionMode);
#endif
 PO TFORMAT, totalCycles);
}

void read_C(p) char *p;
/*===================*/
{
 SEEKBEG(Input, C_offset)
 SI"cycle = %ld\nblocks = %ld\n", &cycle, &blocks);
#ifdef QUANTUM
 SI"selectionMode = %d\n", &selectionMode);
#endif
 SI"totalCycles = %ld", &totalCycles);
}

void write_H_short()
/*================*/
{
 write_RnGn();
 WL(&Accepted, 1) 
 WD(x, N_fict)  WL(dens, N_x)  WD(&lambdaS, 1) 
#ifdef QUANTUM
 WL(&AcceptedT, 1) 
 WD(&alfaS, 1)  WD(&gamaS, 1)  WD(&eps_virS, 1) 
 if(nQ > 0) { WD(QbetaTS, nQ) }
#endif
 if(qFlag) { WD(&qInvS, 1) }
}

void write_H()
/*==========*/
{
 write_H_short();
 WD(&lambdaErr2, 1) 
#ifdef QUANTUM
 WD(&alfaErr2, 1)  WD(&gamaErr2, 1)  WD(&eps_virErr2, 1) 
#endif
 if(qFlag) { WD(&qInvErr2, 1) }
}

void read_H(p) char *p;
/*===================*/
{
 SEEKBEG(Input, H_offset)
 read_RnGn(p);
 RL(&Accepted, 1) 
 RD(x, N_fict)  RL(dens, N_x)  RD(&lambdaS, 1) 
#ifdef QUANTUM
 RL(&AcceptedT, 1) 
 RD(&alfaS, 1)  RD(&gamaS, 1)  RD(&eps_virS, 1) 
 if(nQ > 0) { RD(QbetaTS, nQ) }
#endif
 if(qFlag) { RD(&qInvS, 1) }
 RD(&lambdaErr2, 1) 
#ifdef QUANTUM
 RD(&alfaErr2, 1)  RD(&gamaErr2, 1)  RD(&eps_virErr2, 1) 
#endif
 if(qFlag) { RD(&qInvErr2, 1) }
}

void write_A()
/*==========*/
{
/* Note the value of  cycle  at the beginning of each A block is always
   written separately prior to the call to write_A()
*/ 
 WD(&lambda, 1)  WD(&lambda2, 1) 
#ifdef QUANTUM
 WD(&alfa, 1)  WD(&gama, 1)  WD(&eps_vir, 1) 
 WD(&alfa2, 1)  WD(&gama2, 1)  WD(&eps_vir2, 1) 
 if(nQ > 0) { WD(QbetaT, nQ) }
#endif
 if(qFlag) { WD(&qInv, 1)  WD(&qInv2, 1) }
}

void read_A(p) char *p;
/*===================*/
{
 RD(&lambda, 1)  RD(&lambda2, 1) 
#ifdef QUANTUM
 RD(&alfa, 1)  RD(&gama, 1)  RD(&eps_vir, 1) 
 RD(&alfa2, 1)  RD(&gama2, 1)  RD(&eps_vir2, 1) 
 if(nQ > 0) { RD(QbetaT, nQ) }
#endif
 if(qFlag) { RD(&qInv, 1)  RD(&qInv2, 1) }
}

void write_Lm(p) char *p;
/*=====================*/
{
 SEEKBEG(Output, Lm_offset)
 PO space43);
 SEEKBEG(Output, Lm_offset)
 PO DateAndTime());
}


void OutData()
/*==========*/
{
int out = 0;  /* out will be set to 1 only if an output file name is entered
		 interactively  */
char *p = "OutData";
char over[2];

#ifdef DEBUG
 printAll(1, "beginning of OutData()");
#endif

while(OutputFileName[0] == '\0')
	{ out = 1;
	  PE"\7Enter output file name: ");
	  SC"%s", OutputFileName);
	}
	     
if(actionFlag == INIT || strcmp(InputFileName,OutputFileName))
{

 /* Here, an output file is being built from scratch: */

Try:
 if( (Output = fopen(OutputFileName,"a")) == NULL)
				  FIERROR(OutputFileName)
 if(ftell(Output) != 0L)
	{ PE"\7File %s exist! Overwrite? (y/n) ... ", OutputFileName);
	  SC"%s", over);
	  if(over[0] != 'y') {  PE"\7Enter new output file name: ");
				SC"%s", OutputFileName);
				out = 1;
				goto Try;
			     }
	}
 FC Output);
/* if( (Output = fopen(OutputFileName,"wb+")) == NULL) */
 if( (Output = fopen(OutputFileName,"w+")) == NULL)
				  FIERROR(OutputFileName)
 if(out) PR"\nOutput file: %s\n\n",OutputFileName);


 write_offsets();    /* At this point, this is just to hold place.
			The real offsets will be obtained later.     */
 PO"\nThe first line above contains all needed offsets (in binary).\n");
 PO"Most data below are also in binary.\n\n");

 PO"Created: ");  PO DateAndTime());
 PO"\nLast modified: ");
 Lm_offset = ftell(Output);
 PO space43);   /* to hold 43 bytes for the last-modification date */
#ifdef QUANTUM
 PO"\n\nQUANTUM");
#else
 PO"\n\nCLASSICAL");
#endif
 PO" 1-PARTICLE MONTE-CARLO SIMULATION");
 PO"       Version # %d\n\n", VERSION);

 PO"Potential # %d\n", potFlag);
 V_offset = (*outV)(Output);
 PO"\nbeta = %.18g\n", beta);

 N_offset = ftell(Output);
 write_N();

 PO"\nP1:\n");
 P1_offset = ftell(Output);
 write_P1();

 PO"\n\nP2:\n");
 P2_offset = ftell(Output);
 write_P2();

 PO"\n\nC:\n");
 C_offset = ftell(Output);
 write_C();

 PO"\n\nH:\n");
 H_offset = ftell(Output);
 write_H();
 /* Here, lambdaS, lambdaErr2, etc. are written out to hold the place for
      future updates.
    lambdaS, etc. will contain the sums of the lambda's for all block
      averages already written into the file.
    lambdaErr2, etc. will be the rms errors calculated in the final run.
    Before some additional sampling blocks are added to the file in
      a later run, the quantitities lambdaErr2, alfaErr2, gamaErr2, qInvErr2,
      and eps_virErr2 are all reset to 0 in Sample().

    The configuration is needed to be able to restart the calculation
    in case of a crash.
 */

 /* Block averages at the end of each block: */
 PO"\n\nA:\n");
 A_offset = ftell(Output);
 WL(&totalCycles, 1)	/* 0-th blocks holds the total number of cycles
			   done before any block averages stored in the
			   given file were being calculated.  */
 write_A();
 /* Again, at this time, the block averages lambda, etc., were written out
    to hold the space so as to enable to calculate offset_Incr that is
    equal to the number of bytes needed to write all data for one sampling
    block. Block averages lambda, ... are stored to be able to calculate the
    rms errors in the final run.
 */
 offset_Incr = ftell(Output) - A_offset;

 /* Now, the actual values of offsets are written into the reserved space at
    the beginning of the Output:  */
 SEEKBEG(Output, 0L)
 write_offsets();

#ifdef DEBUG
 PR"\noffset_Incr = %ld bytes\n\n", offset_Incr);
#endif

}
else
{

 /* Here, an existing output file is updated with the results of additional
    mixing, or prepared for sampling or for adding additional samplings:  */

/* if( (Output = fopen(OutputFileName,"rb+")) == NULL) */
 if( (Output = fopen(OutputFileName,"r+")) == NULL)
				  FIERROR(OutputFileName)
 if(out) PR"\nOutput file: %s\n\n",OutputFileName);
 setbuf(Output, (char *)NULL);

 SEEKBEG(Output, P2_offset)
 write_P2();

 SEEKBEG(Output, C_offset)
 write_C();

 SEEKBEG(Output, H_offset)
 write_H();

 if(actionFlag == MIX)
  { SEEKBEG(Output, A_offset)
    WL(&totalCycles, 1)	/* 0-th blocks holds the total number of cycles
			   done before any block averages stored in the
			   given file were being calculated.  */
  }
}

 write_Lm(p);

 if(actionFlag <= MIX)  FC Output);

#ifdef DEBUG
 printAll(1, "end of OutData()");
#endif

}  /*  ---------------  end of OutData() ----------------  */


void dynDefs()
/*==========*/
{
  dens = (long int *)malloc((unsigned)(N_x * sizeof(long int)));
  densA = (float *)malloc((unsigned)(N_x * sizeof(float)));

#ifdef QUANTUM
  if(N_fict < 2) { PR"\nNumber of fictitious particles increased to 2!\n\n");
		   N_fict = 2;
		 }
  x = (double *)malloc((unsigned)((N_fict + 1) * sizeof(double)));
  if(nQ > 0) {
	betaTilde = (double *)malloc((unsigned)(nQ * sizeof(double)));
	QbetaT = (double *)malloc((unsigned)(nQ * sizeof(double)));
	QbetaTS = (double *)malloc((unsigned)(nQ * sizeof(double)));
  }
#endif

#ifdef DEBG1
 if(actionFlag < FINAL)
	  bins = (short int *)
		 malloc((unsigned)((N_fict + 1) * sizeof(short int)));
  /*  Used in   MCcycleR1(), MCcycleR2(), MCcycleR3()  */
#endif
}


#ifdef QUANTUM

void dynDefs2()
/*===========*/
{
  int d1, d2;

  N_fict_1 = N_fict - 1;
  t_N_fict = 2 * N_fict;
  oP_N_fict = N_fict + (N_fict + 1) / 2;
  oM_N_fict = 3 * N_fict - oP_N_fict;
  if(MCcycle == MCcycleP1s || MCcycle == MCcycleR1s)      d1 = d2 = 0;
    else if(MCcycle == MCcycleP1 || MCcycle == MCcycleP2) d1 = d2 = N_fict;
    else if(MCcycle == MCcyclePC)			  d1 = d2 = t_N_fict;
    else if(MCcycle == MCcycleR1 || MCcycle == MCcycleR2 ||
	    MCcycle == MCcycleR3)	{  d1 = N_fict; d2 = t_N_fict;  }
    else if(MCcycle == MCcycleR4)	{  d1 = oP_N_fict; d2 = oM_N_fict;  }
    else  {  PR"\nMCcycle undefined in  dynDefs2() !!!\n\n"); exit(-1);  }
  if(d1) randNum1 = (double *)malloc((unsigned)(d1 * sizeof(double)));
  if(d2) randNum2 = (double *)malloc((unsigned)(d2 * sizeof(double)));
}

void betaTout(bp,Nline) double *bp; int Nline;     /* used with the Q option */
/*==========================================*/
/* Nline  is the # of elements of the array  bp  to be printed on one line */
{
  int n;

  PR" %g", *bp);
  for(n = 1; ; n++)
	{ if(nQ == n) { PR"\n"); return; }
	  if(n % Nline == 0) PR"\n    ");  else PR",");
	  PR" %g", bp[n]);
	}
}

#endif  /* QUANTUM */


void InData()
/*=========*/
{
#define DIM     5
char *p = "InData";
/* static char *mode[DIM] = { "rb", "rb", "rb", "rb+", "rb" }; */
static char *mode[DIM] = { "r", "r", "r", "r+", "r" };
long ccl;

#ifdef DEBUG
 PE"\nIndata() started\n\n");
#endif

 if(MAXaction != (DIM + 2))
  { PR"\nCheck DIM in InData() and the initialization of *mode[]\n\n");
    exit(-1);
  }
#undef DIM

 if( (Input = fopen(InputFileName,mode[actionFlag])) == NULL)
				  FIERROR(InputFileName)
 setbuf(Input, (char *)NULL);

 read_offsets(p);
 read_head(p);
 read_N(p);
 dynDefs();
 read_P1(p);
 read_P2(p);
 read_C(p);
 read_H(p);
#ifdef QUANTUM
 x[N_fict] = x[0];
#endif

 if(blocks != 0) {
   SEEKBEG( Input, A_offset + blocks * offset_Incr )
   /* This is to ensure the correct continuation of an interrupted job  */
   RL(&ccl, 1)
   if(ccl != cycle)
	{ PR"\nInput file corrupted: the two values of \'cycle\' differ!\n\n");
	  exit(-1);
	}
 }
 if(actionFlag < FINAL) FC Input);

#ifdef DEBUG
 printAll(1, "input finished in Indata()");
#endif

 checkCL();
#ifdef QUANTUM
 if(actionFlag < FINAL) dynDefs2();
#endif

#ifdef DEBUG
 if(actionFlag < FINAL) printAll(1, "end of Indata(): after checkCL");
#endif

/* Print out all parameters, plot potential: */
printParams();

}  /*  ------------- end of InData() ----------------  */


void ConfigHead()
/*=============*/
{
if(PlotInt) {
#ifdef QUANTUM
      PR"\n\n Cycle                  Configuration of fictitious particles\n");
#else
      PR"\n\n Cycle                           Particle position\n");
#endif
	    }
}


void printParams()
/*==============*/
/* Print out all important parameters at the beginning of the listing.  */
{
 static char *fmt = "%8ld ", *fmtV = "%9g";
 static int  fmt_len = 9;
 int i;

 PR potBlurb); (void)(*outV)(Listing);
 PR"Beta = %.8g\n", beta);
#ifdef QUANTUM
 PR"No. of fictitious particles = %d\n", N_fict);
 PR"Selection mode = %s\t\t", SelectMode[selectionMode]);
#endif
 PR"Max displacement = %.8g\n", MaxShft);
#ifdef QUANTUM
 if(MaxShftT > 0.) { PR"Whole-polymer moves are ON:");
		     PR"  MaxShftT = %.8g,  intT = %d\n", MaxShftT, intT);
		   }
#endif
 if (actionFlag == INIT) { PR"Mixing: at least %ld cycles.\n", MaxCycles);
#ifdef QUANTUM
			   PR"Initial configuration: %s\n",
				initPoints[initPointFlag]);
#endif
			 }
 if (actionFlag == MIX) { PR"Mixing: previously %ld cycles, now up to\
 %ld cycles.\n", totalCycles, MaxCycles);
#ifdef QUANTUM
			  PR"Initial configuration was: %s\n",
				initPoints[initPointFlag]);
#endif
			}
 if (actionFlag == SAMPLE)
	{ PR"Sampling: Total # of MC cycles to be sampled = %ld.\n", MaxCycles);
	  PR"Sampling will be done every ");
	  if(SamplingInt == 1) PR"cycle.");
		else	       PR"%ld cycles.", SamplingInt);
	  if(SamplingBlk == 1) PR"\nEvery sampling");
		else	       PR"\nThe average of every %ld samplings",
				   SamplingBlk);
	  PR" will be appended to the output file.\n");
	}
 if (actionFlag == FINAL || actionFlag == EPS)
	{ PR"Final results:\nSampling was done every ");
	  if(SamplingInt == 1) PR"cycle.");
		else	       PR"%ld cycles.", SamplingInt);
	  if(SamplingBlk == 1) PR"\nEvery sampling");
		else	       PR"\nThe average of every %ld samplings",
				   SamplingBlk);
	  PR" was recorded in the input file.\n");
	  PR"Total number of records (sampling blocks) is %ld, which\
 corresponds\n", blocks);
	  PR"to %ld MC cycles.\n", cycle);
	}
 if(qFlag) PR"Absolute value of Q_P is being calculated\n");
#ifdef QUANTUM
 if(nQ > 0) {
   PR"Relative values of Q_P are calculated at %d values of\n beta\~ =", nQ);
   betaTout(betaTilde, 10);
 }
#endif
 if(actionFlag > INIT) PR"Input file: %s\n",InputFileName);
 if(actionFlag < FINAL && OutputFileName[0] != '\0')
			PR"Output file: %s\n",OutputFileName);
 if( (actionFlag == SAMPLE || actionFlag == FINAL) &&
	DensityFileName[0] != '\0') PR"Density file: %s\n",DensityFileName);
 PR"\n");

 out_RnGn_state(0, Listing);
 /* Initialize line-printer type plotting routines:  */
 if(PlotWidth) initPlot((float)Cx, (float)CE, xMinAt, xMaxAt, N_fict,
	  PlotWidth, fmt, fmt_len, Listing);

 /* Plot potential:    */
 if(actionFlag < FINAL && PlotWidth)
    { PR"\n");
      for(i = -2 * fmt_len; i < PlotWidth; i += 2) PR" ");
      PR"V(x)\n");
      plotFun1(V, fmtV);
      ConfigHead();
    }
} 


void Initialize()
/*=============*/
{
 int i;
#ifdef QUANTUM
 int Nfh;
 float shift, xStep;
#endif

#ifdef QUANTUM
 if(nQ > 0) {
	T_a:
	PE"Enter the # of beta\~ for the calculation of Q_P(beta\~)/Q_P(beta) ... ");
	SC"%d", &nQ);
	if(nQ <=0) {
		PE"Must be positive, try again ...");
		goto T_a;
	}
 }
#endif
 checkCL();
 dynDefs();
#ifdef QUANTUM
 dynDefs2();
 for(i = 0; i < nQ; i++) { double aux; PE"Enter beta\~ # %d ... ", i + 1);
			   SC"%lf", betaTilde + i);
 }
#endif

/* Print out all parameters, plot potential: */
printParams();

 for(i = 0; i < N_x; i++) dens[i] = 0L;

 /* Generate initial configuration: */
 /* xRet1 and xRet2 are used only here; always in atomic units.
    They are set by (*inV)()					 */
 totalCycles = 0L;
#ifdef QUANTUM
 x[0] = x[N_fict] = xRet1; /* We keep one more point identical with the
      first one to make simpler the summation of (x[i+1]-x[i])^2.   */	
 if(initPointFlag == FOLDED)
	{ Nfh = N_fict / 2;
	  xStep = (xRet2 - xRet1) / Nfh;
	  for (i = 1; i <= Nfh; i++) x[i] = xRet1 + i * xStep;
	  if( (N_fict % 2) )
		{ shift = xRet2 - xStep * 0.5;
		  for (i = 0; i < Nfh; i++) x[i + Nfh + 1] = shift - i * xStep;
		}
	    else
		for (i = 1; i < Nfh; i++) x[i + Nfh] = xRet2 - i * xStep;
        }
   else
	{ xStep = (xRet2 - xRet1) / (N_fict - 1);
	  for (i = 1; i < N_fict; i++) x[i] = xRet1 + i * xStep;
	}
#else
 x[0] = 0.5 * (xRet1 + xRet2);
#endif

 C4 = N_fict / (2 * betaAt);
#ifdef DEBUG
 printAll(1, "end of Initialize()");
#endif

 }


void updateDens()
/*=============*/
{
register int i;
register int ind;

for(i = 0; i < N_fict; i++)
	{ ind = (x[i] - xMinAt) / xStepAt;
	  if(ind >= 0 && ind < N_x) ++(dens[ind]);
	}
}


void OutDensity()
/*=============*/
{
int i;
float x;

if(DensityFileName[0] != '\0')
 { if( (Density = fopen(DensityFileName,"w")) == NULL)
				  FIERROR(DensityFileName)
 }
else
 { Density = Listing;
   PD"\nDensity: Numerical values for finer plotting:\n");
   PD"         x                   dens(x)");
 }

x = xMinAt + 0.5 * xStepAt;
PD"\n %.18g  %.18g\n", Cx * x, densA[0]);
for(i = 1; i < N_x; i++)
  { x += xStepAt; PD" %.18g  %.18g\n", Cx * x, densA[i]); }

if(Density != Listing) FC Density);

}  /* --------------->  end of:  OutDensity()  <------------------ */


void MCcycleP1s()
/*=============*/
{
 /* Single stream using 1st RNG, (*RnGn1)().		 */
 /* Periodic selection of the next particle to be moved. */

#ifdef QUANTUM
 register int i;
#endif
 register double Xnew;
 register double deltaX;
 register double deltaU;
#ifdef DEBG1
 double xold = x[0];
 int j;
#endif

 deltaX = MaxShftAt * (2. * (*RnGn1)() - 1.);
 Xnew = x[0] + deltaX;
#ifdef QUANTUM
 deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
	  + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
#else
 deltaU = (*V)(Xnew) - (*V)(x[0]);
#endif
 if( deltaU < 0 || ((*RnGn1)() < exp(-betaAt * deltaU)) )
			{ x[0] = Xnew; Accepted++; }
#ifdef DEBG1
 PE"x: %g", xold); for(j = 1; j <= N_fict; j++) PE" %g", x[j]);
 PE"\nV(x'new) = %g, V(x'old) = %g\n", (*V)(Xnew), (*V)(xold));
 PE"deltaX = %g, deltaU = %g, x'new - x'old = %g\n",
		deltaX, deltaU, x[0] - xold);
#endif

#ifdef QUANTUM
 x[N_fict] = x[0];
 for(i = 1; i < N_fict; i++)
   {
#ifdef DEBG1
     xold = x[i];
#endif
     deltaX = MaxShftAt * (2. * (*RnGn1)() - 1.);
     Xnew = x[i] + deltaX;
     deltaU = C3 * deltaX * ( Xnew + x[i] - x[i + 1] - x[i - 1] )
          + ((*V)(Xnew) - (*V)(x[i])) / (double)N_fict;
     if( deltaU < 0 || ((*RnGn1)() < exp(-betaAt * deltaU)) )
                            { x[i] = Xnew; Accepted++; }
#ifdef DEBG1
     PE"x:"); for(j = 0; j < i; j++) PE" %g", x[j]);
     PE" %g",xold); for(j = i + 1; j <= N_fict; j++) PE" %g", x[j]);
     PE"\nV(x'new) = %g, V(x'old) = %g\n", (*V)(Xnew), (*V)(xold));
     PE"deltaX = %g, deltaU = %g, x'new - x'old = %g\n",
		deltaX, deltaU, x[i] - xold);
#endif
   }
#endif
}  /* --------------->  end of:  MCcycleP1s()  <------------------ */


void MCcycleP1()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(N_fict, N_fict),
    which places them into randNum1[] and randNum2[].
    Periodic selection of the next particle to be moved.		*/

#ifdef QUANTUM
 register int i;
#endif
 register double Xnew;
 register double deltaX;
 register double deltaU;
#ifdef DEBG1
 double xold = x[0];
 int j;
#endif

 (*refresh_rand)(N_fict, N_fict, randNum1, randNum2);

 deltaX = MaxShftAt * (2. * randNum1[0] - 1.);
 Xnew = x[0] + deltaX;
#ifdef QUANTUM
 deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
	  + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
#else
 deltaU = (*V)(Xnew) - (*V)(x[0]);
#endif
 if( deltaU < 0 || (randNum2[0] < exp(-betaAt * deltaU)) )
			{ x[0] = Xnew; Accepted++; }
#ifdef DEBG1
 PE"x: %g", xold); for(j = 1; j <= N_fict; j++) PE" %g", x[j]);
 PE"\nV(x'new) = %g, V(x'old) = %g\n", (*V)(Xnew), (*V)(xold));
 PE"random numbers: %g, %g\n", randNum1[0], randNum2[0]);
 PE"deltaX = %g, deltaU = %g, x'new - x'old = %g\n",
		deltaX, deltaU, x[0] - xold);
#endif

#ifdef QUANTUM
 x[N_fict] = x[0];
 for(i = 1; i < N_fict; i++)
   {
#ifdef DEBG1
     xold = x[i];
#endif
     deltaX = MaxShftAt * (2. * randNum1[i] - 1.);
     Xnew = x[i] + deltaX;
     deltaU = C3 * deltaX * ( Xnew + x[i] - x[i + 1] - x[i - 1] )
          + ((*V)(Xnew) - (*V)(x[i])) / (double)N_fict;
     if( deltaU < 0 || (randNum2[i] < exp(-betaAt * deltaU)) )
                            { x[i] = Xnew; Accepted++; }
#ifdef DEBG1
     PE"x:"); for(j = 0; j < i; j++) PE" %g", x[j]);
     PE" %g",xold); for(j = i + 1; j <= N_fict; j++) PE" %g", x[j]);
     PE"\nV(x'new) = %g, V(x'old) = %g\n", (*V)(Xnew), (*V)(xold));
     PE"rand: %g, %g\n", randNum1[i], randNum2[i]);
     PE"deltaX = %g, deltaU = %g, x'new - x'old = %g\n",
		deltaX, deltaU, x[i] - xold);
#endif
   }
#endif
}  /* --------------->  end of:  MCcycleP1()  <------------------ */


#ifdef QUANTUM

#define RANc (randNum1[j] += randNum2[j],\
	 randNum1[j] > 1 ? randNum1[j++] - 1. : randNum1[j++])
 
void MCcyclePC()
/*============*/
{
 /* This is actually a single stream situation using a combined generator
    obtained by the combination of r1 and r2 according to
    P. L'Ecuyer, Comm.ACM 31, 742 (1988).

    Two independent streams generated by (*refresh_rand)(2*N_fict, 2*N_fict),
    which places them into randNum1[] and randNum2[].
    Then use  (randNum1[i]+randNum2[i]) mod 1  as a single stream number.

    Periodic selection of the next particle to be moved.		*/

 register int i;
 register int j = 0;
 register double Xnew;
 register double deltaX;
 register double deltaU;

 (*refresh_rand)(t_N_fict, t_N_fict, randNum1, randNum2);

 deltaX = MaxShftAt * (2. * RANc - 1.);
 Xnew = x[0] + deltaX;
 deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
	  + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
 if( deltaU < 0 || (RANc < exp(-betaAt * deltaU)) )
			{ x[0] = Xnew; Accepted++; }

 x[N_fict] = x[0];
 for(i = 1; i < N_fict; i++)
   {
     deltaX = MaxShftAt * (2. * RANc - 1.);
     Xnew = x[i] + deltaX;
     deltaU = C3 * deltaX * ( Xnew + x[i] - x[i + 1] - x[i - 1] )
          + ((*V)(Xnew) - (*V)(x[i])) / (double)N_fict;
     if( deltaU < 0 || (RANc < exp(-betaAt * deltaU)) )
                            { x[i] = Xnew; Accepted++; }
   }
}  /* --------------->  end of:  MCcyclePC()  <------------------ */


void MCcycleR1s()
/*=============*/
{
 /* Single stream using 1st RNG, (*RnGn1)().		*/
 /* Random selection of the next particle to be moved.	*/

 register int i;
 register int ind;
 register double Xnew;
 register double deltaX;
 register double deltaU;

 for(i = 0; i < N_fict; i++)
   { ind = (*RnGn1)() * N_fict;
     deltaX = MaxShftAt * (2. * (*RnGn1)() - 1.);
     Xnew = x[ind] + deltaX;
     if(ind)
	{ deltaU = C3 * deltaX * ( Xnew + x[ind] - x[ind + 1] - x[ind - 1] )
		   + ((*V)(Xnew) - (*V)(x[ind])) / (double)N_fict;
          if( deltaU < 0 || ((*RnGn1)() < exp(-betaAt * deltaU)) )
					{ x[ind] = Xnew; Accepted++; }
	}
     else
	{ deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
		   + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
          if( deltaU < 0 || ((*RnGn1)() < exp(-betaAt * deltaU)) )
					{ x[0] = Xnew; Accepted++; }
	  x[N_fict] = x[0];
	}
   }
}  /* --------------->  end of:  MCcycleR1s()  <------------------ */


void MCcycleP2()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(N_fict, N_fict),
    which places them into randNum1[] and randNum2[].
    Periodic selection of the next particle to be moved.
    Random numbers are taken alternately from randNum1[] and
    randNum2[] only when actually needed  either for particle
    moves or acceptance decisions.					*/

 register int i;
 register int j = 0;
 double *p[2];
 register double Xnew;
 register double deltaX;
 register double deltaU;

 p[0] = randNum1;
 p[1] = randNum2;

#define RANN (*(p[(j++) % 2]++))

/***************************************************************************
The following commented out version of declarations gave identical results,
but it was somewhat slower (e.g., for HO, r1ranmar, r2ranmar, 100000 cycles,
beta=3, it needed 14.35 CPU sec, while the above version only 14.23 CPU sec).

 register int i;
 double *p1 = randNum1;
 double *p2 = randNum2;
 double **p = &p1;
 register double ran;
 register double Xnew;
 register double deltaX;
 register double deltaU;

#define RANN (ran = **p, (*p)++, p = (p == &p1) ? &p2 : &p1, ran)
****************************************************************************/

 (*refresh_rand)(N_fict, N_fict, randNum1, randNum2);

 deltaX = MaxShftAt * (2. * RANN - 1.);
 Xnew = x[0] + deltaX;
 deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
	  + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
 if( deltaU < 0 || (RANN < exp(-betaAt * deltaU)) )
			{ x[0] = Xnew; Accepted++; }

 x[N_fict] = x[0];
 for(i = 1; i < N_fict; i++)
   {
     deltaX = MaxShftAt * (2. * RANN - 1.);
     Xnew = x[i] + deltaX;
     deltaU = C3 * deltaX * ( Xnew + x[i] - x[i + 1] - x[i - 1] )
          + ((*V)(Xnew) - (*V)(x[i])) / (double)N_fict;
     if( deltaU < 0 || (RANN < exp(-betaAt * deltaU)) )
                            { x[i] = Xnew; Accepted++; }
   }

}  /* --------------->  end of:  MCcycleP2()  <------------------ */


void MCcycleR1()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(N_fict, 2*N_fict),
    which places them into randNum1[] (r1) and randNum2[] (r2).
    Random selection of the next particle to be moved: r1 for the
    selection of particles, randNum2[2*i] for generating moves and
    randNum2[2*i + 1] for decision making.				   */

 register int i;
 register int ind;
 register int j;
 register double Xnew;
 register double deltaX;
 register double deltaU;

 (*refresh_rand)(N_fict, t_N_fict, randNum1, randNum2);

 for(i = 0, j = 0; i < N_fict; i++, j +=2)
   { ind = randNum1[i] * N_fict;
#ifdef DEBG1
     bins[ind]++;
#endif
     deltaX = MaxShftAt * (2. * randNum2[j] - 1.);
     Xnew = x[ind] + deltaX;
     if(ind)
	{ deltaU = C3 * deltaX * ( Xnew + x[ind] - x[ind + 1] - x[ind - 1] )
		   + ((*V)(Xnew) - (*V)(x[ind])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[j + 1] < exp(-betaAt * deltaU)) )
					{ x[ind] = Xnew; Accepted++; }
	}
     else
	{ deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
		   + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[j + 1] < exp(-betaAt * deltaU)) )
					{ x[0] = Xnew; Accepted++; }
	  x[N_fict] = x[0];
	}
   }


#ifdef DEBG1
 for(i = 0; i < N_fict; i++) PE"%d ", bins[i]); PE"\n");
#endif

}  /* --------------->  end of:  MCcycleR1()  <------------------ */


void MCcycleR2()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(N_fict, 2*N_fict),
    which places them into randNum1[] (r1) and randNum2[] (r2).
    Random selection of the next particle to be moved: r1 for the
    selection of particles, r2 for generating moves and decision making.
    r2 numbers are taken from randNum2[] only when needed - any unused
    numbers are thus at the end of randNum2[].				   */

 register int i;
 register int ind;
 register int j = 0;
 register double Xnew;
 register double deltaX;
 register double deltaU;

 (*refresh_rand)(N_fict, t_N_fict, randNum1, randNum2);

 for(i = 0; i < N_fict; i++)
   { ind = randNum1[i] * N_fict;
#ifdef DEBG1
     bins[ind]++;
#endif
     deltaX = MaxShftAt * (2. * randNum2[j++] - 1.);
     Xnew = x[ind] + deltaX;
     if(ind)
	{ deltaU = C3 * deltaX * ( Xnew + x[ind] - x[ind + 1] - x[ind - 1] )
		   + ((*V)(Xnew) - (*V)(x[ind])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[j++] < exp(-betaAt * deltaU)) )
					{ x[ind] = Xnew; Accepted++; }
	}
     else
	{ deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
		   + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[j++] < exp(-betaAt * deltaU)) )
					{ x[0] = Xnew; Accepted++; }
	  x[N_fict] = x[0];
	}
   }


#ifdef DEBG1
 for(i = 0; i < N_fict; i++) PE"%d ", bins[i]); PE"\n");
#endif

}  /* --------------->  end of:  MCcycleR2()  <------------------ */


void MCcycleR3()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(N_fict, 2*N_fict),
    which places them into randNum1[] (r1) and randNum2[] (r2).
    Random selection of the next particle to be moved: r1 for generating
    moves, r2 for the selection of particles and decision making.	   */

 register int i;
 register int ind;
 register double Xnew;
 register double deltaX;
 register double deltaU;

 (*refresh_rand)(N_fict, t_N_fict, randNum1, randNum2);

 for(i = 0; i < N_fict; i++)
   { ind = randNum2[2 * i] * N_fict;
#ifdef DEBG1
     bins[ind]++;
#endif
     deltaX = MaxShftAt * (2. * randNum1[i] - 1.);
     Xnew = x[ind] + deltaX;
     if(ind)
	{ deltaU = C3 * deltaX * ( Xnew + x[ind] - x[ind + 1] - x[ind - 1] )
		   + ((*V)(Xnew) - (*V)(x[ind])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[2 * i + 1] < exp(-betaAt * deltaU)) )
					{ x[ind] = Xnew; Accepted++; }
	}
     else
	{ deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
		   + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
          if( deltaU < 0 || (randNum2[2 * i + 1] < exp(-betaAt * deltaU)) )
					{ x[0] = Xnew; Accepted++; }
	  x[N_fict] = x[0];
	}
   }

#ifdef DEBG1
 for(i = 0; i < N_fict; i++) PE"%d ", bins[i]); PE"\n");
#endif

}  /* --------------->  end of:  MCcycleR3()  <------------------ */



void MCcycleR4()
/*============*/
{
 /* Two independent streams generated by (*refresh_rand)(1.5*N_fict,
    1.5*N_fict), which places them into randNum1[] and randNum2[].
    Random selection of the next particle to be moved.
    Random numbers are taken alternately from randNum1[] and
    randNum2[] only when actually needed  either for particle
    selection, particle moves or acceptance decisions.
 */
 register int i;
 register int ind;
 register int j = 0;
 double *p[2];
 register double Xnew;
 register double deltaX;
 register double deltaU;

 p[0] = randNum1;
 p[1] = randNum2;

#define RANN1 (*(p[(j++) % 2]++))

 (*refresh_rand)(oP_N_fict, oM_N_fict, randNum1, randNum2);

 for(i = 0; i < N_fict; i++)
   { ind = RANN1 * N_fict;
     deltaX = MaxShftAt * (2. * RANN1 - 1.);
     Xnew = x[ind] + deltaX;
     if(ind)
	{ deltaU = C3 * deltaX * ( Xnew + x[ind] - x[ind + 1] - x[ind - 1] )
		   + ((*V)(Xnew) - (*V)(x[ind])) / (double)N_fict;
          if( deltaU < 0 || (RANN1 < exp(-betaAt * deltaU)) )
					{ x[ind] = Xnew; Accepted++; }
	}
     else
	{ deltaU = C3 * deltaX * ( Xnew + x[0] - x[1] - x[N_fict_1] )
		   + ((*V)(Xnew) - (*V)(x[0])) / (double)N_fict;
          if( deltaU < 0 || (RANN1 < exp(-betaAt * deltaU)) )
					{ x[0] = Xnew; Accepted++; }
	  x[N_fict] = x[0];
	}
   }

}  /* --------------->  end of:  MCcycleR4()  <------------------ */


double Pot(shift) double shift;
/*===========================*/
/* Calculates the total external potential energy of the whole polymer
 * shifted by 'shift'
 */
{
 int i;
 double  sum = 0.;

 if(shift) for(i = 0; i < N_fict; i++ ) sum += (*V)(x[i] + shift);
    else   for(i = 0; i < N_fict; i++ ) sum += (*V)(x[i]);

 return(sum / (double)N_fict);
}


double MovePolymer()
/*================*/
/* Attemps to move the whole polymer using  MaxShtT.
 * Returns the external potential energy of the whole polymer after
 * the move.
 */
{
 int i;
 double deltaX, Vold, Vnew;

 Vold = Pot(0.);

 deltaX = MaxShftTAt * (2. * (*RnGn2)() - 1.);
 Vnew = Pot(deltaX);
 if( Vnew < Vold || ((*RnGn2)() < exp(-betaAt * (Vnew - Vold))) )
		{ for(i = 0; i < N_fict; i++) x[i] += deltaX;
		  x[N_fict] = x[0];
		  AcceptedT++;
		  return(Vnew); 
		}

 return(Vold);
}

#endif   /*  QUANTUM  */


void Mix()
/*======*/
{
int i;
double moves, Norm;
void updateDens();
char flg[2];

if(actionFlag == MIX)
 { lambdaErr2 = 0.; lambdaS = 0.;
   if(qFlag) { qInvErr2 = 0.; qInvS = 0.; }
#ifdef QUANTUM
   alfaErr2 = 0.; gamaErr2 = 0.; eps_virErr2 = 0.;
   alfaS = 0.; gamaS = 0.; eps_virS = 0.;
   for(i = 0; i < nQ; i++) QbetaTS[i] = 0.;
#endif
 }

if( strcmp(InputFileName,OutputFileName) )
	 { cycle = 0L;
	   Accepted = 0L;
#ifdef QUANTUM
	   AcceptedT = 0L;
#endif
	   for(i = 0; i < N_x; i++) dens[i] = 0L;
	   blocks = 0L;
	 }

/*  plotDens(cycle,x);  */
if(PlotWidth) plotPoints(totalCycles,x);

Continue:

#ifdef DEBUG
PE"\nMix: cycle = %ld, totalCycles = %ld, MaxCycles = %ld, PlotInt = %ld\n",
				 cycle, totalCycles, MaxCycles, PlotInt);
#endif

(void)times(&cpuT);		/* Timing of the MC proper started */
usLast = (double)cpuT.tms_utime;
sysLast = (double)cpuT.tms_stime;

while(cycle < MaxCycles)
  { cycle++;
    totalCycles++;
    (*MCcycle)();
#ifdef QUANTUM
    if(MaxShftT > 0. && !(cycle % intT) ) (void)MovePolymer();
#endif
    updateDens();
    if( PlotInt && PlotWidth && !(cycle % PlotInt) )
					plotPoints(totalCycles,x);
					/* originally:  plotDens(cycle,x); */
  }

(void)times(&cpuT);		/* Timing of the MC proper ended */

if(PlotInt && PlotWidth) xRange();
moves = N_fict * (double)cycle;
Norm = moves * Cx * xStepAt;

#ifdef DEBUG
{ double sum = 0.;
  for(i = 0; i < N_x; i++)
	 { densA[i] = (double)dens[i] / Norm; sum += dens[i]; }
  PR"\n\nTotal # of moves = %g, sum of the contents of all density bins\
 = %g", moves, sum);
}
#else
for(i = 0; i < N_x; i++) densA[i] = (double)dens[i] / Norm;
#endif

if(PlotWidth) {
#ifdef QUANTUM
  PR"\n\n\t\t   Density of fictitious particles\n");
#else
  PR"\n\n\t\t\tParticle density\n");
#endif
  Histogram(densA, N_x, xMinAt, xStepAt);
}
OutDensity();

PR"\nAcceptance ratio = %g\n", (double)Accepted / moves);
#ifdef QUANTUM
if(MaxShftT > 0.) PR"Whole-polymer moves: Acceptance ratio = %g\n",
			 (double)AcceptedT * intT / (double)cycle);
#endif

PR"\nCPU time in seconds used for Monte Carlo proper:\n     user %g,\
 system %g, total %g\n",
	(float)(cpuT.tms_utime - usLast)/ 60.,
	(float)(cpuT.tms_stime - sysLast)/ 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime - usLast - sysLast) / 60.);

Prompt:
PE"\n\7Continue mixing, write to Output and quit,\
 or quit (no write)? (c/w/q) ... ");
SC"%s", flg);
switch(flg[0])
 { case 'c': PE"\7New mixing limit (MaxCycles)? ... ");
	     SC"%ld", &MaxCycles);
	     PE"\n");
	     ConfigHead();
	     goto Continue;
   case 'w': OutData();
	     return;
   case 'q': PE"\n");
	     return;
 }
goto Prompt;

}
/*  ------------ end of Mix() --------------  */


void Sample()
/*=========*/
{
char *p = "Sample";
int i, samplings = 0, start = 1;
#ifdef DEBUG
double moves;
#endif
double la, qI;			/* la, al, ga, ep_v, qI, Qb[i] are the	  */
#ifdef QUANTUM			/* contributions from a single sampling;  */
double sum, al, ga, ep_v,	/* lambda, alfa, ... are block averages;  */
				/* lambda2, ... block averages of squares */ 
       *betaT1, *betaT2;  /* coeff's in the exponent - atomic units */

if(nQ > 0) {
 betaT1 = (double *)malloc((unsigned)(nQ * sizeof(double)));
 betaT2 = (double *)malloc((unsigned)(nQ * sizeof(double)));

 for(i = 0; i < nQ; i++) {
		betaT2[i] = (beta - betaTilde[i]) * CE;
		betaT1[i] = - betaT2[i] * beta / betaTilde[i];
 }
}
#endif

lambdaErr2 = 0.;
if(qFlag) qInvErr2 = 0.;
#ifdef QUANTUM
alfaErr2 = 0.; gamaErr2 = 0.; eps_virErr2 = 0.;
#endif
OutData();  /* Opens Output, writes/updates some parameters. */

/*  If this is the very beginning of a new sampling process
    (immediately after mixing, or when the output file is different from
    the input one), start from zeros again:  */
if( fseek(Output, 0L, SEEK_END) ) err_fseek(p);
if( (A_offset + offset_Incr) == ftell(Output) )
	{ cycle = 0L;
          Accepted = 0L;
#ifdef QUANTUM
	  AcceptedT = 0L;
#endif
	  for(i = 0; i < N_x; i++) dens[i] = 0L;
	  blocks = 0L;
	  lambdaS = 0.;
	  if(qFlag) qInvS = 0.;
#ifdef QUANTUM
	  alfaS = 0.; gamaS = 0.; eps_virS = 0.;
	  for(i = 0; i < nQ; i++) QbetaTS[i] = 0.;
#endif
	}
/*  else { It is ensured in checkCL() that SamplingInt
	     and SamplingBlk are the same for the whole file. }  */

#ifdef DEBUG
PE"\nSample: cycle = %ld, totalCycles = %ld, MaxCycles = %ld, PlotInt = %ld\n",
				 cycle, totalCycles, MaxCycles, PlotInt);
#endif

(void)times(&cpuT);		/* Timing of the MC proper started */
usLast = (double)cpuT.tms_utime;
sysLast = (double)cpuT.tms_stime;

while(cycle < MaxCycles)
  { cycle++;
    totalCycles++;
    (*MCcycle)();
#ifdef QUANTUM
    if( MaxShftT > 0. && !(cycle % intT) ) la = MovePolymer();
       /* Also returns the external potential energy
			     of the whole polymer */
    else la = Pot(0.);
#endif
    updateDens();
    if( PlotInt && PlotWidth && !(cycle % PlotInt) )
		 { plotPoints(cycle,x);
#ifdef DEBUG
		   moves = N_fict * (double)cycle;
		   PR"Acceptance ratio = %g\n", (double)Accepted / moves);
#ifdef QUANTUM
		   if(MaxShftT > 0.)
			PR"Whole-polymer moves: Acceptance ratio = %g\n",
			       (double)AcceptedT * intT / (double)cycle);
#endif
#endif
		 }
    if( !(cycle % SamplingInt) )
	{
#ifdef QUANTUM
	  {
	   /*  harmonic potential between the N_fict particles:  */
	   /*  The surplus (N_fict + 1)th particle is always identical
						    with the 0th one */
	   int i1;
	   double diff;
	   sum = 0.;
	   for(i = 0; i < N_fict; i++ )
		 { i1 = i + 1; diff = x[i1] - x[i]; sum += SQR(diff); }
	   al = C2 * sum;
	  }

	  ga = al - la;

	  sum = 0.;
	  for(i = 0; i < N_fict; i++ ) sum += (*epsVir)(x[i]);
	  ep_v = sum / N_fict;

	  lambda += la;
	  alfa += al;
	  gama += ga;
	  eps_vir += ep_v;
	  alfa2 += SQR(al);
	  gama2 += SQR(ga);
	  eps_vir2 += SQR(ep_v);
	  for(i = 0; i < nQ; i++)
		QbetaT[i] += exp(betaT1[i] * al + betaT2[i] * la);
#else
	  la = (*V)(x[0]);
	  lambda += la;
#endif
	  lambda2 += SQR(la);
	  if(qFlag) {
#ifdef QUANTUM
			qI = al + la;
#else
			qI = la;
#endif
			qI = exp(betaAt * qI);
			qInv += qI;
			qInv2 += SQR(qI);
		    }
	  samplings++;
#ifdef DEBUG
    	  printAll(1, "Sample() - end of a sampling");
#endif

	  if(samplings == SamplingBlk)
		{
		 lambda /= SamplingBlk;
		 lambda2 /= SamplingBlk;
		 lambdaS += lambda;
#ifdef QUANTUM
		 alfa /= SamplingBlk;
		 gama /= SamplingBlk;
		 eps_vir /= SamplingBlk;
		 alfa2 /= SamplingBlk;
		 gama2 /= SamplingBlk;
		 eps_vir2 /= SamplingBlk;
		 alfaS += alfa;
		 gamaS += gama;
		 eps_virS += eps_vir;
		 for(i = 0; i < nQ; i++) {
			QbetaT[i] /= SamplingBlk;
			QbetaTS[i] += QbetaT[i];
		 }
#endif
		 if(qFlag) {
#ifdef QUANTUM
			     qInv /= (SamplingBlk * NormFactor);
			     qInv2 /= (SamplingBlk * NormFactor * NormFactor);
#else
			     qInv /= SamplingBlk;
			     qInv2 /= SamplingBlk;
#endif
			     qInvS += qInv;
			   }
		 blocks++;

		  /*  Update Output:  write out the configuration and the
		      block averages at the end of each block:            */
		 if(start) { start = 0;
			     SEEKBEG(Output, A_offset + blocks * offset_Incr)
			     /* This is to ensure the correct continuation 
				of an interrupted previous run		   */
			   }
			   else if( fseek(Output, 0L, SEEK_END) ) err_fseek(p);
		 WL(&cycle, 1)
		 write_A();
		 SEEKBEG(Output, C_offset)
		 write_C();
		 SEEKBEG(Output, H_offset)
		 write_H_short();
		 write_Lm(p);

#ifdef DEBUG
		 printAll(0, "Sample() - end of a sampling block");
#endif
#ifdef QUANTUM
		 alfa = gama = eps_vir = 0.;
		 alfa2 = gama2 = eps_vir2 = 0.;
		 for(i = 0; i < nQ; i++) QbetaT[i] = 0.;
#endif
		 lambda = lambda2 = 0.;
		 if(qFlag) { qInv = qInv2 = 0.; }
		 samplings = 0;
		}
	}  /* ----> end of:  if( !(cycle % SamplingInt) ) <---- */
  }  /* ----------> end of: while(cycle < MaxCycles) <------------- */

(void)times(&cpuT);		/* Timing of the MC proper ended */

if(PlotInt && PlotWidth) xRange();

}  /* --------------------->  end of:  Sample()  <---------------------- */


void Final()
/*========*/
{
int sample = (actionFlag == SAMPLE || actionFlag == REPAIR);
char *p = "Final";
long i, i1, corrupt_blocks = 0L;
double  moves, Norm,
	lambdaA, epsA, qInvA,   /* averages */
	lambdaErr, epsErr, qInvErr;  /* rms errors (standard deviations) */
#ifdef QUANTUM
double  alfaA, gamaA, eps_virA,   /* averages */
	eps_virErr, alfaErr, gamaErr,  /* rms errors */
	Qmin, aux, N2 = 0.5 * N_fict;
int imin;
#endif
double  CEE = CE * CE;

#ifdef DEBUG
printAll(0, "Final() - beginning");
#endif

/* If (lambdaErr2 != 0.), then the rms errors have already been calculated
   in a previous run, and recorded in the input file.			   */

if(sample) Input = Output; else Output = Input;
moves = (double)N_fict * (double)cycle;
Norm = moves * Cx * xStepAt;

#ifdef DEBUG
{ double sum = 0.;
  for(i = 0; i < N_x; i++)
	 { densA[i] = (double)dens[i] / Norm; sum += dens[i]; }
  PR"\n\nTotal # of moves = %g, sum of the contents of all density bins\
 = %g", moves, sum);
}
#else
for(i = 0; i < N_x; i++) densA[i] = (double)dens[i] / Norm;
#endif

if(PlotWidth) {
#ifdef QUANTUM
  PR"\n\n\t\t   Density of fictitious particles\n");
#else
  PR"\n\n\t\t\tParticle density\n");
#endif
  Histogram(densA, N_x, xMinAt, xStepAt);
}
OutDensity();

PR"\n\n          Acceptance ratio = %g\n", (double)Accepted / moves);
#ifdef QUANTUM
if(MaxShftT > 0.) PR"          Whole-polymer moves: Acceptance ratio = %g\n",
			 (double)AcceptedT * intT / (double)cycle);
#endif
PR"\n");
/*.....................................*/
  PR"\n          Total averages:\n");
  lambdaA = lambdaS / blocks;
  if(qFlag) qInvA = qInvS / blocks;
#ifdef QUANTUM
  alfaA = alfaS / blocks;
  gamaA = gamaS / blocks;
  eps_virA = eps_virS / blocks;
  epsA = C4 - gamaA;
#else
  epsA = C4 + lambdaA;
#endif
  PR"          <eps> = %g\n", CE * epsA);
#ifdef QUANTUM
  PR"          <eps_vir> = %g\n", CE * eps_virA);
  PR"          <alfa> = %g\n", CE * alfaA);
  PR"          <gama> = %g\n", CE * gamaA);
#endif
  PR"          <lambda> = %g\n\n", CE * lambdaA);
  if(qFlag) {
    PR"          1/Q = %g         Q = %g\n", qInvA, 1. / qInvA);
    PR"To get the absolute value of Q, the above value must be multiplied\n");
    PR"by the volume (in external units of length) of the P-dimensional\n");
    PR"space accessible to the fictitious particles (length of the Hilbert\n");
    PR"space raised to the P'th power; P=N_fict)\n\n");
  }
#ifdef QUANTUM
  if(! sample) {
    PE"Enter index of beta\~ (1 to %d) with respect to which to calculate\n",
									nQ);
    PE"the ratios of Q (anything else => w.r.t. the minimum of Q) ... ");
    SC"%d", &imin);
    imin--;
    Qmin = QbetaTS[imin] / pow(betaTilde[imin], N2);
  }
  if(sample || imin < 0 || imin >= nQ) {
    imin = 0;
    Qmin = QbetaTS[imin] / pow(betaTilde[imin], N2);
    for(i = 1; i < nQ; i++)
	{ aux = QbetaTS[i] / pow(betaTilde[i], N2);
	  if(aux < Qmin) { Qmin = aux; imin = i; }
	}
  }
  for(i = 0; i < nQ; i++)
    PR"        Q(%g) = Q(beta) * %g = Q(%g) * %g\n", betaTilde[i],
	pow(beta / betaTilde[i], N2) * QbetaTS[i] / blocks,
	betaTilde[imin],
	QbetaTS[i] / (pow(betaTilde[i], N2) * Qmin));
  PR"\n");
#endif
/*.....................................*/
PR"\n          Errors:\n");
if (lambdaErr2 == 0.)
 {
  SEEKBEG(Input, A_offset + offset_Incr)
	/* Skip the initial-configuration block that contains no block
	   averages yet							*/
  RL(&cycle, 1)
  for(i = 1; i <= blocks; i++)
   { read_A(p);
#ifdef DEBUG
     printAll(0, "Final() - end of input of a block");
#endif
     i1 = i * SamplingBlk * SamplingInt;
     if(cycle != i1)
	{ PR"\nFinal(): Block # %ld in Input file corrupted!\n", i);
	  PR"           cycle = %ld,  but block*SamplingBlk*SamplingInt\
 = %ld\n\n", cycle, i1);
	  corrupt_blocks++;
	}
        else
	{ lambdaErr2 += (lambda2 + lambdaA * (lambdaA - 2. * lambda));
	  if(qFlag) qInvErr2 += (qInv2 + qInvA * (qInvA - 2. * qInv));
#ifdef QUANTUM
	  alfaErr2 += (alfa2 + alfaA * (alfaA - 2. * alfa));
	  gamaErr2 += (gama2 + gamaA * (gamaA - 2. * gama));
	  eps_virErr2 += (eps_vir2 + eps_virA * (eps_virA - 2. * eps_vir));
#endif
	}
#ifdef ultrix
     if(!fread((void *)&cycle,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&cycle,sizeof(long),(size_t)1,Input))
#endif
       { if( feof(Input) ) break;
	 else  PR"\nFinal(): No value for cycle read in, but not EOF\n");
       }
   }

  i1 = (blocks > i);
  if(i1)
	{ PR"%ld blocks at the end of Input file missing during error\
 calculation!\n\n", blocks - i);
	  blocks = i;
	}
  if(corrupt_blocks)
	{ PR"\nContribution of %ld corrupted blocks not considered\
 for error calculation!\n\n", corrupt_blocks);
	  blocks -= corrupt_blocks;
	}

  lambdaErr2 /= blocks;
  if(qFlag) qInvErr2 /= blocks;
#ifdef QUANTUM
  eps_virErr2 /= blocks;
  alfaErr2 /= blocks;
  gamaErr2 /= blocks;
#endif

  if( !(i1 || corrupt_blocks) )
   { /* Write out the errors to their reserved place: */
     SEEKBEG(Output, H_offset)
     write_H();
   }

 }

  lambdaErr = sqrt(lambdaErr2);
  if(qFlag) qInvErr = sqrt(qInvErr2);
#ifdef QUANTUM
  eps_virErr = sqrt(eps_virErr2);
  alfaErr = sqrt(alfaErr2);
  gamaErr = sqrt(gamaErr2);
  epsErr = gamaErr;
#else
  epsErr = lambdaErr;
#endif

#ifdef QUANTUM
PR"          rms(eps) = %g\t\t<del eps^2> = %g\n",
					CE * epsErr, CEE * gamaErr2);
PR"          rms(eps_vir) = %g\t<del eps_vir^2> = %g\n",
					CE * eps_virErr, CEE * eps_virErr2);
PR"          rms(alfa) = %g\t\t<del alfa^2> = %g\n",
					CE * alfaErr, CEE * alfaErr2);
PR"          rms(gama) = %g\t\t<del gama^2> = %g\n",
					CE * gamaErr, CEE * gamaErr2);
#else
PR"          rms(eps) = %g\t\t<del eps^2> = %g\n",
					CE * epsErr, CEE * lambdaErr2);
#endif
PR"          rms(lambda) = %g\t<del lambda^2> = %g\n",
					CE * lambdaErr, CEE * lambdaErr2);
if(qFlag) PR"          rms(1/Q) = %g\t<del (1/Q)^2> = %g\n\n",
					qInvErr, qInvErr2);

FC Input);

if( ! sample)
  { PE"\nIs it O.K. to append data ");
#ifdef QUANTUM
    PE"for N = %d ", N_fict);
    PE"to files epsN, eps_virN, ..., d-epsN, ... ? (y/n)  ");
#else
    PE"to files eps, eps_vir, ..., d-eps, ... ? (y/n)  ");
#endif
    SC"%s",p);
    if(*p != 'y') { PE"\nNo files appended\n\n");
		    return;
		  }

#ifdef QUANTUM
    PE"Only epsN, eps_virN and qN, or all files? (e/a)  ");
#else
    PE"Only eps and qq, or all files? (e/a)  ");
#endif
    SC"%s",p);

#ifdef QUANTUM
    if( (Output = fopen("epsN","A")) == NULL)  FIERROR("epsN")
    PO"%d %g\n", N_fict, CE * epsA);
    FC Output);
    if( (Output = fopen("eps_virN","A")) == NULL)  FIERROR("eps_virN")
    PO"%d %g\n", N_fict, CE * eps_virA);
    FC Output);
    if(qFlag) {
		if( (Output = fopen("qN","A")) == NULL)  FIERROR("qN")
		PO"%d %g\n", N_fict, 1. / qInvA);
		FC Output);
	      }
    if(*p == 'e') return;
    if( (Output = fopen("lambdaN","A")) == NULL)  FIERROR("lambdaN")
    PO"%d %g\n", N_fict, CE * lambdaA);
    FC Output);
    if( (Output = fopen("gamaN","A")) == NULL)  FIERROR("gamaN")
    PO"%d %g\n", N_fict, CE * gamaA);
    FC Output);
    if( (Output = fopen("alfaN","A")) == NULL)  FIERROR("alfaN")
    PO"%d %g\n", N_fict, CE * alfaA);
    FC Output);
    if( (Output = fopen("d-epsN","A")) == NULL)  FIERROR("d-epsN")
    PO"%d %g\n", N_fict, CEE * gamaErr2);
    FC Output);
    if( (Output = fopen("d-eps_virN","A")) == NULL)  FIERROR("d-eps_virN")
    PO"%d %g\n", N_fict, CEE * eps_virErr2);
    FC Output);
    if( (Output = fopen("d-lambdaN","A")) == NULL)  FIERROR("d-lambdaN")
    PO"%d %g\n", N_fict, CEE * lambdaErr2);
    FC Output);
    if( (Output = fopen("d-alfaN","A")) == NULL)  FIERROR("d-alfaN")
    PO"%d %g\n", N_fict, CEE * alfaErr2);
    FC Output);
#else
    if( (Output = fopen("eps","A")) == NULL)  FIERROR("eps")
    PO"%ld %g\n", blocks * SamplingBlk, CE * epsA);
    FC Output);
    if(qFlag) {
		if( (Output = fopen("qq","A")) == NULL)  FIERROR("qq")
		PO"%d %g\n", N_fict, 1. / qInvA);
		FC Output);
	      }
    if(*p == 'e') return;
    if( (Output = fopen("lambda","A")) == NULL)  FIERROR("lambda")
    PO"%ld %g\n", blocks * SamplingBlk, CE * lambdaA);
    FC Output);
    if( (Output = fopen("d-eps","A")) == NULL)  FIERROR("d-eps")
    PO"%ld %g\n", blocks * SamplingBlk, CEE * lambdaErr2);
    FC Output);
#endif
  }

#ifdef DEBUG
  printAll(0, "Final() - end");
#endif

}  /* --------------------->  end of:  Final()  <---------------------- */


void mult_Final()
/*=============*/
{
long max_blocks, /* = current max # of blocks from the beginning of
		    InputFile to be taken into account for the calculation
		    of averages and errors.   */
     init_cycle, /* determines the first value of  max_blocks  */
     iBl, i1, corrupt_blocks, blcks, config, SKIP, blockIncrmnt;
int  n, errors_too, others, actual = 0;
char *p = "mult_Final", p1[2], *mw = "w", *mA = "A", *mode = mw;
double  lambdaA, epsA, qInvA;   /* averages */
#ifdef QUANTUM
double  alfaA, gamaA, eps_virA;   /* averages */
#endif
double  CEE = CE * CE;

#ifdef DEBUG
printAll(0, "mult_Final() - beginning");
#endif

PE"\nIs it O.K. to rewrite files eps, eps_vir, ..., d-eps, ... ? (y/n)  ");
SC"%s",p1);
if(p1[0] != 'y') { PE"\nNo files modified\n\n");
		   return;
		 }

PE"Averages only, or averages + errors? (a/e)  ");
SC"%s",p1);
errors_too = (p1[0] != 'a');
#ifdef QUANTUM
PE"Only eps, eps_vir and qq, or all files? (e/a)  ");
#else
PE"Only eps and qq, or all files? (e/a)  ");
#endif
SC"%s",p1);
others = (p1[0] != 'e');
do
  { PE"Enter # of MC cycles to skip at the beginning of the file\n");
    PE"    (will be rounded to Sampling blocks; 0 = nothing skipped):  ");
    SC"%ld", &SKIP);
  } while(SKIP < 0L);
PE"Enter # of MC cycles up to which to average initially\n");
PE"    (will be truncated; made larger than %ld):  ",
				 SKIP);
SC"%ld", &init_cycle);
do
  { PE"Enter # of MC cycles to be repeatedly added to the averages:  ");
    SC"%ld", &blockIncrmnt);
  } while(blockIncrmnt <= 0L);
if(SKIP)
  { PE"Write/record the actual # of samplings averaged, or the total # from\n");
    PE"    the beginning of the file? (a/t)  ");
    SC"%s",p1);
    actual = (p1[0] == 'a');
  }

i1 = SamplingBlk * SamplingInt;
SKIP /= i1;
blockIncrmnt = (blockIncrmnt + 0.5 * i1) / i1;  /* on the command line, in
	the si option, one gives the increment in MC cycles, here it must
	be converted into sampling blocks */
if(blockIncrmnt < 1L) blockIncrmnt = 1L;
iBl = (init_cycle / i1 + 0.5 * blockIncrmnt) / blockIncrmnt;
max_blocks = iBl * blockIncrmnt;
if(max_blocks <= SKIP) max_blocks  = SKIP + blockIncrmnt;
			  /* This is initial value of max_blocks */
lambdaS = 0.;
if(qFlag) qInvS = 0.;
#ifdef QUANTUM
alfaS = gamaS = eps_virS = 0.;
for(n = 0; n < nQ; n++) QbetaTS[n] = 0.;
#endif
corrupt_blocks = 0L;
iBl = SKIP + 1;
SEEKBEG(Input, A_offset + offset_Incr * iBl)
	/* Skip the initial-configuration block that contains no block
	   averages yet, and additional SKIP blocks			*/
RL(&cycle, 1)

while(max_blocks <= blocks)
 {

  PE" >>>>>  ");  /* Continues with:   PE"%s   %ld samplings\n", mode, config);
		     after the current pass through while is finished. */
  lambdaErr2 = 0.;
  if(qFlag) qInvErr2 = 0.;
#ifdef QUANTUM
  alfaErr2 = gamaErr2 = eps_virErr2 = 0.;
#endif
  blcks = max_blocks;

  for( ; iBl <= max_blocks; iBl++) /* For second and all subsequent passes
				      through Master File, calculation of
				      the sums can resume where the calculation
				      of errors ended in the previous pass. */
   { read_A(p);
#ifdef DEBUG
     printAll(0, "mult_Final() - end of input of a block");
#endif
     i1 = iBl * SamplingBlk * SamplingInt;
     if(cycle != i1)
	{ PR"\nmult_Final(): Block # %ld in Input file corrupted!\n", iBl);
	  PR"           cycle = %ld,  but block*SamplingBlk*SamplingInt\
 = %ld\n\n", cycle, i1);
	  corrupt_blocks++;
	}
        else
	{ lambdaS += lambda;
	  if(qFlag) qInvS += qInv;
#ifdef QUANTUM
	  alfaS += alfa;
	  gamaS += gama;
	  eps_virS += eps_vir;
	  for(n = 0; n < nQ; n++) QbetaTS[n] += QbetaT[n];
#endif
	}
#ifdef ultrix
     if(!fread((void *)&cycle,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&cycle,sizeof(long),(size_t)1,Input))
#endif
       { if( feof(Input) ) break;
	 else  PR"\nmult_Final(): No value for cycle read in, but not EOF\n");
       }
   }

  i1 = (max_blocks > iBl);
  if(i1)
	{ PR"%ld blocks at the end of Input file missing during the\
 calculation of averages!\n\n", max_blocks - iBl);
	  blcks = iBl;
	}
  if(corrupt_blocks)
	{ PR"\nContribution of %ld corrupted blocks not considered\
 for the calculation of averages!\n\n", corrupt_blocks);
	  blcks = max_blocks - corrupt_blocks;
	}
  config = (actual ? (blcks - SKIP) : blcks) * SamplingBlk;

  lambdaA = lambdaS / (blcks - SKIP);
  if(qFlag) qInvA = qInvS / (blcks - SKIP);
#ifdef QUANTUM
  alfaA = alfaS / (blcks - SKIP);
  gamaA = gamaS / (blcks - SKIP);
  eps_virA = eps_virS / (blcks - SKIP);
  epsA = C4 - gamaA;
  if( (Output = fopen("eps_vir",mode)) == NULL)  FIERROR("eps_vir")
  PO"%ld %g\n", config, CE * eps_virA);
  FC Output);
  if(others) {
	if( (Output = fopen("gama",mode)) == NULL)  FIERROR("gama")
	PO"%ld %g\n", config, CE * gamaA);
	FC Output);
	if( (Output = fopen("alfa",mode)) == NULL)  FIERROR("alfa")
	PO"%ld %g\n", config, CE * alfaA);
	FC Output);
  }
  if(qFlag) {
	if( (Output = fopen("qq",mode)) == NULL)  FIERROR("qq")
	PO"%ld %g\n", config, 1. / qInvA);
	FC Output);
  }
  if(nQ > 0) { double N2 = 0.5 * (double)N_fict;
	if( (Output = fopen("QQ",mode)) == NULL)  FIERROR("QQ")
	PO"%ld ", config);
	for(n = 0; n < nQ; n++)
	    PO" %g", pow(beta / betaTilde[n], N2) *
					 QbetaTS[n] / (blcks - SKIP));
	PO"\n");
	FC Output);
  }
#else
  epsA = C4 + lambdaA;
#endif
  if( (Output = fopen("eps",mode)) == NULL)  FIERROR("eps")
  PO"%ld %g\n", config, CE * epsA);
  FC Output);
  if(others) {
	if( (Output = fopen("lambda",mode)) == NULL)  FIERROR("lambda")
	PO"%ld %g\n", config, CE * lambdaA);
	FC Output);
  }

  if(errors_too) {

    corrupt_blocks = 0L;
    blcks = max_blocks;
    iBl = SKIP + 1;
    SEEKBEG(Input, A_offset + offset_Incr * iBl)
	/* Skip the initial-configuration block that contains no block
	   averages yet, and additional SKIP blocks			*/
    RL(&cycle, 1)
    for( ; iBl <= max_blocks; iBl++)
     { read_A(p);
#ifdef DEBUG
       printAll(0, "mult_Final() - end of input of a block");
#endif
       i1 = iBl * SamplingBlk * SamplingInt;
       if(cycle != i1)
	{ PR"\nmult_Final(): Block # %ld in Input file corrupted!\n", iBl);
	  PR"           cycle = %ld,  but block*SamplingBlk*SamplingInt\
 = %ld\n\n", cycle, i1);
	  corrupt_blocks++;
	}
        else
	{ lambdaErr2 += (lambda2 + lambdaA * (lambdaA - 2. * lambda));
	  if(qFlag) qInvErr2 += (qInv2 + qInvA * (qInvA - 2. * qInv));
#ifdef QUANTUM
	  alfaErr2 += (alfa2 + alfaA * (alfaA - 2. * alfa));
	  gamaErr2 += (gama2 + gamaA * (gamaA - 2. * gama));
	  eps_virErr2 += (eps_vir2 + eps_virA * (eps_virA - 2. * eps_vir));
#endif
	}
#ifdef ultrix
     if(!fread((void *)&cycle,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&cycle,sizeof(long),(size_t)1,Input))
#endif
	{ if( feof(Input) ) break;
	  else  PR"\nmult_Final(): No value for cycle read in, but not EOF\n");
	}
     }

    i1 = (max_blocks > iBl);
    if(i1)
	{ PR"%ld blocks at the end of Input file missing during error\
 calculation!\n\n", max_blocks - iBl);
	  blcks = iBl;
	}
    if(corrupt_blocks)
	{ PR"\nContribution of %ld corrupted blocks not considered\
 for error calculation!\n\n", corrupt_blocks);
	  blcks = max_blocks - corrupt_blocks;
	}

    lambdaErr2 /= (blcks - SKIP);
    if(qFlag) qInvErr2 /= (blcks - SKIP);
#ifdef QUANTUM
    eps_virErr2 /= (blcks - SKIP);
    alfaErr2 /= (blcks - SKIP);
    gamaErr2 /= (blcks - SKIP);
#endif
    config = (actual ? (blcks - SKIP) : blcks) * SamplingBlk;

#ifdef QUANTUM
    if( (Output = fopen("d-eps",mode)) == NULL)  FIERROR("d-eps")
    PO"%ld %g\n", config, CEE * gamaErr2);
    FC Output);
    if( (Output = fopen("d-eps_vir",mode)) == NULL)  FIERROR("d-eps_vir")
    PO"%ld %g\n", config, CEE * eps_virErr2);
    FC Output);
    if(others) {
	if( (Output = fopen("d-lambda",mode)) == NULL)  FIERROR("d-lambda")
	PO"%ld %g\n", config, CEE * lambdaErr2);
	FC Output);
	if( (Output = fopen("d-alfa",mode)) == NULL)  FIERROR("d-alfa")
	PO"%ld %g\n", config, CEE * alfaErr2);
	FC Output);
    }
#else
    if( (Output = fopen("d-eps",mode)) == NULL)  FIERROR("d-eps")
    PO"%ld %g\n", config, CEE * lambdaErr2);
    FC Output);
#endif
    if(qFlag) {
	if( (Output = fopen("d-qInv",mode)) == NULL)  FIERROR("d-qInv")
	PO"%ld %g\n", config, qInvErr2);
	FC Output);
    }

  }  /* --- end of:  if(errors_too) {   --- */

#ifdef QUANTUM
  PE"%s   %ld samplings\t<eps> = %g,  <eps_vir> = %g\n",
			mode, config, CE * epsA, CE * eps_virA);
#else
  PE"%s   %ld samplings\t<eps> = %g\n",	mode, config, CE * epsA);
#endif
  max_blocks += blockIncrmnt;
  mode = mA;

 }  /* ---  end of:  while(max_blocks <= blocks)  --- */

FC Input);

#ifdef DEBUG
printAll(0, "mult_Final() - end");
#endif

}  /* ------------------->  end of:  mult_Final()  <-------------------- */



void displayFile(Listing)  FILE *Listing;
/*=======================================	similar to InData() */
{
long oEnd, blockEnd;
int i;
char *p = "displayFile";
static char  flg[2] = "n";

 checkCL();

/* if( (Input = fopen(InputFileName,"rb")) == NULL) */
 if( (Input = fopen(InputFileName,"r")) == NULL)
				  FIERROR(InputFileName)
 setbuf(Input, (char *)NULL);

  PR"\n\n\t\t\tTHE CONTENTS OF THE FILE: "); PR InputFileName);

 read_offsets(p);
  PR"\n\nLm_offset = %ld, V_offset = %ld, N_offset = %ld\n",
	 Lm_offset, V_offset, N_offset);
  PR"P1_offset = %ld, P2_offset = %ld, C_offset = %ld\n",
	 P1_offset, P2_offset, C_offset);
  PR"H_offset = %ld, A_offset = %ld, offset_Incr = %ld",
	 H_offset, A_offset, offset_Incr);

 for(i = 0; i < 10; i++) fputs(fgets(text,MAXLETT,Input), Listing);

 read_head(p);
 (void)(*outV)(Listing);
 (void)fgets(text,MAXLETT,Input);

 /* beta  &  N - block: */
#ifdef QUANTUM
 for(i = 0; i < 7; i++) fputs(fgets(text,MAXLETT,Input), Listing);
#else
 for(i = 0; i < 5; i++) fputs(fgets(text,MAXLETT,Input), Listing);
#endif
 read_N(p);
 dynDefs();
   
 PR"\nP1:\n");
 read_P1(p);
  PR"beta = %.16g, betaAt = %.16g\n", beta, betaAt);
  PR"Cx = %.16g, CE = %.16g\n", Cx, CE);
  PR"xMin = %.8g, xMax = %.8g\n", xMin, xMax);
  PR"xMinAt = %.8g, xMaxAt = %.8g\n", xMinAt, xMaxAt);
  PR"xStepAt = %.8g, C4 = %.16g\n", xStepAt, C4);
#ifdef QUANTUM
  PR"C1 = %.16g, C2 = %.16g, C3 = %.16g\n", C1, C2, C3);
  if(nQ > 0) { PR"betaTilde ="); betaTout(betaTilde, 10); }
#endif

 PR"\nP2:\n");
 read_P2(p);
  PR"MaxShft = %.16g, MaxShftAt = %.16g\n", MaxShft, MaxShftAt);
  PR"MaxCycles = %ld, SamplingInt = %ld, SamplingBlk = %ld, PlotInt = %ld\n",
	 MaxCycles, SamplingInt, SamplingBlk, PlotInt);
#ifdef QUANTUM
  PR"MaxShftT = %.16g, MaxShftTAt = %.16g\n", MaxShftT, MaxShftTAt);
  PR"intT = %d\n", intT);
#endif

 (void)fgets(text,MAXLETT,Input);
 /* C - block: */
#ifdef QUANTUM
 for(i = 0; i < 6; i++) fputs(fgets(text,MAXLETT,Input), Listing);
#else
 for(i = 0; i < 5; i++) fputs(fgets(text,MAXLETT,Input), Listing);
#endif

 PR"\nH:");
 read_H(p);
  out_RnGn_state(1, Listing);
  PR"Accepted = %ld\n", Accepted);
#ifdef QUANTUM
  PR"AcceptedT = %ld\n", AcceptedT);
#endif
  PR"x:"); for(i = 0; i < N_fict; i++) PR" %g", x[i]);
  PR"\ndens:"); for(i = 0; i < N_x; i++) PR" %ld", dens[i]); PR"\n");
  PR"lambdaS = %.16g, ", lambdaS);
#ifdef QUANTUM
  PR"alfaS = %.16g\ngamaS = %.16g, ", alfaS, gamaS);
  PR"eps_virS = %.16g\n", eps_virS);
  if(nQ > 0) { PR"QbetaTS ="); betaTout(QbetaTS, 5); }
#endif
  if(qFlag) PR"qInvS = %.16g\n", qInvS);
  PR"lambdaErr2 = %.16g, ", lambdaErr2);
#ifdef QUANTUM
  PR"alfaErr2 = %.16g\ngamaErr2 = %.16g, ", alfaErr2, gamaErr2);
  PR"eps_virErr2 = %.16g\n", eps_virErr2);
#endif
  if(qFlag) PR"qInvErr2 = %.16g\n", qInvErr2);

 if( fseek(Input, 0L, SEEK_END) ) err_fseek(p);
 oEnd = ftell(Input) - A_offset;
 blockEnd = oEnd / offset_Incr;
 PR"\nA:");
 SEEKBEG(Input, A_offset)
 blocks = 0;
 while( flg[0] != 'q' )
   {
#ifdef ultrix
     if(!fread((void *)&cycle,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&cycle,sizeof(long),(size_t)1,Input))
#endif
       { if(feof(Input))
		{ PR"\nEnd of file (at %ld bytes from A:) reached/passed.",
				oEnd);
		  PR"\n                     %ld = %ld * offset_Incr + %ld\n\n",
				oEnd, blockEnd, oEnd % offset_Incr);
		}
		else  PR"\nNo value for cycle read in, but not EOF.\n");
       }
       else
       { PR"\n    BLOCK # %ld\n", blocks);
	 read_A(p);
	 PR"cycle = %ld\n", cycle);
	 PR"lambda = %.16g, lambda2 = %.16g\n", lambda, lambda2);
#ifdef QUANTUM
	 PR"alfa = %.16g, alfa2 = %.16g\n", alfa, alfa2);
	 PR"gama = %.16g, gama2 = %.16g\n", gama, gama2);
	 PR"eps_vir = %.16g, eps_vir2 = %.16g\n", eps_vir, eps_vir2);
	 if(nQ > 0) { PR"QbetaT ="); betaTout(QbetaT, 5); }
#endif
	 if(qFlag) PR"qInv = %.16g, qInv2 = %.16g\n", qInv, qInv2);
       }
     (void)fflush(Listing);
     PE"\n\n\7Continue with the next block, jump to a block, or quit?\
 (n/j/q) ... ");
     SC"%s", flg);
     if(flg[0] == 'j') { PE"\7Give block number ... ");
			 SC"%ld", &blocks);
			 if(blocks < 0L) blocks = 0L;
			 if(blocks > blockEnd) blocks = blockEnd + 1L;
			 SEEKBEG( Input, A_offset + blocks * offset_Incr )
		       }
		       else blocks++; 
   }

FC Input);

}  /*  ------------- end of displayFile() ----------------  */

void repairFile()
/*=============*/
{
int n;
char *p = "repairFile";
long iBl, i1, ccl, cycle0;
double laS = 0., qIS = 0.;
#ifdef QUANTUM
double alS = 0., gaS = 0., ep_vS = 0., *QbS;
#endif
			/* laS, alS, gaS, ep_vS, qIS are the new sums  */
static char  flg[2] = "n";

#ifdef QUANTUM
if(nQ > 0) QbS = (double *)malloc((unsigned)(nQ * sizeof(double)));
#endif

PR"\n\n\t\t\t\tRepairing file %s\n\n", InputFileName);

/* if( (Input = fopen(InputFileName,"rb+")) == NULL) */
if( (Input = fopen(InputFileName,"r+")) == NULL)
				  FIERROR(InputFileName)
setbuf(Input, (char *)NULL);
Output = Input;
 read_offsets(p);
 read_head(p);
 read_N(p);
 dynDefs();
 read_P1(p);
 read_P2(p);
 read_C(p);
 read_H(p);
#ifdef QUANTUM
 x[N_fict] = x[0];
#endif
 SEEKBEG(Input, A_offset)
 RL(&cycle0, 1);

 PE"Enter the new # of blocks and total # of cycles to be written into the\n");
 PE"file being repared: ");
 SC"%ld %ld", &blocks, &cycle);

 totalCycles = cycle + cycle0;
 SEEKBEG(Output, C_offset)
 write_C();

 if(blocks != 0) {
   SEEKBEG( Input, A_offset + blocks * offset_Incr )
   /* This is to ensure the correct continuation of an interrupted job  */
   RL(&ccl, 1)
   if(ccl != cycle)
	{ PR"\nThe two values of \'cycle\' still differ!!!!\n\n");
	  exit(-1);
	}
 }

 SEEKBEG( Input, A_offset + offset_Incr )
 RL(&cycle, 1)
 for(iBl = 1; iBl <= blocks; iBl++)
   { read_A(p);
     i1 = iBl * SamplingBlk * SamplingInt;
     if(cycle != i1)
	{ PR"\nrepairFile(): Block # %ld in Input file corrupted!\n", iBl);
	  PR"           cycle = %ld,  but block*SamplingBlk*SamplingInt\
 = %ld\n\n", cycle, i1);
	  exit(-1);
	}
        else
	{ laS += lambda;
	  if(qFlag) qIS += qInv;
#ifdef QUANTUM
	  alS += alfa;
	  gaS += gama;
	  ep_vS += eps_vir;
	  for(n = 0; n < nQ; n++) QbS[n] += QbetaT[n];
#endif
	}
#ifdef ultrix
     if(!fread((void *)&cycle,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&cycle,sizeof(long),(size_t)1,Input))
#endif
       { if( feof(Input) ) break;
	 else  PR"\nrepairFile(): No value for cycle read in, but not EOF\n");
       }
   }

PR"\n\n          New                old             sums:\n");
PR"lambda:   %g          %g\n", laS, lambdaS);
#ifdef QUANTUM
PR"alfa:     %g          %g\n", alS, alfaS);
PR"gama:     %g          %g\n", gaS, gamaS);
PR"eps_vir:  %g          %g\n", ep_vS, eps_virS);
#endif
if(qFlag) PR"qInv:     %g          %g\n", qIS, qInvS);

#ifdef QUANTUM
PR"\n New QbetaTS:"); betaTout(QbS, 5);
PR"\n Old QbetaTS:"); betaTout(QbetaTS, 5);
#endif

PE"\nWrite new averages into the file? (y/n) ... ");
SC"%s", flg);
if(flg[0] == 'y') { lambdaS = laS;
		    if(qFlag) qInvS = qIS;
#ifdef QUANTUM
		    alfaS = alS;
		    gamaS = gaS;
		    eps_virS = ep_vS;
		    alfaErr2 = 0;
		    gamaErr2 = 0;
		    eps_virErr2 = 0;
		    for(n = 0; n < nQ; n++) QbetaTS[n] = QbS[n];
#endif
		    lambdaErr2 = 0;
		    if(qFlag) qInvErr2 = 0;

		    SEEKBEG(Output, H_offset)
		    write_H();
		  }
write_Lm(p);
PR"\n");

}  /* ----------------->  end of:  repairFile()  <--------------------- */
