/*****************************************************************************/
/********************** (C) 1992 by Miroslav Kolar ***************************/
/*									     */
/*	Romberg integration of the effective 1D estimators obtained by	     */
/*  the approximate analytic integration (up to the 2nd order in potential)  */
/*  in the N_fict-1 dimensional subspace perpendicular to the body diagonal, */
/*	    of the path-integral quantum Monte-Carlo estimators		     */
/*    for a single particle in a one-dimensional external potential V(x).    */
/*   This version is for some even number of fictitious particles only!!!    */
/*									     */
/*  Internally, potential is assumed to be scaled to certain atomic-like     */
/*  units, so that the hamiltonian is  -(1/2)(d^2/dx^2) + V(x)		     */
/*									     */
/*****************************************************************************/

/*
External files required: 1Dpotential.c
			 NRqromb.c
			 LinePlots.c
			 DateAndTime.c

 List of functions in this file:
	void printChoices(N, descr) int N; char *descr;
	void usage(mode) int mode;
	void defaults(mode) int mode;
	void help1()
	void help()
	main(argc, argv)  int argc; char *argv[];
	void checkCL()
	void update_N_fict()
	void dynDefs()
	void setKappa();
	void printParams()
	void OutDensity()
	void writeOut()
	double fq1(u)  double u;
	double fq2(u)  double u;
	double fal1(u)  double u;
	double fal2(u)  double u;
	double fla1(u)  double u;
	double fla2(u)  double u;
	double fev1(u)  double u;
	double fev2(u)  double u;
	float DensIntegrate(X)  double X;
	float rhoHOexactPrim2(X)  double X;
	float rhoHOexact(X)  double X;

External functions:
	char *DateAndTime();			(in  DateAndTime.c)
	void initPlot(), plotFun1(), Histogram();
						(in  LinePlots.c)
        Currently available potentials are defined in 1Dpotential.c:
		double (*V)();
		double (*dV)();      = dV(x)/dx
		double (*epsVir)();  = V(x) + 0.5*(x-Shift)* dV(x)/dx
		void   (*inV)();     input of the parameters of the potential
		long   (*outV)();    output of the parameters of the potential
		double (*V2_4)();  = V''(x) / 4
		double (*V23_2)();  = (V''(x) + 0.25 * (x-Shift) * V'''(x)) / 2
		void choosePotential();

	Romberg integration:
		double qromb(func)  double (*func)();   (in NRqromb.c)
		void qromb_init()			(	"    )
*/

#include <errno.h>
#include <sys/times.h>
#include <stdio.h>                  
#include <math.h>
#include <string.h>

/*#define DEBUG*/

#define PE         (void)fprintf(stderr,
#define PR         (void)fprintf(Listing,
#define SC         (void)scanf(
#define FC         (void)fclose(
#define ILLEGAL    { PE" Illegal option '%s' ignored!\n",argv[i]); illeg++; }
#define FIERROR(x) { PR" File '%s': %s\n", x, sys_errlist[errno]); exit(-1); }

#define MAXLETT    101   /* max length of file names, text lines, etc.
			    Must be at least 10  */

#define CUBE	   0  /* Integration inside the hypercube only */
#define ALL	   1  /* Space perpendicular to diagonal extended to infinity */

#define AVERAGES   0
#define RHO1	   1
#define RHOm	   2

/* Defaults:  */
#define DEFN_fict	6	/* # of fictitious particles  */
#define DEFpotFlag	0	/* see 1Dpotential.c   */
#define DEFPlotWidth	65
#define DEFN_x		10	/* # of points to calculate density at  */
#define DEFEPSromb	1e-7	/* rel accuracy in Romberg integration  */
#define DEFJMAXromb	18	/* max # of steps in Romberg integration  */
#define DEFKromb	5	/* # of points in Romberg extrapolation  */


/* ------ Global declarations: ------- */
#ifdef ultrix
extern int  errno;
#else
int  errno;
#endif
extern char *sys_errlist[];
extern double exp(), erf(), sqrt();
extern clock_t times();
extern void exit();
/*
#ifdef ultrix
        extern void *malloc();
#else
        extern char *malloc();
#endif
*/

extern char *DateAndTime();			/* in DateAndTime.c  */
extern void initPlot(), plotFun1(), Histogram();
						/* in LinePlots.c    */
extern double qromb();
extern void qromb_init();                       /* in NRqromb.c      */

static struct tms cpuT;		/*    - " -        */

static char *author = "                     Written by M. Kolar (C) 1992\n",
	    *space = "\n                                    ",
	    *progName;	/*  to be set equal to  argv[0]  */

static int outsideFlag = ALL,
	   actionFlag = AVERAGES; /* : Q and energy averages calculated */
		   /* = RHO1 : rho(xValue) calculated			*/
		   /* = RHOm : whole array of rho(.) calculated		*/
static char *CUBEstring = "CUBE",
	    *DENSstring = "rho";

static int  PlotWidth = DEFPlotWidth;
		/*  Width (in characters) of the line-printer plots  */

FILE *Listing = stdout;
static char ListingFileName[MAXLETT];
	 /* To print out parameters, final results, configurations during
	    mixing, etc.    If no file is given, stdout is used.   */

int	   N_fict = DEFN_fict,
	   N_fictMax = DEFN_fict; /* used when actionFlag = RHO1 */
static int N_fict_1,  /* = N_fict - 1 */
	   NkapDbl;   /* = (N_fict - 2)/2 = # of doubly degenerate eigenvalues
			  of the polymer "spring-force" matrix A  */
static double N_fictO2;  /* = N_fict * 0.5  */

double  beta = -1.,  /* beta = 1/(k*T);
			k - Boltzman constant, T - temperature  */
	m = -1.,     /* mass of the real particle */
	C1,          /* = N_fict * mAt / (2*pi*hAt*hAt*betaAt)      */
	C2,          /* = N_fict * mAt / (2*hAt*hAt*betaAt*betaAt)  */
	C3,          /* = 2 * C2                                    */
	C4,          /* = N_fict / (2*betaAt)			    */
	C4s,	     /* = C4^2					    */
	s2piBeta;    /* = sqrt(2*pi*betaAt)			    */

float 	      xMin = 1e+30, xMax = -1e+30; /*  integration limits   */
static double xValue = 0., /* to calculate a single density at (DENS on) */
	      NormFactor;  /* to normalize Q, density */

/* Density of (fictitious) particles: */
static int    N_x = DEFN_x;	  /*  # of points to calculate density at  */
static double *xd;	/*  At xd[i] calculate 		    | dyn defined  */
float	      *dens;	/*  density dens[i]		    | in dynDefs() */

static double *kappa, /* doubly degenerate eigenvalues of A |   - " -      */
	      kappa0; /* eigenvalue 4 */
	/* All kappa are actually multiplied by C4 */

/* Integrals: */
static double
   alfa, lambda, eps_vir,
	  /* See Herman, Bruskin, Berne, J.Chem.Phys. 76, 5150 (1982) */
   Q;  /* canonical sum (coordinate dependent partition function) */

/* Atomic-like values used consistently in the internal representation:     */
static double	betaAt;		/* mAt = 1, hAt = 1 */
double		xMinAt, xMaxAt,
		LAt;		/*  = (xMaxAt - xMinAt)		     */
static double	xStepAt,	/*  = (xMaxAt - xMinAt) / (N_x - 1)  */
		xValueAt;       /* to calculate a single density at  */
double	Cx = 0.,  /* coord. transformation: x = Cx * xAt	   */
	CE = 0.;  /* energy transformation: E = CE * E_atomic,
					    beta = betaAt / CE	   */
	/* For harmonic oscillator: Cx = CE = 1			   */

#define P1237 (float)1.273e29
#define M1237 (float)-1.273e29
static float  p1273 = P1237, clxMin = P1237, 
	      m1273 = M1237, clxMax = M1237;

/* Romberg: */
double EPSromb = DEFEPSromb;
int JMAXromb = DEFJMAXromb,
    Kromb = DEFKromb;

static double SQRTPI; /* = sqrt(Pi) */

/*-----------------------------------------------------------------------*/
/*     Currently available potentials are defined in 1Dpotential.c       */

static int potFlag = DEFpotFlag;
	 /* equals to the number of the selected potential */

double (*V)();		/* pointer to the potential function		*/
double (*dV)();		/* pointer to dV(x)/dx				*/
double (*epsVir)();	/* pointer to the function returning
			V(x) + 0.5*(x-Shift)* dV(x)/dx  for the selected V(x),
			where Shift is nonzero for a potential defined in
			finite interval only, and is selected so that
			the virial estimator gives correct averages	  */
void   (*inV)();	/* pointers to functions for input and output   */
long   (*outV)();	/* of the parameters of the potential		*/
float  xRet1, xRet2;  /* (approximate) classical return points of the
                          potential for energy = Vminimum + 2 / beta,
                          or some other suitable limits			*/
extern double (*V2_4)(),   /* pointer to  V''(x) / 4			 */
	      (*V23_2)();  /* pointer to (V''(x)+0.25*(x-Shift)*V'''(x))/2 */
extern char *potBlurb,   /* potential description for printParams()      */
	    *potLabel[], /* potential name abbreviations to be used on
			    command line [and in help() and short help]  */
	    *potHelp[];  /* more information on potentials for help()    */
extern int  numPot,      /* total number of currectly defined potentials */
	    HOpotFlag;   /* the value of potFlag for Harmonic Oscillator */
extern void choosePotential();  /* and perform potential initialization  */

/* Unused here but required in 1Dpotential.c: */
FILE *Input;
void err_fseek(p) char *p; { }
void err_fread(p) char *p; { }

/*	End of definitions for the file 1Dpotential.c			 */
/*-----------------------------------------------------------------------*/

/* Some of the functions defined below the Main:  */
void   checkCL(), update_N_fict(), dynDefs(), setKappa(), OutDensity(),
       writeOut();
float  DensIntegrateOld(), DensIntegrate(), rhoHOexactPrim2(), rhoHOexact();
double fq1(), fq2(), fal1(), fal2(), fla1(), fla2(), fev1(), fev2();


void printChoices(N, descr) int N; char *descr[];
/*=============================================*/
{
int n;

PE"{"); PE *descr);
for(n = 1; n < N; n++) { PE" "); PE descr[n]); }
PE"}");
}

void usage(mode) int mode;
/*======================*/
{
PE"\7\nUsage:"); if(mode) PE"\n");
PE"  %s options [ > file]\n", progName);
if(mode)
 PE"\n  options: pPotential");
 else
 { PE"  options: p");
   printChoices(numPot, potLabel);
 }
PE" lListFile\n");
PE"           fNfict bBeta xXmin XXmax hNhist eEPS jJMAX kRextr pwPlotWidth\n");
PE"           {%s %sXvalue %s}                     ( -h  ?  -help )\n",
		CUBEstring, DENSstring, DENSstring);
}

void defaults(mode) int mode;
/*=========================*/
{
PE"Defaults:\n ");
PE"  p%s  f%d  e%.16g  j%d  k%d  h%d  pw%d\n", potLabel[DEFpotFlag],
	DEFN_fict, DEFEPSromb, DEFJMAXromb, DEFKromb, DEFN_x, DEFPlotWidth);
PE"     Integration extended to infinity in the perpendicular direction\n");
if(mode)  PE"Example: %s  b0.4 rho h12 x0.23 X15.0\n\n", progName);
}


void help1()
/*========*/
{
usage(0);
defaults(0);
PE"########### If you need more help, enter command: %s -help ###########\n\n",
	progName);
}

void help()
/*=======*/
{
	int n;
PE"     Romberg integration of the effective 1D estimators obtained by\n");
PE"  the approximate analytic integration (up to the 2nd order in V(x)) in\n");
PE"  the N_fict-1 dimensional subspace perpendicular to the body diagonal,\n");
PE"         of the path-integral quantum Monte-Carlo estimators\n");
PE"   for a single particle in a one-dimensional external potential V(x).\n");
PE"  This version is for (some) even numbers of fictitious particles only.\n");
PE author);
usage(1);
PE"\n  Only some of these options need be given in any order. All essential\n");
PE"  parameters not supplied will be prompted for.\n\n");
PE"  - As a default, energy averages are calculated and the approximate\n");
PE"    analytic integration is extended to infinity in the direction\n");
PE"    perpendicular to the diagonal of the hypercube. To limit the \n");
PE"    integration to the hypercube, specify the option %s.\n\n", CUBEstring);
PE"  Potential:  "); PE potLabel[0]); PE potHelp[0]);
for(n = 1; n < numPot; n++) {
PE"\n              "); PE potLabel[n]); PE potHelp[n]);
			    }
PE"\n\n  Nfict:      Number of fictitious particles in the polymer\n\n");
PE"  Beta:       1/(kT)\n\n");
PE"  Xmin, Xmax:  integration limits (along one side of the hypercube)\n\n");
PE"  Nhist: # of points to calculate the density at (and nothing else).\n");
PE"         Applies only when  %s  is specified (without any Xvalue)\n\n",
					 DENSstring);
PE"  Xvalue: only a single density calculated at x=Xvalue for increasing Nfict\n\n");
PE"        Density calculation doesn't require any further numerical\n");
PE"        integration. Analytical integration extended to infinity was\n");
PE"        done after higher than 2nd order terms in V(x) were neglected.\n\n");
PE"  EPS: Relative accuracy of the Romberg integration\n\n");
PE"  JMAX: Max # of steps of the Romberg integration\n\n");
PE"  Rextr: # of points used in Romberg extrapolation\n\n");
PE"  PlotWidth: Width (in characters) of all line-printer type plots\n");
PE"             (PlotWidth = 0 means no plotting at all)\n\n");
defaults(1);
exit(0);
}
/* End of help() */


/*=====================================*/
/*    MAIN program starts here!!!!     */
/*=====================================*/
main(argc, argv)  int argc; char *argv[];
/*=====================================*/
{
  int	i, atoi(), n,
	illeg = 1, /* see #define ILLEGAL  */
	clpotFlag;
  long atol();
  int  strcmp(), strncmp();
  char *ARG, *strcpy();
  double atof(), aux;

#ifdef DEBUG
 PE"\nAvailable potentials:  ");
 for(n = 0; n < numPot; n++) { PE potLabel[n]); PE"   "); }
 PE"\n");
#endif

progName = argv[0];
for (i=1; i<argc; i++)
  {
   ARG = argv[i];
#ifdef DEBUG
   PE"\n..... processing '%s'\n",ARG);
#endif
   if ( ! strcmp(ARG,"-help") || ! strcmp(ARG,"-HELP") )  help();
   if ( ! strcmp(ARG,"-h") || ! strcmp(ARG,"-H") ||
	! strcmp(ARG,"?") || ! strcmp(ARG,"-?") )  { help1(); exit(0); }

     else if(! strcmp(ARG,CUBEstring))	    outsideFlag = CUBE;
     else if(! strncmp(ARG,DENSstring,(n = (int)strlen(DENSstring))) )
           { if((int)strlen(ARG) == n) actionFlag = RHOm; else
		{ actionFlag = RHO1; ARG += n; xValue = atof(ARG); }
	   }
     else if(*ARG == 'e')
           { aux = atof(++ARG); if(aux > 0.) EPSromb = aux; else ILLEGAL }
     else if(! strncmp(ARG,"pw",2))
       { ARG += 2; if(isdigit(*ARG)) PlotWidth = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'p')
           { ++ARG; for(n = 0; n < numPot; n++)
			{ if(! strncmp(ARG,potLabel[n],strlen(potLabel[n])) )
						{ clpotFlag = n; break; }
			}
             if( clpotFlag == -1) { ILLEGAL } else potFlag = clpotFlag;
           }
     else if(*ARG == 'l') strcpy(ListingFileName,++ARG);
     else if(*ARG == 'h')
       { ++ARG; if(isdigit(*ARG)) N_x = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'f')
       { ++ARG; if(isdigit(*ARG)) N_fict = (N_fictMax = atoi(ARG));
							else ILLEGAL }
     else if(*ARG == 'j')
       { ++ARG; if(isdigit(*ARG)) JMAXromb = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'k')
       { ++ARG; if(isdigit(*ARG)) Kromb = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'x') clxMin = (float)atof(++ARG);
     else if(*ARG == 'X') clxMax = (float)atof(++ARG);
     else if(*ARG == 'b')
           { aux = atof(++ARG); if(aux > 0.) beta = aux; else ILLEGAL }
     else ILLEGAL
  }

if (illeg == argc) help1(); 

if(actionFlag == RHO1) N_fict = 2;

/* Initialize Listing File:  */
if(ListingFileName[0] != '\0')
 { if( (Listing = fopen(ListingFileName,"A")) == NULL)
				  FIERROR(ListingFileName)
 }

PR space); PR DateAndTime());
PR"\n\n    %s:  ", progName);
if(actionFlag == AVERAGES) {
PR"Romberg integration of the effective 1D estimators obtained by\n");
PR"   the approximate 2nd-order N_fict-1 dimensional analytic integration\n");
PR"            of the path-integral quantum Monte-Carlo estimators\n");
}
else {
PR"Approximate density obtained by analytical integration after\n");
PR"    higher than 2nd-order terms in V(x) were neglected in the primitive\n");
PR"            path-integral quantum Monte-Carlo density matrix\n");
}
PR"                     for a single particle in a V(x).\n");
PR author); PR"\n");

checkCL();
dynDefs();
SQRTPI = sqrt(M_PI);

switch (actionFlag)
 {
   case AVERAGES: if(outsideFlag == CUBE)
			{ PR"\nCUBE  NOT IMPLEMENTED\n\n"); exit(0);
			  /*
			  PE"Q:\n");
			  Q = qromb(fq1);
			  PE"alfa:\n");
			  alfa = qromb(fal1);
			  PE"lambda:\n");
			  lambda = qromb(fla1);
			  PE"eps_vir:\n");
			  eps_vir = qromb(fev1);
			  */
			}
			else
			{ PE"Integrating ... Q ");
			  Q = qromb(fq2);
			  PE"  alpha ");
			  alfa = qromb(fal2);
			  PE"  lambda ");
			  lambda = qromb(fla2);
			  PE"  eps_vir ");
			  eps_vir = qromb(fev2);
			  PE"\n\n");
			}
		  writeOut();
		  break;

   case	RHO1:	  PR"\n P\t  rho(%g)\t\t\t  (P = Nfict)\n", xValue);
		  PR" %d\t %g\n", N_fict, DensIntegrate(xValueAt));
		  while(N_fict < N_fictMax)
			{ update_N_fict();
			  PR" %d\t %g\n", N_fict, DensIntegrate(xValueAt));
			}
		  PR"\n");
		  if(potFlag == HOpotFlag)
		    { if(N_fict == 2)
			PR"\tExact primitive density for N_fict=%d is %g\n\n",
					N_fict, rhoHOexactPrim2(xValueAt));
		      PR"\tExact quantum density is %g\n\n",
						rhoHOexact(xValueAt));
		    }
		  break;

   case	RHOm:	  for(i = 0; i < N_x; i++) dens[i] = DensIntegrate(xd[i]);
		  OutDensity();
 }

(void)times(&cpuT);
PR"Total (us+sys) CPU time used by the whole job: %g sec\n\n",
	(float)(cpuT.tms_utime + cpuT.tms_stime) / 60.);
FC Listing);
exit(0);

}  /* ------------------------ end of Main --------------------------  */


void checkCL()
/*============         Use all command line values that make sense  */
{
void printParams();

if(clxMin == p1273 && clxMax == m1273) PE"\n  CHECK xMin AND xMax !!!\n\n");
if(clxMin != p1273) xMin = clxMin;
if(clxMax != m1273) xMax = clxMax;

while(beta <= 0.0)
	 { PE"\7Enter positive value of beta: "); SC"%lf", &beta); }

choosePotential(potFlag, 0);  /* among others, it calculates
		      Cx and CE, and C1, C2 and C3 */

while(xMax <= xMin)
	{ PR"\nFor integration limits, enter xMin < xMax: ");
	  SC"%f%f", &xMin, &xMax);
	}

/*
if( actionFlag == RHO1 && (xValue < xMin || xValue > xMax) )
	{
	 PR"\n x = %g to calculate the density at is outside of (%g,%g)!\n\n",
			xValue, xMin, xMax);
	 exit(-1);
	}
*/

if(actionFlag == AVERAGES)
  { if(N_fict > 20) { int i, dif, mdif = abs(N_fict - i);

		      for(i = 64; ; i *=2) { dif = abs(N_fict - i);
					     if(dif > mdif) break;
					     mdif = dif;
				           }
		      N_fict = i / 2;
		      NkapDbl = N_fict / 2;
		    }
	else	    N_fict = 2 * (NkapDbl = N_fict / 2); 

    N_fictO2 = NkapDbl--;
  }
  else N_fictO2 = 0.5 * (double)N_fict;

N_fict_1 = N_fict - 1;

xMaxAt = xMax / Cx;
xMinAt = xMin / Cx;
LAt = xMaxAt - xMinAt;
xStepAt = LAt / (N_x - 1);
betaAt = beta * CE;
C4 = N_fictO2 / betaAt;
if(actionFlag == RHO1) xValueAt = xValue / Cx;

if(actionFlag == AVERAGES)
  { /* Calculate normalization factor
       [ includes: sqrt((double)N_fict * pow(M_PI, (double)N_fict_1)) ]:
    */
    NormFactor = pow(C4, N_fictO2) * sqrt((double)N_fict / M_PI);
  }
s2piBeta = sqrt(2. * M_PI * betaAt);
C4s = C4 * C4;

/* Print out all parameters, plot potential: */
printParams();

}  /* ------------------- end of checkCL() -----------------------  */


void update_N_fict()
/*================*/
/* Used when actionFlag = RHO1 */
{
N_fict++;
N_fictO2 = 0.5 * (double)N_fict;
C4 = N_fictO2 / betaAt;
C1 = C4 / M_PI;
s2piBeta = sqrt(2. * M_PI * betaAt);
C4s = C4 * C4;
}


void dynDefs()
/*==========*/
{
 int  i;

 if(actionFlag == AVERAGES) {
	if(NkapDbl > 0) kappa = (double *)malloc(NkapDbl * sizeof(double));
	setKappa();
	qromb_init();
 }

 if(actionFlag == RHOm)
   {  xd = (double *)malloc(N_x * sizeof(double));
      dens = (float *)malloc(N_x * sizeof(float));
      xd[N_x - 1] = xMaxAt;
      for(i = N_x - 2; i >= 0; i--) xd[i] = xd[i + 1] - xStepAt;
   }
}


void setKappa()
/*===========*/
/* Calculate doubly degenerate eigenvalues
   of the polymer "spring-force" matrix A
*/
{
 int  i, j;
 double C9, S9s3, SQRT;

 switch(N_fict)
   {
	case 10:
	case 20: kappa[0] = 0.5 * (3. - (SQRT = sqrt(5.)));
		 kappa[1] = 0.5 * (3. + SQRT);
		 kappa[2] = 0.5 * (5. - SQRT);
		 kappa[3] = 0.5 * (5. + SQRT);
		 if(N_fict == 10) break;
		 kappa[4] = 2.;
		 kappa[5] = sqrt(kappa[2]);
		 kappa[6] = 2. - kappa[5];
		 kappa[5] += 2.;
		 kappa[7] = sqrt(kappa[3]);
		 kappa[8] = 2. - kappa[7];
		 kappa[7] += 2.;
		 break;

	case 14: kappa[0] = .19806226419516174752779536098;
		 kappa[1] = 3.80193773580483825247220463902;
		 kappa[2] = 1.55495813208737119142219487100;
		 kappa[3] = 2.44504186791262880857780512900;
		 kappa[4] = .75302039628253293894999023199;
		 kappa[5] = 3.24697960371746706105000976801;
		 break;

	case 18: kappa[0] = 1.;
		 kappa[1] = 3.;
		 C9 = cos(M_PI / 9.);
		 S9s3 = sin(M_PI / 9.) * sqrt(3.);
		 kappa[2] = 2. * (1. - C9);
		 kappa[3] = 2. * (1. + C9);
		 kappa[4] = 2. - C9 + S9s3;
		 kappa[5] = 2. + C9 + S9s3;
		 kappa[6] = 2. - C9 - S9s3;
		 kappa[7] = 2. + C9 - S9s3;
		 break;

	case 12: kappa[2] = 2.;
		 kappa[3] = 2. - (SQRT = sqrt(3.));
		 kappa[4] = 2. + SQRT;
	case  6: kappa[1] = 3.;
		 kappa[0] = 1.;
		 break;

	case 16:
	case  8: kappa[2] = 2. - M_SQRT2;
		 kappa[1] = 2. + M_SQRT2;
		 if(N_fict == 16) { kappa[3] = 2 - (SQRT = sqrt(kappa[1]));
				    kappa[4] = 2 + SQRT;
				    kappa[5] = 2 - (SQRT = sqrt(kappa[2]));
				    kappa[6] = 2 + SQRT;
				  }
	case  4: kappa[0] = 2.;
	case  2: break;
/*
	case 11: kappa[0] = .31749293433763766227637670216;
		 kappa[1] = 1.16916997399622714894145170154;
		 kappa[2] = 2.28462967654657028088758533723;
		 kappa[3] = 3.30972146789057012811385014493;
		 kappa[4] = 3.91898594722899477978073611413;
		 break;

	case 13: kappa[0] = .22908794869358020819924895597;
		 kappa[1] = .86387050653768839497638488174;
		 kappa[2] = 1.75892663948935389330186462510;
		 kappa[3] = 2.70920977408507125193927578520;
		 kappa[4] = 3.49702149634220219726926119940;
		 kappa[5] = 3.94188363485210405431396455259;
		 break;

	case 17: kappa[0] = .13505554119128839085376821636;
		 kappa[1] = .52198216555868176815093138025;
		 kappa[2] = 1.10852328844692346520708490124;
		 kappa[3] = 1.81546328107339600952069778569;
		 kappa[4] = 2.54732598014416572707815587087;
		 kappa[5] = 3.20526927275851277835717630997;
		 kappa[6] = 3.70043427145922830426828784590;
		 kappa[7] = 3.96594619936780355656389768971;
		 break;

	case 19: kappa[0] = .10836551659873064196066857143;
		 kappa[1] = .42171898120721280156203770120;
		 kappa[2] = .90610368375514625057647450661;
		 kappa[3] = 1.50902902571840170215541816441;
		 kappa[4] = 2.16515869094466464920068786848;
		 kappa[5] = 2.80339084930593891503368331949;
		 kappa[6] = 3.35456314325148214952430196899;
		 kappa[7] = 3.75894750241297814278170950976;
		 kappa[8] = 3.97272260680544474720501838964;
		 break;
*/
	default: kappa[0] = 2.;
		 for(i = 1, j = 0; i < NkapDbl; j++)
			{ kappa[i++] = 2. - (SQRT = sqrt(kappa[j]));
			  kappa[i++] = 2. + SQRT;
			}
   }

#ifdef DEBUG
 PE"kappa:\n");
 for(i = 0; i < NkapDbl; ) { PE"%.13f", kappa[i]);
			     if(++i % 5) PE" "); else PE"\n");
			   }
 { double K = 0;
   for(i = 0; i < NkapDbl; i++) K += (1. / kappa[i]);
   K = (2. * K + 0.25) / (double)(N_fict * N_fict);
   PE"\nK = %g,  K - 1/12 = %g\n\n", K, K - 1. / 12.);
 }
#endif

 for(i = 0; i < NkapDbl; i++) kappa[i] *= C4;
 kappa0 = C4 * 4.;

}  /* ------------------- end of setKappa() ----------------------  */


void printParams()
/*==============*/
/* Print out all important parameters at the beginning of the listing.  */
/* (Called at the end of  checkCL() )					*/
{
 static char *fmt = "%8ld ", *fmtV = "%9g";
 static int  fmt_len = 9;
 int i;

 PR potBlurb); (void)(*outV)(Listing);
 PR"Beta = %.8g\n", beta);
 if(actionFlag == RHO1)
	PR"Maximum no. of fictitious particles = %d\n", N_fictMax);
    else
	{ PR"No. of fictitious particles = %d\n", N_fict);
	  PR"xMin = %g,  xMax = %g\n", xMin, xMax);
	}
 if(actionFlag == AVERAGES)
	   { PR"Analytic integration ");
	     if(outsideFlag == CUBE) PR"limited to the hypercube\n");
		else  PR"extended to infinity\n");
	     PR"Romberg: EPS = %g, JMAX = %d, K = %d\n",
					EPSromb, JMAXromb, Kromb);
	   }
	else
	   PR"Analytic integration extended to infinity\n");

#ifdef DEBUG
 if(actionFlag == AVERAGES) {
   PE"N_fict, N_fict_1, N_fictO2, NkapDbl: %d, %d, %g, %d\n",
		N_fict, N_fict_1, N_fictO2, NkapDbl);
   PE"C1 = %g, NormFactor = %g\n", C1, NormFactor);
 }
#endif

 if(PlotWidth)
   { /* Initialize line-printer type plotting routines:  */
     initPlot((float)Cx, (float)CE, (float)xMinAt, (float)xMaxAt,
			1, PlotWidth, fmt, fmt_len, Listing);

     /* Plot potential:    */
     PR"\n");
     for(i = -2 * fmt_len; i < PlotWidth; i += 2) PR" ");
     PR"V(x)\n");
     plotFun1(V, fmtV);
   }
 PR"\n");
}


void OutDensity()
/*=============*/
{
int i;
float maxD = 0.;

PR"\n\t\t\tDensity");
if(PlotWidth)
  { PR" (normalized to Q)\n");
    for(i = 0; i < N_x; i++)
	if(fabs((double)dens[i]) > maxD) maxD = dens[i];
    if(maxD > 1000.) PR"\n\t\t     Density is too big to be plotted!\n");
    else Histogram(dens, N_x, (float)(xMinAt - 0.5 * xStepAt), (float)xStepAt);
  }
  else PR"\n");
maxD = 0.;
for(i = 0; i < N_x; i++) maxD += dens[i];
maxD *= (xStepAt * Cx);
PR"\nNumerically obtained estimate for Q = %g\n", maxD);

/*
for(i = 0; i < N_x; i++) PR"  %.6f    %g\n",
		 Cx * xd[i], dens[i] / 31.414);
exit(0);
*/

PR"\n      x        rho(x)           rho(x)\n");
PR"               norm\'ed to 1     normalized to Q\n");
PR"               numerically\n");
for(i = 0; i < N_x; i++) PR"  %.6f    %g     \t%g\n",
		 Cx * xd[i], dens[i] / maxD, dens[i]);

if(potFlag == HOpotFlag)
  { if(N_fict > 2)
	{
	  PR"\n    x         Exact quantum rho(x)\n");
	  for(i = 0; i < N_x; i++)
		PR"  %.6f    %g\n", (maxD = Cx * xd[i]), rhoHOexact(maxD));
	}
    else
	{
	  PR"\n    x    Exact quantum rho(x)     Exact primitive rho(x)\n");
	  for(i = 0; i < N_x; i++)
	    PR"  %.6f    %g     \t%g\n",
		 (maxD = Cx * xd[i]), rhoHOexact(maxD), rhoHOexactPrim2(maxD));
	}
  }
PR"\n");

}  /* --------------->  end of:  OutDensity()  <------------------ */


void writeOut()
/*===========*/
{
   double eps, kin;

   PR"Q = %.16g\n", Q * NormFactor);
   alfa /= (2. * Q * betaAt);
   lambda /= Q;
   kin = CE * (C4 - alfa);
   eps = kin + CE * lambda;
   eps_vir *= (CE / Q);
   PR"<alfa> = %.16g, <lambda> = %.16g\n", CE * alfa, CE * lambda);
   PR"<kin> = %.16g, 1/(2*beta) = %.16g\n", kin, 0.5 / beta);
   PR"<eps> = %.16g, <eps_vir> = %.16g\n", eps, eps_vir);
   if(potFlag == HOpotFlag)
     {	double	Eb2 = exp(-0.5 * betaAt),
		Eb = exp(-betaAt);

	PR"\tExact primitive-algorithm values for N_fict=%d:\n", N_fict);
	if(N_fict > 2)
	    { double R = betaAt / N_fict,
		     s = sqrt(4. + R * R),
		     f = 1. + 0.5 * R * (R + s),
		     fN = pow(f,(double)N_fict);

	      PR"\tQ = %.16g, <eps> = %.16g\n", sqrt(fN) / (fN - 1.),
		 (0.5 + 1./(fN - 1.)) * (R + s - 2./s) / f);
	    }
	else
	    { double b8 = 8. + betaAt * betaAt,
		     b16 = 8. + b8,
		     B = CE / (betaAt * b16);
	      
	      PR"\tQ = %.16g\n", 4./(betaAt * sqrt(b16)));
	      PR"\t<alfa> = %.16g, <lambda> = %.16g\n",
						8. * B, (lambda = b8 * B));
	      PR"\t<eps> = <eps_vir> = %.16g\n", 2. * lambda);
	    }
	PR"\tExact quantum values:\n");
	PR"\tQ = %.16g, <E> = %.16g\n", Eb2 / (1. - Eb),
		 0.5 * (1. + Eb) / (1. - Eb));
     }
   PR"\n");
   (void)fflush(Listing);
}


double fq1(u)  double u;
/*====================*/
{
}

double fq2(u)  double u;
/*====================*/
{
 int i;
 double b = exp(-betaAt * (*V)(u)),
	V2u = (*V2_4)(u),
	V2uP = (V2u > 0. ? V2u / C4 : 0.);

 /* doubly degenerate eigenvalues (even N_fict): */
 for(i = 0; i < NkapDbl; i++) b /= (kappa[i] + V2uP);
 /* a single eigenvalue equal to 4: */
 b /= sqrt(kappa0 + V2uP);
 return b;
}

double fal1(u)  double u;
/*=====================*/
{
}

double fal2(u)  double u;
/*=====================*/
{
 int i;
 double b = exp(-betaAt * (*V)(u)),
	s = 0.,
	alf = (*V2_4)(u),
	V2uP = (alf > 0. ? alf / C4 : 0.);

 for(i = 0; i < NkapDbl; i++) { alf = kappa[i] + V2uP;
				b /= alf;
				s += kappa[i] / alf;
			      }
 alf = kappa0 + V2uP;
 b /= sqrt(alf);
 s = 2. * s + kappa0 / alf;
 return  b * s;
}

double fla1(u)  double u;
/*=====================*/
{
}

double fla2(u)  double u;
/*=====================*/
{
 int i;
 double s = 0.,
	V2u = (*V2_4)(u),
	V2uPlain = (V2u > 0. ? V2u : 0.),
	V2uP = V2uPlain / C4,
	alf,
	Vu = (*V)(u),
	b = exp(-betaAt * Vu);

 for(i = 0; i < NkapDbl; i++) { alf = kappa[i] + V2uP;
				b /= alf;
				s += 1. / alf;
			      }
 alf = kappa0 + V2uP;
 b /= sqrt(alf);
 s = 2. * s + 1. / alf;
 return b * (Vu + V2uPlain * s / (double)N_fict);
/* return b * (Vu + V2u * s / (double)N_fict); */
}

double fev1(u)  double u;
/*=====================*/
{
}

double fev2(u)  double u;
/*=====================*/
{
 int i;
 double s = 0.,
	alf = (*V2_4)(u),
	V2uP = (alf > 0. ? alf / C4 : 0.),
	b = exp(-betaAt * (*V)(u));

 for(i = 0; i < NkapDbl; i++) { alf = kappa[i] + V2uP;
				b /= alf;
				s += 1. / alf;
			      }
 alf = kappa0 + V2uP;
 b /= sqrt(alf);
 s = 2. * s + 1. / alf;
/*
 return b * ((*epsVir)(u) + (*V23_2)(u) * s / N_fict);
*/
 alf = (*V23_2)(u);
 return b * ((*epsVir)(u) + (alf > 0. ? alf * s / N_fict : 0.)); 
}


float DensIntegrateOld(X)  double X;
/*================================*/
/* Returns densities scaled back by 1/Cx to external units */
{
 double b, v, aux, del, f, SqRoot, fP, fP2, N_fictM2 = N_fict - 2.;

 v = (*V)(X);
 aux = (*V2_4)(X);
 del = (aux > 0. ? aux / C4s : 0.);
 b = 0.5 * (*dV)(X);
 if(fabs(del) > 1e-17) {
   aux = 2. * pow(betaAt/N_fict, 3.);
   b = aux * b * b;
   SqRoot = sqrt(del * (del + 4.));
   f = 1. + 0.5 * (del + SqRoot);
   fP2 = pow(f, N_fictM2);
   fP = fP2 * f * f;
   aux = fP - 1. / fP;
   return (float) ( sqrt(C1 * SqRoot/aux) * exp(b * ( 1. / f + N_fictM2 / del +
	   ((2. * (f + 1.) + fP2 * (1. - 3. * f) + (f - 3.) / fP2) /
	   (del * (f - 1.)) - SqRoot / fP) / aux) - betaAt * v) / Cx );
 }
 else {
   b *= betaAt; 
   return (float) ( exp((b * b * (1. - 1. / (double)(N_fict * N_fict)) / 6. -
			v) * betaAt) / (s2piBeta * Cx) );
 }
}  /* --------------->  end of:  DensIntegrateOld()  <------------------ */


float DensIntegrate(X)  double X;
/*=============================*/
/* Returns densities scaled back by 1/Cx to external units */
{
 double b, v, v24, aux, del, f, SqR1, SqR2, fPp, fPm;

 v = (*V)(X);
 aux = (*V2_4)(X);
 v24 = (aux > 0. ? aux : 0.);
 b = 0.5 * (*dV)(X);
 if(fabs(v24) > 1e-17) {
   del = v24 / C4s;
   SqR1 = sqrt(v24) / C4;
   SqR2 = sqrt(del + 4.);
   f = 1. + 0.5 * SqR1 * (SqR1 + SqR2);
   aux = pow(f, N_fictO2);
   fPm = (aux - (fPp = 1. / aux)) / SqR1;
   fPp += aux; fPp /= SqR2;
   return (float) ( sqrt(C1 / (fPp * fPm)) *
	    exp(betaAt * (0.5 * b * b / v24 * (1. - fPm / (fPp * N_fict)) - v))
	    / Cx );
 }
 else {
   b *= betaAt; 
   return (float) ( exp((b * b * (1. - 1. / (double)(N_fict * N_fict)) / 6. -
			v) * betaAt) / (s2piBeta * Cx) );
 }
}  /* --------------->  end of:  DensIntegrate()  <------------------ */


float rhoHOexactPrim2(X)  double X;
/*===============================*/
{
 double b = 8. + betaAt * betaAt;

	return (float)( 2. * exp(-0.25 * betaAt * (8. + b) * X * X / b)
		 / (sqrt(M_PI *betaAt * b) * Cx) );
}


float rhoHOexact(X)  double X;
/*==========================*/
{
   double	Eb = exp(-betaAt),
		tEavEx = (1. + Eb) / (1. - Eb),
		Qex = exp(-0.5 * betaAt) / (1. - Eb);

   return (float) (Qex / sqrt(tEavEx * M_PI) * exp(-X * X / tEavEx) / Cx);
}
