/*****************************************************************************/
/********************** (C) 1992 by Miroslav Kolar ***************************/
/*									     */
/*			Direct Halton/Ranmar integration		     */
/*		  of the approximate density matrix normally		     */
/*        used in the path-integral quantum Monte-Carlo simulation	     */
/*    for a single particle in a one-dimensional external potential V(x)     */
/*									     */
/*  Internally, potential is assumed to be scaled to certain atomic-like     */
/*  units, so that the hamiltonian is  -(1/2)(d^2/dx^2) + V(x)		     */
/*									     */
/*****************************************************************************/

/*
External files required: 1D.h
			 1Dpotential.c
			 RanGen1.c
			 ranmar1.c
			 RanGen.c  (function Eratosthenes() only)
			 LinePlots.c
			 DateAndTime.c

 List of functions in this file:
	void printChoices(N, descr) int N; char *descr;
	void usage(mode) int mode;
	void defaults(mode) int mode;
	void help1()
	void help()
	main(argc, argv)  int argc; char *argv[];
	void Integration()
	void checkCL()
        void printAll(mode, msg) int mode; char *msg;
	void err_fseek(proc)  char *proc;
	void err_fread(proc)  char *proc;
	void write_offsets()
	void read_offsets(p) char *p;
	void read_head(p) char *p;
	void write_N();
	void read_N(p) char *p;
	void write_P1()
	void read_P1(p) char *p;
	void write_P2()
	void read_P2(p) char *p;
	void write_C()
	void read_C(p) char *p;
	void write_A()
	void read_A(p) char *p;
	void write_Lm(p) char *p;
	void OutData()
	void dynDefs()
	void xdDefs()
	void InData()
	void printParams(old_deltaX)  double old_deltaX;
	void plotDensity()
	void OutDensity(mode)  int mode;
	void writeOut()
	void Integrate()
	void DiagIntegrate()
	void DensIntegrate()
	void displayFile(Listing)  FILE *Listing;
	void Final()

External functions:
	char *DateAndTime();			(in  DateAndTime.c)
	void initPlot(), plotFun1(), Histogram();
						(in  LinePlots.c)
        Currently available potentials are defined in 1Dpotential.c:
		double (*V)();
		double (*epsVir)();  = V(x) + 0.5*(x-Shift)* dV(x)/dx
		void   (*inV)();     input of the parameters of the potential
		long   (*outV)();    output of the parameters of the potential
		void choosePotential();

	void Eratosthenes();		(in RanGen.c)
        Random number generators are defined in RanGen1.c.
*/

#include <errno.h>
#include <sys/times.h>
#include "1D.h"

/* #define DEBUG */
/* #define DEBUG1 */

#define SC         (void)scanf(
#define SI         (void)fscanf(Input,
#define FC         (void)fclose(
#define ILLEGAL    { PE" Illegal option '%s' ignored!\n",argv[i]); illeg++; }
#define FIERROR(x) { PR" File '%s': %s\n", x, sys_errlist[errno]); exit(-1); }
#define max(x, y)  ((x) < (y) ? (y) : (x))

#define MAXLETT    101   /* max length of file names, text lines, etc.
			    Must be at least 10  */
/* Some additional #defines can be found in 1D.h and before the OutData()
function. */

#define HALTON 	   0
#define RANMAR 	   1
#define CONT	   2
#define DISPLAY	   3
#define EPS	   4
#define MAXaction  5	/* Current maximum number of the actions
			   performed by the program. */

#define DIAG	   0	/* Diagonal integration - no densities	    */
#define FULL	   1	/* Full volume integration - all densities  */
#define DENS	   2	/* A single density only - "diagonal style" */

#define REMOVE	   0	/* For diag integration - no points outside */
#define ALL	   1	/*   -  "  -    - all generated  points considered */

/* Defaults:  */
#define DEFactionFlag		HALTON
#define DEFpotFlag		0      /* see 1Dpotential.c   */
#define DEFPlotInt		0
#define DEFPlotWidth		65
#define DEFMaxPoints		100000L
		/* Default maximum number of integration grid points.  */
#define DEFWriteInt		10000
#define DEFN_x			10
#define DEFN_fict		6
#define DEFseed			32737
#define DEFExponent		50.


/* ------ Global declarations: ------- */
#if defined(ultrix) || defined(__osf__)
extern int  errno;
extern void setbuf();
extern void *malloc();
#else
int  errno;
extern int setbuf();
extern char *malloc();
#endif
extern char *sys_errlist[];
extern double exp();
extern clock_t times();
extern void exit();

extern char *DateAndTime();			/* in DateAndTime.c  */
extern void initPlot(), plotFun1(), Histogram();
						/* in LinePlots.c    */

static clock_t usLast, sysLast;	/* used for timing */
static struct tms cpuT;		/*    - " -        */

static char *author = "                     Written by M. Kolar (C) 1992\n",
	    *space = "\n                                    ",
	    *space43 = "                                           ",
	    *progName;	/*  to be set equal to  argv[0]  */

 /* All important data enabling to continue/restart an integration are stored
    in what is referred below as a Master File. Depending on the action
    performed in a given run of the program (determined by  actionFlag), only
    an Input Master File, or only an Output Master File, or both are required.
    At the same time, Input and Output Master Files can be identical (only one
    of them is opened at any given time).
 */
static int actionFlag = DEFactionFlag;
	/* = HALTON - start a new integration using Halton numbers with
		      a random selection of values of n0.  [Output only]
	   = RANMAR - start a new integration using RANMAR random number
		      generator.  [Output only]
	   = CONT   - continue: add additional grid points to the integration
		      from the input file.  [Input/Output]
	   = DISPLAY - display the contents of a Master File.  [Input only]
	   = EPS     - write eps, eps_vir (or rho) into separate files as
		       function of integration points. [Input only]
	*/
static int rgFlag = DEFactionFlag,   /* either equal to HALTON or RANMAR
					whenever actually needed	*/
	   integrationFlag = DIAG,
	   outsideFlag = REMOVE,
	   noOutputFlag = 0;
static char *FULLstring = "FULL",
	    *ALLstring = "ALL",
	    *DENSstring = "rho";
static double Exponent = DEFExponent, /* to determine the width of the	*/
	      deltaX;	/* integration channel along the diagonal - so	*/
  /* as to fit in only the Boltzman factors larger than exp(-Exponent)	*/

static int PlotInt = DEFPlotInt,
		/*  Number of integration points between consecutive plots
		    of particle density.  */
	   PlotWidth = DEFPlotWidth;
		/*  Width (in characters) of the line-printer plots  */

static long
 point = 0L,	/* integration grid point counter */
 MaxPoints = DEFMaxPoints,
		/* Maximum number of grid points.  */
 outsidePoints = 0L,
		/* for DIAG, counts points that fall outside of the cube  */
 blocks = 0L,   /* counts blocks of integrals written in the  Output file */
 Lm_offset = 0L, /* offset for Last modification date in file  Output. */
 V_offset = 0L,  /* offset for possible use by (*inV)() and (*outV)(). */
 N_offset = 0L,  /* offset for N_fict.				       */
 P1_offset = 0L, /* offset for "constant" parameters.		       */
 P2_offset = 0L, /* offset for re-writeable parameters.		       */
 C_offset = 0L,  /* offset for current point and blocks values.	       */
 A_offset = 0L,  /* offset of the block averages section in the
		    file  Output.   */
 offset_Incr = 0L; /* diff of offsets of two consecutive recordings of
		      integrals in the file  Output.   */
static int WriteInt = DEFWriteInt;  
		/*  Number of integration points between consecutive
		    recordings of the current values of the integrals
		    in the Output file.  */

FILE *Input = NULL, *Output = NULL, *Listing = stdout;
static FILE *Density = NULL;
static char InputFileName[MAXLETT], OutputFileName[MAXLETT],
	/* Both names can be identical. Thus a Master File from a previous run
	   can be updated: Some parameters will be (over)written
	   at the beginning of the file, new values of integrals (for higher
	   numbers of grid points) will be appended at the end.
        */
	    ListingFileName[MAXLETT], /* To print out parameters, final results,
				     configurations during mixing, etc.
				     If no file is given, stdout is used.   */
	    DensityFileName[MAXLETT]; /* To output the final density plot. If
				      no file is given, append to Listing.  */

char    text[MAXLETT];  /* Auxilliary string storage */
double  beta = -1.,  /* beta = 1/(k*T);
			k - Boltzman constant, T - temperature  */
	m = -1.,     /* mass of the real particle */
	C1,          /* = N_fict * mAt / (2*pi*hAt*hAt*betaAt)      */
	C2,          /* = N_fict * mAt / (2*hAt*hAt*betaAt*betaAt)  */
	C3,          /* = 2 * C2                                    */
	C4;          /* = N_fict / (2*betaAt)			    */

int	    N_fict = DEFN_fict;    /* number of the fictitious particles  */
static int  N_fict_1;		   /*   = N_fict - 1			  */

float 	      xMin = 1e+30, xMax = -1e+30; /*  integration limits   */
static double xValue = 0., /* to calculate a single density at (DENS on) */
	      NormFactor;  /* to normalize Q, density */

/* Density of (fictitious) particles: */
static int    N_x = DEFN_x;	  /*  # of points to calculate density at  */
static double *xd,	/*  At this values (xd)		    | defined	   */
	      *densS;	/*  calculate density (densS).	    | dynanically  */
float	      *dens;	/*  Normalized density.		    | in	   */
static double *XD, *FD, *orthBas, *xi; 	/*		    | dynDefs()	   */


/* Other accumulators: */
static double
   alfaS = 0., lambdaS = 0., eps_virS = 0.,
	  /* See Herman, Bruskin, Berne, J.Chem.Phys. 76, 5150 (1982) */
	  /* eps_virS doesn't contain factor 1/N_fict; this must be added
	     during printing */
   QS = 0.;  /* canonical sum (coordinate dependent partition function) */

/* Atomic-like values used consistently in the internal representation:     */
static double	betaAt,		/* mAt = 1, hAt = 1 */
		xMinAt, xMaxAt,
		LAt,		/*  = (xMaxAt - xMinAt)		     */
		xStepAt,	/*  = (xMaxAt - xMinAt) / (N_x - 1)  */
		xValueAt;       /* to calculate a single density at  */
double	Cx = 0.,  /* coord. transformation: x = Cx * xAt	   */
	CE = 0.;  /* energy transformation: E = CE * E_atomic,
					    beta = betaAt / CE	   */
	/* For harmonic oscillator: Cx = CE = 1			   */


/* Command Line values:  */
static int clPlotInt = -1,
	   clPlotWidth = -1,
	   clN_fict = -1,
	   clpotFlag = -1,
	   clN_x = -1,
	   clseed, clseedFlag = 0,
	   clWriteInt = -1 ;
static long clMaxPoints = -1L;
static double clbeta = -1.;

#define P1237 (float)1.273e29
#define M1237 (float)-1.273e29
static float  p1273 = P1237, clxMin = P1237, 
	      m1273 = M1237, clxMax = M1237;


/*-----------------------------------------------------------------------*/
/*     Currently available potentials are defined in 1Dpotential.c       */

static int potFlag = DEFpotFlag;
	 /* equals to the number of the selected potential */

double (*V)();		/* pointer to the potential function		*/
double (*dV)();		/* pointer to dV(x)/dx				*/
double (*epsVir)();	/* pointer to the function returning
			V(x) + 0.5*(x-Shift)* dV(x)/dx  for the selected V(x),
			where Shift is nonzero for a potential defined in
			finite interval only, and is selected so that
			the virial estimator gives correct averages	  */
void   (*inV)();	/* pointers to functions for input and output   */
long   (*outV)();	/* of the parameters of the potential		*/
float  xRet1, xRet2;  /* (approximate) classical return points of the
                          potential for energy = Vminimum + 2 / beta,
                          or some other suitable limits			*/

extern char *potBlurb,   /* potential description for printParams()      */
	    *potLabel[], /* potential name abbreviations to be used on
			    command line [and in help() and short help]  */
	    *potHelp[];  /* more information on potentials for help()    */
extern int  numPot;      /* total number of currectly defined potentials */
extern void choosePotential();  /* and perform potential initialization  */

/*	End of definitions for the file 1Dpotential.c			 */
/*-----------------------------------------------------------------------*/

/* RANMAR from file ranmar1.c is used here: */
extern int rmarin1();
extern double ranmar1();

/*-----------------------------------------------------------------------*/
/* The basic algorithms for the Halton numbers as presented
   in the paper "Fast, portable, and reliable algorithm for ...."
   are in the file RanGen1.c.
   It also contains the generator of the full Halton vector using the 
   first nBase primes, and an unsuccesful "experimental generator"
   based on binary Halton.
*/

extern double init_HaltonBS();
extern double HaltonBS();
extern void init_Halton();
extern double Halton();
extern void init_Halton2();
extern double Halton2();
extern void init_vHalton();
extern void rand_init_vHalton();
extern double vHalton();
/* extern void init_ExprG(); */
/* extern double ExprG();    */
extern int N0[];
/*	End of definitions for the file RanGen1.c			 */
/*-----------------------------------------------------------------------*/

double (*pR)();  /* pointer to the selected rand' num' gen' */
int seed  = DEFseed;  /* seed to start randomly HALTON or RANMAR  */

extern void Eratosthenes();
	/* from RanGen.c;  needed in init_vHalton() and rand_init_vHalton() */
int rg1Flag, rg2Flag;
	/* needed to link functions from RanGen.c; unused here */

/* For printing of usage & defaults, etc.:  */
static char *action[MAXaction];
	    /*  for *potLabel[] and *potHelp[]  see  1Dpotential.c  */


/* Some of the functions defined below the Main:  */
 void   Integration(), checkCL(), dynDefs(), xdDefs(), InData(), Integrate(),
	DiagIntegrate(), DensIntegrate(), displayFile(), Final();
#ifdef DEBUG
 void printAll();
#endif


void printChoices(N, descr) int N; char *descr[];
/*=============================================*/
{
int n;

PE"{"); PE *descr);
for(n = 1; n < N; n++) { PE" "); PE descr[n]); }
PE"}");
}

void usage(mode) int mode;
/*======================*/
{
PE"\7\nUsage:"); if(mode) PE"\n");
PE"  %s options [ > file]\n", progName);
if(mode)
 PE"\n  options: Action pPotential iInFile oOutFile lListFile dDensFile noO\n");
 else
 { PE"  options: ");
   printChoices(MAXaction, action);
   PE" iInFile oOutFile lListFile dDensFile noO\n");
   PE"           p");
   printChoices(numPot, potLabel);
   PE"\n");
 }
PE"           fNfict hNhist xXmin XXmax mpMaxPoints wWriteInt bBeta sSeed\n");
PE"           eExponent piPlotInt pwPlotWidth\n");
PE"           %s %s %sXvalue                       ( -h  ?  -help )\n",
		FULLstring, ALLstring, DENSstring);
}

void defaults(mode) int mode;
/*=========================*/
{
PE"Defaults:\n ");
PE"  %s  p%s  f%d  h%d  mp%ld  w%d  s%d  e%.16g  pi%d  pw%d\n",
        action[DEFactionFlag], potLabel[DEFpotFlag], DEFN_fict, DEFN_x,
	DEFMaxPoints, DEFWriteInt, DEFseed, DEFExponent,
	DEFPlotInt, DEFPlotWidth);
PE"     \"Diagonal\" integration       No points outside the cube\n\n");
if(mode)  PE"Example: %s  b0.4 o93_1 h12 x0.23 X15.0\n\n", progName);
}

void help1()
/*========*/
{
usage(0);
defaults(0);
PE"########### If you need more help, enter command: %s -help ###########\n\n",
	progName);
}

void help()
/*=======*/
{
	int n;
PE"\n         Direct Halton/Ranmar Integration Corresponding to the\n");
PE"             Quantum Monte-Carlo Simulation for a Single\n");
PE"           Particle in a One-Dimensional External Potential\n");
PE author);
usage(1);
PE"\n  Only some of these options need be given in any order. All essential\n");
PE"  parameters not supplied (as defaults, from InFile) will be prompted for.\n");
PE"  Priority: 1. Command line values, 2. InFile values, 3. Defaults\n");
PE"  Tip: start the program interactively, it will ask you for all necessary\n");
PE"  parameters ommitted on the command line, then put it in the background.\n\n");
PE"  Action: %s - start new integration using Halton numbers\n", action[HALTON]);
PE"          %s - start new integration using RANMAR\n", action[RANMAR]);
PE"          %s - continue in the integration from InFile\n", action[CONT]);
PE"          %s - display the contents of InFile\n",	action[DISPLAY]);
PE"          %s - write eps, eps_vir (or rho) into separate files\n",
					action[EPS]);
PE"  (No results are printed during calculation; use Action D or E to reveal them)\n\n");
PE"  Potential:  "); PE potLabel[0]); PE potHelp[0]);
for(n = 1; n < numPot; n++) {
PE"\n              "); PE potLabel[n]); PE potHelp[n]);
			    }
PE"\n\n  Beta:       1/(kT)\n\n");
PE"  Nfict:      # of fictitious particles for quantum MC\n\n");
PE"  Exponent: As a default, integration is done only over a rectangular\n");
PE"            channel along the body diagonal of the Nfict-dimensional space.\n");
PE"            The width of the channel is determined to fit in only\n");
PE"            the product of Boltzman factors larger than exp(-Exponent).\n");
PE"            No density is calculated.\n");
PE"      - As a default, the channel is pointed at the ends (bounded by the\n");
PE"        faces of the Nfict-D cube - the full integration volume). To see\n");
PE"        the effect of using the same width till the end of the channel,\n");
PE"        specify the option %s.\n", ALLstring);
PE"      - To perform the integration over the whole volume, specify the\n");
PE"        option %s. \n\n", FULLstring);
PE"  Xvalue: only a single density calculated at x=Xvalue: integration is\n");
PE"          over an (Nfict-1)-D cube around Xvalue with the length of its\n");
PE"          edge determined by Exponent as above. When %s is specified\n",
			ALLstring);
PE"          integration is always over the whole volume of this smaller\n");
PE"          cube.  Default: only over its portion inside the big cube.\n\n");
PE"      - Options %s and %s are mutually exclusive.\n",
			FULLstring, DENSstring);
PE"        %s has no effect with %s.\n", ALLstring, FULLstring);
PE"        All three have no effect when Action = %s, %s.\n\n",
			action[CONT], action[DISPLAY]);
PE"  MaxPoints: Max number of integration grid points\n\n");
PE"  Seed: abs(Seed) used as the seed for drand48/mrand48 used to select at\n");
PE"        random the seeds of RANMAR, or initial indeces n_0 for different\n");
PE"        components of the Halton vector. For HALTON, if  Seed < 0, then\n");
PE"        n_0 = -Seed  for all Halton components\n\n");
PE"  WriteInt: Interval (# of points) between writes to OutFile\n\n");
PE"  Xmin, Xmax:  integration limits\n");
PE"  Nhist: # of points between Xmin and Xmax to calculate the density at.\n");
PE"         Applies only when %s is specified\n\n", FULLstring);
PE"  PlotInt: Interval (# of points) between plots of density and printing\n");
PE"           of current values of integrals (PlotInt = 0 means no output\n");
PE"           of intermediate results during integration)\n");
PE"           PlotInt (WriteInt) may be adjusted so that PlotInt is a\n");
PE"           fraction of WriteInt\n\n");
PE"  PlotWidth: Width (in characters) of all line-printer type plots\n");
PE"             (PlotWidth = 0 means no plotting at all)\n\n");
PE"  If no DensFile is given, density plot is appended to ListFile. If no\n");
PE"  ListFile is given, standard output is used which can be redirected to\n");
PE"  a file. OutFile can be identical with InFile (if only one is given,\n");
PE"  they are assumed identical). Option \"noO\" means that no OutFile\n");
PE"  is created. If certain command line parameters (Beta, integration\n");
PE"  limits or # of density points, etc.) differ from those in InFile,\n");
PE"  new integration must be started.\n\n");
defaults(1);
exit(0);
}
/* End of help() */



/*=====================================*/
/*    MAIN program starts here!!!!     */
/*=====================================*/
main(argc, argv)  int argc; char *argv[];
/*=====================================*/
{
  int  i, atoi(), n,
       illeg = 1; /* see #define ILLEGAL  */
  long atol();
  int  strcmp(), strncmp();
  char *ARG, *strcpy();
  double atof(), aux;

/*  Define the wording of the command line commands/parameters:  */
 action[HALTON] = "H";
 action[RANMAR] = "R";
 action[CONT] = "C";
 action[DISPLAY] = "D";
 action[EPS] = "E";

#ifdef DEBUG
 PE"\nAvailable potentials:  ");
 for(n = 0; n < numPot; n++) { PE potLabel[n]); PE"   "); }
 PE"\n");
#endif

progName = argv[0];
for (i=1; i<argc; i++)
  {
   ARG = argv[i];
#ifdef DEBUG
   PE"\n..... processing '%s'\n",ARG);
#endif
   if ( ! strcmp(ARG,"-help") || ! strcmp(ARG,"-HELP") )  help();
   if ( ! strcmp(ARG,"-h") || ! strcmp(ARG,"-H") ||
	! strcmp(ARG,"?") || ! strcmp(ARG,"-?") )  { help1(); exit(0); }

     else if(! strcmp(ARG,"noO"))	    noOutputFlag = 1;
     else if(! strcmp(ARG,action[HALTON]))  actionFlag = rgFlag = HALTON;
     else if(! strcmp(ARG,action[RANMAR]))  actionFlag = rgFlag = RANMAR;
     else if(! strcmp(ARG,action[CONT]))    actionFlag = CONT;
     else if(! strcmp(ARG,action[DISPLAY])) actionFlag = DISPLAY;
     else if(! strcmp(ARG,action[EPS]))	    actionFlag = EPS;
     else if(! strcmp(ARG,FULLstring))	    integrationFlag = FULL;
     else if(! strcmp(ARG,ALLstring))	    outsideFlag = ALL;
     else if(! strncmp(ARG,DENSstring,(n = (int)strlen(DENSstring))) )
           { ARG += n; xValue = atof(ARG); integrationFlag = DENS; }
     else if(*ARG == 'e')
           { aux = atof(++ARG); if(aux > 0.) Exponent = aux; else ILLEGAL }
     else if(! strncmp(ARG,"pi",2))
       { ARG += 2; if(isdigit(*ARG)) clPlotInt = atoi(ARG); else ILLEGAL }
     else if(! strncmp(ARG,"pw",2))
       { ARG += 2; if(isdigit(*ARG)) clPlotWidth = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'p')
           { ++ARG; for(n = 0; n < numPot; n++)
			{ if(! strncmp(ARG,potLabel[n],strlen(potLabel[n])) )
						{ clpotFlag = n; break; }
			}
             if( clpotFlag == -1) ILLEGAL
           }
     else if(*ARG == 'i') strcpy(InputFileName,++ARG);
     else if(*ARG == 'o') strcpy(OutputFileName,++ARG);
     else if(*ARG == 'l') strcpy(ListingFileName,++ARG);
     else if(*ARG == 'd') strcpy(DensityFileName,++ARG);
     else if(*ARG == 'f')
       { ++ARG; if(isdigit(*ARG)) clN_fict = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'h')
       { ++ARG; if(isdigit(*ARG)) clN_x = atoi(ARG); else ILLEGAL }
     else if(*ARG == 'x') clxMin = (float)atof(++ARG);
     else if(*ARG == 'X') clxMax = (float)atof(++ARG);
     else if(*ARG == 'b')
           { aux = atof(++ARG); if(aux > 0.) clbeta = aux; else ILLEGAL }
     else if(! strncmp(ARG,"mp",2))
       { ARG += 2; if(isdigit(*ARG)) clMaxPoints = atol(ARG); else ILLEGAL }
     else if(*ARG == 'w')
       { ++ARG; if(isdigit(*ARG)) clWriteInt = atoi(ARG); else ILLEGAL }
     else if(*ARG == 's')
       { ++ARG; if(*ARG == '-' || isdigit(*ARG))
		     { clseed = atoi(ARG); clseedFlag = 1; }
		else ILLEGAL
       }
     else ILLEGAL
  }

if (illeg == argc) help1(); 

/* Initialize Files:  */
if(ListingFileName[0] != '\0')
 { if( (Listing = fopen(ListingFileName,"w")) == NULL)
				  FIERROR(ListingFileName)
 }
if(actionFlag == CONT || actionFlag == DISPLAY)
 { if(actionFlag == CONT)
	{ if(InputFileName[0] != '\0' && OutputFileName[0] == '\0')
			strcpy(OutputFileName,InputFileName);
	  if(InputFileName[0] == '\0' && OutputFileName[0] != '\0')
			strcpy(InputFileName,OutputFileName);
	}
   while(InputFileName[0] == '\0')
	{ PR"\nEnter input file name: "); SC"%s", InputFileName); }
 }

if(actionFlag < CONT && ! noOutputFlag)
 { while(OutputFileName[0] == '\0')
	{ PR"\nEnter output file name: "); SC"%s", OutputFileName); }
 }

PR space); PR DateAndTime());
PR"\n\n         Direct Halton/Ranmar Integration Corresponding to the\n");
PR"             Quantum Monte-Carlo Simulation for a Single\n");
PR"           Particle in a One-Dimensional External Potential\n");
PR author); PR"\n");

#ifdef DEBUG
   printAll(0, "main() - after argument processing");
#endif

switch (actionFlag)
 {
   case HALTON:
   case	RANMAR:  checkCL();
		 dynDefs();
		 xdDefs();
		 if(integrationFlag == FULL)
			for(i = 0; i < N_x; i++) densS[i] = 0.;
		 Integration();
		 break;
   case	CONT:	 InData();  /* contains dynDefs() */
		 checkCL();
		 xdDefs();
		 Integration();
		 break;
   case	DISPLAY: InData();
		 checkCL();
		 xdDefs();
		 displayFile(Listing);
		 break;
   case	EPS:	 InData();
		 checkCL();
		 xdDefs();
		 Final();
 }
(void)times(&cpuT);
PR"CPU time used by the whole job:\n    user  %g, system  %g, total  %g\n",
	(float)cpuT.tms_utime / 60., (float)cpuT.tms_stime / 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime) / 60.);
PR DateAndTime()); PR"\n\n");
FC Listing);
exit(0);
}  /* ------------------------ end of Main --------------------------  */


void Integration()
/*==============*/
{
switch (integrationFlag)
 {
   case FULL :	Integrate();
		break;
   case DIAG :	DiagIntegrate();
		break;
   case DENS :  DensIntegrate();
 }
}


void checkCL()
/*============         Activate all command line values that make sense  */
{
int chng = 0;
void printParams();
double old_deltaX = -1;

if(clPlotWidth != -1) PlotWidth = clPlotWidth;
/*	else if(actionFlag == DISPLAY) PlotWidth = 0;  */

if(actionFlag == DISPLAY)
{

  if(clWriteInt != -1) chng = 1;
  if(clPlotInt != -1) chng = 1;
  if(clN_x != -1) chng = 1;
  if(clpotFlag != -1) chng = 1;
  if(clMaxPoints != -1) chng = 1;
  if(clbeta != -1.) chng = 1;
  if(clxMin != p1273) chng = 1;
  if(clxMax != m1273) chng = 1;
  if(clN_fict != -1) chng = 1;
  if(chng)
	{
	 PR"\n When %s is specified, all other command line parameters,\n",
		action[DISPLAY]);
	 PR" except for i... l...  have no effect, of course!\n");
	}

}
else
{

if(clN_x != -1) { if(actionFlag == CONT && clN_x != N_x) chng = 1;
		  N_x = clN_x;
		}
if(clpotFlag != -1) { if(actionFlag == CONT && clpotFlag != potFlag)
				{ PR"\nPotential required on the command line\
 differs from that in input file!\n");
				  exit(-1);
				}
		      potFlag = clpotFlag;
		    }
if(clMaxPoints != -1) MaxPoints = clMaxPoints;
if(clWriteInt != -1) WriteInt = clWriteInt;
if(clPlotInt != -1) PlotInt = clPlotInt;
if(clseedFlag) seed = clseed;
if(clbeta != -1.)
  { if(actionFlag == CONT && clbeta != beta)
	{ PR"\n \'beta\' from the command line differs from that in input\
 file.\n");
	  PR" New integration must be started!\n\n");
	  exit(-1);
	}
    beta = clbeta;
  }
if(clxMin != p1273 && clxMax != m1273 && clxMin < clxMax)
	{ if(actionFlag == CONT && (xMin != clxMin || xMax != clxMax))
								 chng = 1;
	  xMin = clxMin; xMax = clxMax;
	}
if(clxMin == p1273 && clxMax != m1273 && xMin < clxMax)
	{ if(actionFlag == CONT && xMax != clxMax) chng = 1;
	  xMax = clxMax;
	}
if(clxMin != p1273 && clxMax == m1273 && clxMin < xMax)
	{ if(actionFlag == CONT && xMin != clxMin) chng = 1;
	  xMin = clxMin;
	}

if(actionFlag == CONT && chng)
	{ PR"\n Integration limits and/or the number of density points from the\
 command line differ from those in the Input file.\n");
	  PR" New integration must be started!\n\n");
	  exit(-1);
	}

if(clN_fict != -1) { if(actionFlag == CONT && clN_fict != N_fict)
				{ PR"\nNumber of fictitious particles\
 required on the command line differs\n");
				  PR"from that in the input file!\n");
				  exit(-1);
				}
		     N_fict = clN_fict;
		   }

/* Adjust PlotInt (WriteInt) so that PlotInt is a fraction of WriteInt: */
{ /* Round WriteInt to thousands or one significant digit */
  int j = 0, lastD = 0;
  while(WriteInt > 10 && j < 3)
		 { lastD = WriteInt % 10; j++; WriteInt /= 10; }
  if(lastD > 4) WriteInt++;
  while(j-- > 0) WriteInt *= 10;
}
if(PlotInt == 0)
  { if(noOutputFlag) PlotInt = WriteInt; }
  else
  { if(PlotInt > WriteInt) PlotInt = WriteInt;
    else { int	i = (WriteInt + 0.5 * PlotInt) / PlotInt,
		j = 0, lastD = 0, im, iM, dir = -1;
	   while(i > 10) { lastD = i % 10; j++; i /= 10; }
	   if(lastD > 4) i++;
	   while(j-- > 0) i *= 10;
	   im = iM = i;
	   if(im <= 1) dir = 1;
	   while((PlotInt = WriteInt / i) * i != WriteInt)
		if(dir < 0) { dir = 1; i = --im; }
		     else   { if(im > 1) dir = -1; i = ++iM; }
	 }
  }

MaxPoints = max(WriteInt, ((MaxPoints + WriteInt / 2) / WriteInt) * WriteInt);

#ifdef DEBUG
PR"MaxPoints = %ld, WriteInt = %d, PlotInt = %d\n",
		MaxPoints, WriteInt, PlotInt);
#endif

} /* ----  end of:  if(actionFlag == DISPLAY)  ... else ...  ----  */

if(actionFlag == HALTON || actionFlag == RANMAR)
			/*  I.e., a new integration starts  */
 {
  if(N_fict < 2) { PR"\nNumber of fictitious particles increased to 2!\n\n");
		   N_fict = 2;
		 }

  while(beta <= 0.0)
	 { PE"\7Enter positive value of beta: "); SC"%lf", &beta); }

  choosePotential(potFlag, 0);  /* among others, it calculates
		      Cx and CE, and C1, C2 and C3 */

  while(xMax <= xMin)
	{ PR"\nFor integration limits, enter xMin < xMax: ");
	  SC"%f%f", &xMin, &xMax); }

  if( integrationFlag == DENS && (xValue < xMin || xValue > xMax) )
	{
	 PR"\n x = %g to calculate the density at is outside of (%g,%g)!\n\n",
			xValue, xMin, xMax);
	 exit(-1);
	}

 }

xMaxAt = xMax / Cx;
xMinAt = xMin / Cx;
LAt = xMaxAt - xMinAt;
xStepAt = LAt / (N_x - 1);
betaAt = beta * CE;
C4 = N_fict / (2 * betaAt);
if(integrationFlag == DENS) xValueAt = xValue / Cx;

N_fict_1 = N_fict - 1;

if(integrationFlag != FULL)
		 deltaX = sqrt(Exponent * log(10.) / (betaAt * C2 * N_fict));

if(integrationFlag == DENS)
 {  /* Make sure that the little cube is completely inside the big one: */
    if(LAt < deltaX)
	{ old_deltaX = deltaX; 
	  deltaX = LAt;
	}
 }
{  /* Calculate normalization factor (includes transformation of
	integration variables) */
   double fact;
   int i;

   switch (integrationFlag)
    {
      case FULL :  fact = NormFactor = LAt;
		   i = 1;
		   break;
      case DIAG :  fact = deltaX;
		   NormFactor = LAt * sqrt((double) N_fict);
		   i = 1;
		   break;
      case DENS :  fact = NormFactor = deltaX;
		   i = 2;
    }
   for( ; i < N_fict; i++) NormFactor *= fact;
   NormFactor *= pow(C1, 0.5 * N_fict);
}
 
/* Print out all parameters, plot potential: */
printParams(old_deltaX);

#ifdef DEBUG
   printAll(0, "checkCL() - end");
#endif

}  /* ------------------- end of checkCL() -----------------------  */


#ifdef DEBUG
void printAll(mode, msg) int mode; char *msg;
/*=========================================*/
{
 int i;

 PE"\nGLOBAL PARAMETERS ("); PE msg); 
 PE"):\nactionFlag = %d, potFlag = %d\n", actionFlag, potFlag);
 PE"beta = %g, betaAt = %g\n", beta, betaAt);
 PE"xMin = %g, xMinAt = %g, xMax = %g, xMaxAt = %g\n",
	 xMin, xMinAt, xMax, xMaxAt);
 PE"N_x = %d, LAt = %g, xStepAt = %g\n", N_x, LAt, xStepAt);
 PE"Cx = %g, CE = %g, m = %g\n", Cx, CE, m);
 PE"C1 = %g, C2 = %g\nC3 = %g, ", C1, C2, C3);
 PE"C4 = %g\n", C4);
 PE"MaxPoints = %ld, WriteInt = %ld\n", MaxPoints, WriteInt);
 PE"PlotInt = %d, PlotWidth = %d\n", PlotInt, PlotWidth);
 PE"point = %ld, blocks = %ld\n", point, blocks);

 PE"N_fict = %d\n", N_fict);
 PE"lambdaS = %g, alfaS = %g\n", lambdaS, alfaS);
 PE"eps_virS = %g, QS = %g\n\n", eps_virS, QS);
 if(mode && integrationFlag == FULL)
   { PE"xd:"); for(i = 0; i < N_x; i++) PE" %g", xd[i]); PE"\n\n");
     PE"densS:"); for(i = 0; i < N_x; i++) PE" %g", densS[i]); PE"\n\n");
   }

 PE"rgFlag = %d, seed = %d\n", rgFlag, seed);
 PE"integrationFlag = %d\n", integrationFlag);
 if(integrationFlag != FULL)
	PE"Exponent = %g, deltaXatomic = %g, outsideFlag = %d\n",
			 Exponent, deltaX, outsideFlag);
 if(integrationFlag == DENS)
	PE"xValue = %g, xValueAt = %g\n", xValue, xValueAt);

 if( actionFlag == HALTON || actionFlag == RANMAR ||
			 strcmp(OutputFileName,InputFileName) )
  { PE"\nLm_offset = %ld, V_offset = %ld, N_offset = %ld\n",
	 Lm_offset, V_offset, N_offset);
    PE"P1_offset = %ld, P2_offset = %ld, C_offset = %ld\n",
	 P1_offset, P2_offset, C_offset);
    PE"A_offset = %ld, offset_Incr = %ld", A_offset, offset_Incr);
  }

 PE"\n\n");
}
#endif


void err_fseek(proc)  char *proc;
/*=============================*/
{  PR"\7\n"); PR proc);  PR": fseek error\n");  exit(-1);  }

void err_fread(proc)  char *proc;
/*=============================*/
{  PR"\7\n");
   PR proc);
   if(feof(Input))  PR": End of Input file\n");
	else	    PR": fread error\n");
   exit(-1);
}

#define SEEKBEG(F,o) if( fseek(F, o, SEEK_SET) )  err_fseek(p);

void write_offsets()
/*================*/
{
 WL(&Lm_offset, 1)  WL(&V_offset, 1)  WL(&N_offset, 1)  WL(&P1_offset, 1) 
 WL(&P2_offset, 1)  WL(&C_offset, 1)  WL(&A_offset, 1)  WL(&offset_Incr, 1) 
}

void read_offsets(p) char *p;
/*=========================*/
{
 RL(&Lm_offset, 1)  RL(&V_offset, 1)  RL(&N_offset, 1)  RL(&P1_offset, 1) 
 RL(&P2_offset, 1)  RL(&C_offset, 1)  RL(&A_offset, 1)  RL(&offset_Incr, 1) 
}

void read_head(p) char *p;
/*======================*/
{
 SEEKBEG(Input, Lm_offset)
 while(getc(Input) != '\n') ;
 SI"%s", text);
#ifdef DEBUG
 PE"%s\n", text);
#endif
 if( strcmp(text,"DIRECT") )
  { PR"\n Wrong input file!!\n\n");
    exit(-1);
  }
 SI"%s", text);
#ifdef DEBUG
 PE"%s\n", text);
#endif
 if( strcmp(text,"INTEGRATION") )
  { PR"\n Wrong input file!!\n\n");
    exit(-1);
  }

 SI"Potential # %d", &potFlag);
 choosePotential(potFlag, 1);
 (*inV)(V_offset);
} /* read_head() */

void write_N()
/*==========*/
{
 PO"N_fict = %d\n", N_fict);
 PO"xMin = %.8g, xMax = %.8g, N_x = %d\n", xMin, xMax, N_x);
 PO"rgFlag = %d\n", rgFlag);
 PO"integrationFlag = %d, outsideFlag = %d\n", integrationFlag, outsideFlag);
}

void read_N(p) char *p;
/*===================*/
{
 SEEKBEG(Input, N_offset)
 SI"N_fict = %d\n", &N_fict);
 SI"xMin = %*f, xMax = %*f, N_x = %d\n", &N_x);
 SI"rgFlag = %d\n", &rgFlag);
 SI"integrationFlag = %d, outsideFlag = %d\n", &integrationFlag, &outsideFlag);
}

void write_P1()
/*===========*/
{
 WD(&beta, 1)  WD(&Cx, 1)  WD(&CE, 1) WF(&xMin, 1)  WF(&xMax, 1)
 WD(&C1, 1)  WD(&C2, 1)
 if(integrationFlag != FULL) { WD(&Exponent, 1) }
 if(integrationFlag == DENS) { WD(&xValue, 1) }
}

void read_P1(p) char *p;
/*====================*/
{
 SEEKBEG(Input, P1_offset)
 RD(&beta, 1)  RD(&Cx, 1)  RD(&CE, 1) RF(&xMin, 1)  RF(&xMax, 1)
 RD(&C1, 1)  RD(&C2, 1)
 if(integrationFlag != FULL) { RD(&Exponent, 1) }
 if(integrationFlag == DENS) { RD(&xValue, 1) }
}

void write_P2()
/*===========*/
{
 WL(&MaxPoints, 1)
 WI(&PlotInt, 1) WI(&WriteInt, 1)
}

void read_P2(p) char *p;
/*====================*/
{
 SEEKBEG(Input, P2_offset)
 RL(&MaxPoints, 1)
 RI(&PlotInt, 1) RI(&WriteInt, 1)
}

void write_C()
/*==========*/
{
#define IGPOINT   2140000000L
#define FORMAT   "point = %11ld\nblocks = %11ld"
#define WARNING  { PR"\npoint > IGPOINT --> point approaches allowable \
maximum!\n");\
 PR"  For larger values, specify point, etc. as double, and adjust FORMAT\n");\
 PR"  and all other formats involving point, etc., accordingly!\n");\
 PR"  Don't update old files using a wider new format!\n\n"); exit(-1); }

 if(point > IGPOINT)  WARNING
 PO FORMAT, point, blocks);
}

void read_C(p) char *p;
/*===================*/
{
 SEEKBEG(Input, C_offset)
 SI"point = %ld\nblocks = %ld", &point, &blocks);
}

void write_A()
/*==========*/
{
 WL(&point, 1)
 if(integrationFlag != DENS)
	{ WD(&alfaS, 1)  WD(&lambdaS, 1)  WD(&eps_virS, 1) }
 WD(&QS, 1)
 if(integrationFlag == FULL) { WD(densS, N_x) }
 if(integrationFlag == DIAG && outsideFlag == REMOVE) { WL(&outsidePoints, 1) }
}

void read_A(p) char *p;
/*===================*/
{
/* Note the value of  point  at the beginning of each A block is always
   read separately prior to the call to read_A()
*/ 
 if(integrationFlag != DENS)
	{ RD(&alfaS, 1)  RD(&lambdaS, 1)  RD(&eps_virS, 1) }
 RD(&QS, 1)
 if(integrationFlag == FULL) { RD(densS, N_x) }
 if(integrationFlag == DIAG && outsideFlag == REMOVE) { RL(&outsidePoints, 1) }
}

void write_Lm(p) char *p;
/*=====================*/
{
 SEEKBEG(Output, Lm_offset)
 PO space43);
 SEEKBEG(Output, Lm_offset)
 PO DateAndTime());
}


void OutData()
/*==========*/
{
int out = 0;  /* out will be set to 1 only if an output file name is entered
		 interactively  */
char *p = "OutData";
char over[2];

if(noOutputFlag) return;

#ifdef DEBUG
 printAll(1, "beginning of OutData()");
#endif

while(OutputFileName[0] == '\0')
	{ out = 1;
	  PE"\7Enter output file name: ");
	  SC"%s", OutputFileName);
	}
	     
if(actionFlag == HALTON || actionFlag == RANMAR ||
			 strcmp(InputFileName,OutputFileName))
{

 /* Here, an output file is being built from scratch: */

Try:
 if( (Output = fopen(OutputFileName,"a")) == NULL)
				  FIERROR(OutputFileName)
 if(ftell(Output) != 0L)
	{ PE"\7File %s exist! Overwrite? (y/n) ... ", OutputFileName);
	  SC"%s", over);
	  if(over[0] != 'y') {  PE"\7Enter new output file name: ");
				SC"%s", OutputFileName);
				out = 1;
				goto Try;
			     }
	}
 FC Output);
/* if( (Output = fopen(OutputFileName,"wb+")) == NULL) */
 if( (Output = fopen(OutputFileName,"w+")) == NULL)
				  FIERROR(OutputFileName)
 if(out) PR"\nOutput file: %s\n\n",OutputFileName);


 write_offsets();    /* At this point, this is just to hold place.
			The real offsets will be obtained later.     */
 PO"\nThe first line above contains all needed offsets (in binary).\n");
 PO"Most data below are also in binary.\n\n");

 PO"Created: ");  PO DateAndTime());
 PO"\nLast modified: ");
 Lm_offset = ftell(Output);
 PO space43);   /* to hold 43 bytes for the last-modification date */
 PO"\n\nDIRECT INTEGRATION OF THE CLASSICAL PARTITION FUNCTION EQUIVALENT\n");
 PO"TO THE QUANTUM PATH-INTEGRAL MONTE-CARLO APPROXIMATION FOR 1 PARTICLE\n\n");

 PO"Potential # %d\n", potFlag);
 V_offset = (*outV)(Output);
 PO"\nbeta = %.18g\n", beta);

 N_offset = ftell(Output);
 write_N();

 PO"\nP1:\n");
 P1_offset = ftell(Output);
 write_P1();

 PO"\n\nP2:\n");
 P2_offset = ftell(Output);
 write_P2();

 PO"\n\nC:\n");
 C_offset = ftell(Output);
 write_C();

 /* Intermediate values of the integrals: */
 PO"\n\nA:\n");
 A_offset = ftell(Output);
 write_A();
 /* Again, at this time, dummy values of integrals were written out
    to hold the space so as to enable to calculate offset_Incr that is
    equal to the number of bytes needed to write one record.
 */
 offset_Incr = ftell(Output) - A_offset;

 /* Now, the actual values of offsets are written into the reserved space at
    the beginning of the Output:  */
 SEEKBEG(Output, 0L)
 write_offsets();

#ifdef DEBUG
 PR"\noffset_Incr = %ld bytes\n\n", offset_Incr);
#endif

}
else
{

 /* Here, an existing output file is updated:  */

 Output = Input;
 if(out) PR"\nOutput file: %s\n\n",OutputFileName);

 SEEKBEG(Output, P2_offset)
 write_P2();

}

 write_Lm(p);

#ifdef DEBUG
 printAll(1, "end of OutData()");
#endif

}  /*  ---------------  end of OutData() ----------------  */


void dynDefs()
/*==========*/
{
  if(integrationFlag == FULL)
    { xd = (double *)malloc(N_x * sizeof(double));
      densS = (double *)malloc(N_x * sizeof(double));
      dens = (float *)malloc(N_x * sizeof(float));
      FD = (double *)malloc(N_x * sizeof(double));
      if(actionFlag != DISPLAY) XD = (double *)malloc(N_x * sizeof(double));
    }
  if(integrationFlag == DIAG)
	{ orthBas = (double *)malloc(2 * (N_fict - 1) * sizeof(double));
	  xi = (double *)malloc(N_fict * sizeof(double));
	}
}


void xdDefs()
/*=========*/
/* Must follow both  dynDefs()  and  checkCL() */
{
 int i;

 if(integrationFlag == FULL)
   { xd[N_x - 1] = xMaxAt;
     for(i = N_x - 2; i >= 0; i--) xd[i] = xd[i + 1] - xStepAt;

     for(i = 0; i < N_x; i++) FD[i] = exp(-betaAt * (*V)(xd[i]) / N_fict);
   }
}


void InData()
/*=========*/
{
char *p = "InData", *strcat();
/* static char mode[4] = "rb"; */
static char mode[3] = "r";
long cpt;

#ifdef DEBUG
 PE"\nIndata() started\n\n");
#endif

 if(actionFlag == CONT) strcat(mode, "+");
 if( (Input = fopen(InputFileName,mode)) == NULL)
				  FIERROR(InputFileName)
 setbuf(Input, (char *)NULL);

 read_offsets(p);
 read_head(p);
 read_N(p);
 dynDefs();
 read_P1(p);
 read_P2(p);
 read_C(p);

 if(actionFlag != DISPLAY)
  {
    SEEKBEG( Input, A_offset + blocks * offset_Incr )
    /* This is to ensure the correct continuation of an interrupted job  */
    RL(&cpt, 1)
    if(cpt != point)
	{ PR"\nInput file corrupted: the two values of \'point\' differ!\n\n");
	  exit(-1);
	}
    read_A(p);
  }

#ifdef DEBUG
 printAll(1, "input finished in Indata()");
#endif

}  /*  ------------- end of InData() ----------------  */


void printParams(old_deltaX)  double old_deltaX;
/*============================================*/
/* Print out all important parameters at the beginning of the listing.  */
/* (Called at the end of  checkCL() )					*/
{
 static char *fmt = "%8ld ", *fmtV = "%9g";
 static int  fmt_len = 9;
 int i;

 if(actionFlag != DISPLAY) {
   PR potBlurb); (void)(*outV)(Listing);
   PR"Beta = %.8g\n", beta);
   PR"No. of fictitious particles = %d\n", N_fict);
   PR"Max number of grid points = %ld.\n", MaxPoints);
   if(rgFlag == HALTON) PR"HALTON quasi-random integration;");
   if(rgFlag == RANMAR) PR"RANMAR random number generator;");
   PR"  seed = %d\n", seed);
   switch (integrationFlag)
    {
      case FULL :  PR"Full volume integration\n");
		   break;
      case DIAG :  PR"Integration along the diagonal only:\n");
		   break;
      case DENS :  PR"A single density calculation:\n");
    }
   if(integrationFlag != FULL)
     { PR"            Exponent = %.8g, deltaX = %.8g\n", Exponent, Cx * deltaX);
       if(integrationFlag == DENS && old_deltaX > 0)
		PR"                (changed from %.8g to fit into the cube)\n",
							Cx * old_deltaX);
       if(integrationFlag == DIAG)
	   { PR"            \"Outside\" points ");
	     if(outsideFlag == REMOVE) PR"not ");
	     PR"used\n");
	   }
     }
   if(actionFlag == CONT) PR"Input file: %s\n",InputFileName);
   if(OutputFileName[0] != '\0') PR"Output file: %s\n",OutputFileName);
   if(DensityFileName[0] != '\0') PR"Density file: %s\n",DensityFileName);
   PR"\n");
 }

 if(actionFlag != EPS && PlotWidth)
  {
    /* Initialize line-printer type plotting routines:  */
    initPlot((float)Cx, (float)CE, (float)xMinAt, (float)xMaxAt,
				1, PlotWidth, fmt, fmt_len, Listing);

    if(actionFlag != DISPLAY)
	{ /* Plot potential:    */
	  PR"\n");
	  for(i = -2 * fmt_len; i < PlotWidth; i += 2) PR" ");
	  PR"V(x)\n");
	  plotFun1(V, fmtV);
	}
  }
}


void plotDensity()
/*==============*/
{
register int i;
double dnmntr = Cx * LAt * QS;
float maxD = 0;

for(i = 0; i < N_x; i++)
 if(fabs((double)(dens[i] = densS[i] * FD[i] / dnmntr)) > maxD) maxD = dens[i];

if(maxD > 1000.) PR"\n\t\tDensity is too big!\n");
else { PR"\n\t\t   Density (normalized to 1)\n");
       Histogram(dens, N_x, (float)(xMinAt - 0.5 * xStepAt), (float)xStepAt);
     }
}


void OutDensity(mode)  int mode;
/*============================*/
/* mode=1: in Integrate()
       =0: in DisplayFile()
*/
{
int i;
double dnmntr = Cx * LAt * QS,
       fact = NormFactor / (Cx * LAt * (point + outsidePoints));

if(mode && DensityFileName[0] != '\0')
 { if( (Density = fopen(DensityFileName,"w")) == NULL)
				  FIERROR(DensityFileName)
 }
else
 { Density = Listing;
   if(mode) PD"\nNumerical values for finer plotting:\n");
   PD"\nDensity (normalized to 1):\n");
   PD"          x                 rho(x)/Q\n");
 }

for(i = 0; i < N_x; i++) PD" %.18g  %.18g\n", Cx * xd[i],
					      densS[i] * FD[i] / dnmntr);
if(Density == Listing)
 { PD"\nDensity (normalized to Q):\n");
   PD"          x                 rho(x)\n");
   for(i = 0; i < N_x; i++) PD" %.18g  %.18g\n", Cx * xd[i],
					      densS[i] * FD[i] * fact);
 }

if(Density != Listing) FC Density);

}  /* --------------->  end of:  OutDensity()  <------------------ */


void writeOut()
/*===========*/
{
   int i;
   double alfa, lambda, eps, eps_vir, rhoL, x,
	  aux = NormFactor / (point + outsidePoints);
	  /* NormFactor must be corrected to use actual integration volume
	     which is equal to  point/(point + outsidePoints) * rectangular
	     volume */

   PR"Q = %.16g\n", QS * aux);
   PR"lambda * Q = %.16g, eps_vir * Q = %.16g\n",
		 CE * lambdaS * aux, CE * eps_virS * aux / (double)N_fict);
   PR"alfa * Q = %.16g\n", CE * alfaS * aux);
   alfa = alfaS / QS;
   lambda = lambdaS / QS;
   eps = CE * (C4 - alfa + lambda);
   eps_vir = CE *  eps_virS / (QS * N_fict);
   PR"alfa = %.16g, lambda = %.16g\n", CE * alfa, CE * lambda);
   PR"eps = %.16g, eps_vir = %.16g\n", eps, eps_vir);
   if(integrationFlag == FULL)
    { rhoL = densS[N_x - 1] * FD[N_x - 1] / (Cx * LAt * QS);
      x = Cx * xd[N_x - 1];
      PR"rho(%g)/Q = %.10g\n", x, rhoL);
      PR"eps - eps_vir - (N_fict * xMax / (2 * beta)) * rho(%g)/Q = %.10g\n",
		x, eps - eps_vir - N_fict * xMax * rhoL / (2. * beta) );
      aux = 0.5 * (densS[0] * FD[0] + densS[N_x - 1] * FD[N_x - 1]);
      for(i = 1; i < N_x - 1; i++) aux += (densS[i] * FD[i]);
      PR"rho normalization test: %g (should be 1)\n", aux / (QS * (N_x - 1)));
    }
   (void)fflush(Listing);
}


void Integrate()
/*============*/
{
char *p = "Integrate";
int i, start = 1,
    Pause = PlotInt ? PlotInt : WriteInt,
    pointCount = point % Pause;
double x0, xp, x, sumA, sumL, al, la, q, diff, ev, SUM;
extern void srand48();
extern long mrand48();
extern double drand48();
long ij, kl;

 for(i = 0; i < N_x; i++) XD[i] = C3 * xd[i] * xd[i];

 /* Initialize random number generators */
 srand48((long)abs(seed));
 if(rgFlag == HALTON)
	{ pR = vHalton;
	  if(seed <= 0)
		 { for(i = 0; i < N_fict; i++) N0[i] = -seed;
		   init_vHalton(N_fict, N0);
		 }
		 else  rand_init_vHalton(N_fict, (unsigned)mrand48());
	}
  else  { pR = ranmar1;
	  ij = drand48() * 31328; 
	  kl = drand48() * 30081;
	  (void)rmarin1(&ij, &kl);
        }

 OutData();

 /*******************************************************************/
 (void)times(&cpuT);	/* Timing of the integration proper started */
 usLast = (double)cpuT.tms_utime;
 sysLast = (double)cpuT.tms_stime;

 Resume:
 while(pointCount < Pause)
  { point++; pointCount++;

#ifdef DEBUG1
PE"point = %ld     ", point);
#endif

    /* Calculate contribution to QS, alfaS, lambdaS, eps_virS:		   */
    /* The next N_fict calls to the random generator (*pR)() will generate */
    /* all coordinates of one grid point of the random integration grid.   */
    x0 =  xMinAt + LAt * (*pR)();
    ev = (*epsVir)(x0);
    SUM = xp = x =  xMinAt + LAt * (*pR)();  /* x1 */
    sumA = x * x;
    sumL = (*V)(x);
    ev += (*epsVir)(x);
    for(i = 2; i < N_fict; i++)
	{ x =  xMinAt + LAt * (*pR)();
	  diff = x - xp;
	  sumA += diff * diff;
	  sumL += (*V)(x);
	  ev += (*epsVir)(x);
	  xp = x;
	}
    sumA += x * x; 
    SUM += x;
    al = C2 * (2. * x0 * (x0 - SUM) + sumA);
    la = ((*V)(x0) + sumL) / (double)N_fict;
    q = exp(-betaAt * (al + la));
    QS += q;
    alfaS += al * q;
    lambdaS += la * q;
    eps_virS += ev * q;

#ifdef DEBUG1
PE" q = %.16g\n", q);
#endif
    
    /* Calculate contribution to  rho(xd[i]) (particle density): */
    SUM *= C3;
    x = sumL / (double)N_fict + C2 * sumA;
    for(i = 0; i < N_x; i++)
	densS[i] += exp(-betaAt * (XD[i] - xd[i] * SUM + x));
  }	

 if( ! noOutputFlag && !(point % WriteInt) )
 {
	/* Update Output:  write out the current sums: */
	blocks++;
	if(start) { start = 0;
	    SEEKBEG(Output, A_offset + blocks * offset_Incr)
	    /* This is to ensure the correct continuation
	       of an interrupted previous run             */
	}
	else if( fseek(Output, 0L, SEEK_END) ) err_fseek(p);
	write_A();
	SEEKBEG(Output, C_offset)
	write_C();
	write_Lm(p);
 }

 if(PlotInt) { PR"\n\npoint = %ld    QS = %.16g\n", point, QS);
	       writeOut();
	       if(PlotWidth) plotDensity();
	     }

 if(point < MaxPoints) { pointCount = 0; goto Resume; }

 OutDensity(1);

(void)times(&cpuT);	/* Timing of the integration proper ended */
/******************************************************************/

PR
"\nCPU time used for integration proper:\n     user %g, system %g, total %g\n",
	(float)(cpuT.tms_utime - usLast)/ 60.,
	(float)(cpuT.tms_stime - sysLast)/ 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime - usLast - sysLast) / 60.);

}  /* --------------->  end of:  Integrate()  <------------------ */


void DiagIntegrate()
/*================*/
{
char *p = "DiagIntegrate";
int i, start = 1,
    Pause = PlotInt ? PlotInt : WriteInt,
    pointCount = point % Pause,
    iOrthBas = 0, j;
double x0, x, sumA, sumL, q, diff, ev;
extern void srand48();
extern long mrand48();
extern double drand48();
long ij, kl;

 /* Generate orthogonal basis perpendicular to the body diagonal: */
 for(i = 1; i < N_fict; i++)
	{ orthBas[iOrthBas++] = deltaX / sqrt((double)(i * i + i));
		/* this element occurs in a basis vector  i  times */
	  orthBas[iOrthBas++] = -deltaX * sqrt((double)i / (double)(i + 1));
		/* this element occurs in a basis vector once */
	}

 /* Initialize random number generators: */
 srand48((long)abs(seed));
 if(rgFlag == HALTON)
	{ pR = vHalton;
	  if(seed <= 0)
		 { for(i = 0; i < N_fict; i++) N0[i] = -seed;
		   init_vHalton(N_fict, N0);
		 }
		 else  rand_init_vHalton(N_fict, (unsigned)mrand48());
	}
  else  { pR = ranmar1;
	  ij = drand48() * 31328; 
	  kl = drand48() * 30081;
	  (void)rmarin1(&ij, &kl);
        }

 OutData();

 /*******************************************************************/
 (void)times(&cpuT);	/* Timing of the integration proper started */
 usLast = (double)cpuT.tms_utime;
 sysLast = (double)cpuT.tms_stime;

 Resume:
 while(pointCount < Pause)
  { /* Generate a mesh point in a channel along the body diagonal: */
#ifdef DEBUG1
    PE"New point: ");
#endif
    iOrthBas = 0;
    x0 =  xMinAt + LAt * (*pR)();
    for(i = 0; i < N_fict; i++) xi[i] = x0;
    for(j = 1; j < N_fict; j++)
	{ x = (*pR)() - 0.5;
	  x0 = x * orthBas[iOrthBas++];
	  for(i = 0; i < j; i++) xi[i] += x0;
	  xi[j] += (x * orthBas[iOrthBas++]);
	}

    if(outsideFlag == REMOVE)
     { /* Check whether inside the cube (xMaxAt-xMinAt)^N_fict: */
       for(i = 0; i < N_fict; i++)
		if(xi[i] < xMinAt || xi[i] > xMaxAt) break;
       if(i < N_fict) { outsidePoints++; 
#ifdef DEBUG1
			PE"outside!!!\n");
#endif
			continue;
		      }
     }

    point++; pointCount++;

#ifdef DEBUG1
    PE"point = %ld     ", point);
#endif

    /* Calculate contribution to QS, alfaS, lambdaS, eps_virS:		   */
    x0 =  xi[0];
    ev = (*epsVir)(x0);
    sumL = (*V)(x0);
    diff = x0 - xi[N_fict_1];
    sumA = diff * diff;
    for(i = 1; i < N_fict; i++)
	{ x =  xi[i];
	  diff = x - x0;
	  sumA += diff * diff;
	  sumL += (*V)(x);
	  ev += (*epsVir)(x);
	  x0 = x;
	}
    sumA *= C2;
    sumL /= (double)N_fict;
    q = exp(-betaAt * (sumA + sumL));
    QS += q;
    alfaS += sumA * q;
    lambdaS += sumL * q;
    eps_virS += ev * q;

#ifdef DEBUG1
    PE" q = %.16g\n", q);
#endif

  }  /*  -----  end of:  while(pointCount < Pause)  -----   */

 if( ! noOutputFlag && !(point % WriteInt) )
    {
	/* Update Output:  write out the current sums: */
	blocks++;
	if(start) { start = 0;
		    SEEKBEG(Output, A_offset + blocks * offset_Incr)
		    /* This is to ensure the correct continuation
		       of an interrupted previous run		*/
		  }
	else if( fseek(Output, 0L, SEEK_END) ) err_fseek(p);
	write_A();
	SEEKBEG(Output, C_offset)
	write_C();
	write_Lm(p);
    }

 if(PlotInt) {  PR"\n\npoint = %ld    QS = %.16g\n", point, QS);
		if(outsideFlag == REMOVE)
		    PR"# of outside (rejected) points: %ld\n", outsidePoints);
		writeOut();
	     }

 if(point < MaxPoints) { pointCount = 0; goto Resume; }

(void)times(&cpuT);	/* Timing of the integration proper ended */
/******************************************************************/

PR
"\nCPU time used for integration proper:\n     user %g, system %g, total %g\n",
	(float)(cpuT.tms_utime - usLast)/ 60.,
	(float)(cpuT.tms_stime - sysLast)/ 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime - usLast - sysLast) / 60.);

}  /* --------------->  end of:  DiagIntegrate()  <------------------ */


void DensIntegrate()
/*================*/
{
char *p = "DensIntegrate";
int i, start = 1,
    Pause = PlotInt ? PlotInt : WriteInt,
    pointCount = point % Pause;
double xp, x, sumA, sumL, diff, L0, xCentre;
extern void srand48();
extern long mrand48();
extern double drand48();
long ij, kl;

 /* Initialize random number generators: */
 srand48((long)abs(seed));
 if(rgFlag == HALTON)
	{ pR = vHalton;
	  if(seed <= 0)
		 { for(i = 0; i < N_fict_1; i++) N0[i] = -seed;
		   init_vHalton(N_fict_1, N0);
		 }
		 else  rand_init_vHalton(N_fict_1, (unsigned)mrand48());
	}
  else  { pR = ranmar1;
	  ij = drand48() * 31328; 
	  kl = drand48() * 30081;
	  (void)rmarin1(&ij, &kl);
        }

 /* Make sure that the little cube is completely inside the big one: */
 if(LAt <= deltaX) { /* Correction in checkCL(): deltaX = LAt!!! */
		     xCentre = xMinAt + 0.5 * LAt;
		   }
 else if(xValueAt + 0.5 * deltaX > xMaxAt) xCentre = xMaxAt - 0.5 * deltaX;
 else if(xValueAt - 0.5 * deltaX < xMinAt) xCentre = xMinAt + 0.5 * deltaX;
 else xCentre = xValueAt;

 OutData();

 /*******************************************************************/
 (void)times(&cpuT);	/* Timing of the integration proper started */
 usLast = (double)cpuT.tms_utime;
 sysLast = (double)cpuT.tms_stime;

 L0 = (*V)(xValueAt);

 Resume:
 while(pointCount < Pause)
  { point++; pointCount++;

    /* Calculate contribution to rho(xValue):		   */
    x =  xCentre + ((*pR)() - 0.5) * deltaX;
    sumL = L0 + (*V)(x);
    diff = x - xValueAt;
    sumA = diff * diff;
    for(i = 1; i < N_fict_1; i++)
	{ xp = x;
	  x =  xCentre + ((*pR)() - 0.5) * deltaX;
	  diff = xp - x;
	  sumA += diff * diff;
	  sumL += (*V)(x);
	}
    diff = x - xValueAt;
    sumA += diff * diff;
    sumA *= C2;
    sumL /= (double)N_fict;
    QS += exp(-betaAt * (sumA + sumL));

  }  /*  -----  end of:  while(pointCount < Pause)  -----   */

 if( ! noOutputFlag && !(point % WriteInt) )
    {
	/* Update Output:  write out the current sum: */
	blocks++;
	if(start) { start = 0;
		    SEEKBEG(Output, A_offset + blocks * offset_Incr)
		    /* This is to ensure the correct continuation
		       of an interrupted previous run		*/
		  }
	else if( fseek(Output, 0L, SEEK_END) ) err_fseek(p);
	write_A();
  	SEEKBEG(Output, C_offset)
	write_C();
	write_Lm(p);
    }

 if(PlotInt) {  PR"\n\npoint = %ld    QS = %.16g\n", point, QS);
		PR"         rho(%g) = %.16g (normalized to Q)\n",
		    xValue, QS * NormFactor / (Cx * (point + outsidePoints)));
	     }

 if(point < MaxPoints) { pointCount = 0; goto Resume; }

(void)times(&cpuT);	/* Timing of the integration proper ended */
/******************************************************************/

PR
"\nCPU time used for integration proper:\n     user %g, system %g, total %g\n",
	(float)(cpuT.tms_utime - usLast)/ 60.,
	(float)(cpuT.tms_stime - sysLast)/ 60.,
	(float)(cpuT.tms_utime + cpuT.tms_stime - usLast - sysLast) / 60.);

}  /* --------------->  end of:  DensIntegrate()  <------------------ */


void displayFile(Listing)  FILE *Listing;
/*=====================================*/
/* Similar to InData(), however it is assumed here that InData() was called
   before the invocation of this function */
{
long oEnd, blockEnd;
int i;
char *p = "displayFile";
static char  flg[2] = "n";

  PR"\n\n\t\t\tTHE CONTENTS OF THE FILE: "); PR InputFileName);

 SEEKBEG(Input, 0L)
 read_offsets(p);
  PR"\n\nLm_offset = %ld, V_offset = %ld, N_offset = %ld\n",
	 Lm_offset, V_offset, N_offset);
  PR"P1_offset = %ld, P2_offset = %ld, C_offset = %ld\n",
	 P1_offset, P2_offset, C_offset);
  PR"A_offset = %ld, offset_Incr = %ld", A_offset, offset_Incr);

 for(i = 0; i < 12; i++) fputs(fgets(text,MAXLETT,Input), Listing);

 read_head(p);
 (void)(*outV)(Listing);
 (void)fgets(text,MAXLETT,Input);

 for(i = 0; i < 6; i++) fputs(fgets(text,MAXLETT,Input), Listing);
   
 PR"\nP1:\n");
  PR"beta = %.16g, betaAt = %.16g\n", beta, betaAt);
  PR"Cx = %.16g, CE = %.16g\n", Cx, CE);
  PR"xMin = %.8g, xMax = %.8g\n", xMin, xMax);
  PR"xMinAt = %.16g, xMaxAt = %.16g\n", xMinAt, xMaxAt);
  PR"xStepAt = %.16g, C4 = %.16g\n", xStepAt, C4);
  PR"C1 = %.16g, C2 = %.16g, C3 = %.16g\n", C1, C2, C3);
   if(integrationFlag != FULL)
      PR"Exponent = %.16g, deltaX = %.16g\n", Exponent, Cx * deltaX);

 PR"\nP2:\n");
 SEEKBEG(Input, P2_offset)
 read_P2(p);
  PR"MaxPoints = %ld, WriteInt = %ld, PlotInt = %ld\n",
	 MaxPoints, WriteInt, PlotInt);

 (void)fgets(text,MAXLETT,Input);
 for(i = 0; i < 4; i++) fputs(fgets(text,MAXLETT,Input), Listing);

 if( fseek(Input, 0L, SEEK_END) ) err_fseek(p);
 oEnd = ftell(Input) - A_offset;
 blockEnd = oEnd / offset_Incr;
 PR"\nA:");
 SEEKBEG(Input, A_offset)
 blocks = 0;
 while( flg[0] != 'q' )
   {
#ifdef ultrix
     if(!fread((void *)&point,sizeof(long),(size_t)1,Input))
#else
     if(!fread((char *)&point,sizeof(long),(size_t)1,Input))
#endif
       { if(feof(Input))
		{ PR"\nEnd of file (at %ld bytes from A:) reached/passed.",
				oEnd);
		  PR"\n                     %ld = %ld * offset_Incr + %ld\n\n",
				oEnd, blockEnd, oEnd % offset_Incr);
		}
		else  PR"\nNo value for point read in, but not EOF.\n");
       }
       else
       { PR"\n    BLOCK # %ld\n", blocks);
	 read_A(p);
	 PR"point = %ld    QS = %.16g\n", point, QS);
	 if(integrationFlag == DIAG)
		PR"# of outside (rejected) points: %ld\n", outsidePoints);
	 if(point > 0 && integrationFlag == DENS)
		PR"         rho(%g) = %.16g (normalized to Q)\n",
		 xValue, QS * NormFactor / (Cx * point));
	 else
	 {
	   if(QS != 0.)
	    { writeOut();
#ifdef DEBUG
	      printAll(1, "displayFile() - before density output/plotting");
#endif
	      if(integrationFlag == FULL)
		{ if(PlotWidth) plotDensity();  else OutDensity(0); }
	    }
	   else
	    {
	      if(alfaS != 0.) PR"alfaS = %.16g != 0 !!!\n", alfaS);
	      if(lambdaS != 0.) PR"lambdaS = %.16g != 0 !!!\n", lambdaS);
	      if(eps_virS != 0.) PR"eps_virS = %.16g != 0 !!!\n", eps_virS);
	      if(integrationFlag == FULL)
		   for(i = 0; i < N_x; i++)
			if(densS[i] != 0.)
				PR"densS[%d] = %.16g != 0 !!!\n", i, densS[i]);
	    }
	 }
       }
     PE"\n\n\7Continue with the next block, jump to a block, or exit?\
 (n/j/q) ... ");
     SC"%s", flg);
     if(flg[0] == 'j') { PE"\7Give block number ... ");
			 SC"%ld", &blocks);
			 if(blocks < 0L) blocks = 0L;
			 if(blocks > blockEnd) blocks = blockEnd + 1L;
			 SEEKBEG( Input, A_offset + blocks * offset_Incr )
		       }
		       else blocks++; 
   }

FC Input);

}  /*  ------------- end of displayFile() ----------------  */


void Final()
/*========*/
/* It is assumed here that InData() was called
   before the invocation of this function */
{
int iBl;
char *p = "Final", p1[2];
FILE *Output0, *Output1;

if(integrationFlag == DENS)
	PE"\nIs it O.K. to rewrite file rho? (y/n)  ");
else
	PE"\nIs it O.K. to rewrite files eps, eps_vir, Q? (y/n)  ");
SC"%s",p1);
if(p1[0] != 'y') { PE"\nNo files modified\n\n");
		   return;
		 }

if(integrationFlag == DENS)
  { if( (Output = fopen("rho","w")) == NULL)  FIERROR("rho") }
else
  { if( (Output = fopen("eps","w")) == NULL)  FIERROR("eps")
    if( (Output1 = fopen("eps_vir","w")) == NULL)  FIERROR("eps_vir")
    if( (Output0 = fopen("Q","w")) == NULL)  FIERROR("Q")
  }

SEEKBEG(Input, A_offset + offset_Incr)
	/* Skip the initial-configuration block that contains no sums yet */

for(iBl = 1; iBl <= blocks; iBl++)
 {
#ifdef ultrix
   if(!fread((void *)&point,sizeof(long),(size_t)1,Input))
#else
   if(!fread((char *)&point,sizeof(long),(size_t)1,Input))
#endif
       { if(feof(Input))
		{ PR"\nUnexpected end of input file!\n");
		  break;
		}
       }
   read_A(p);
   if(integrationFlag == DENS)
	PO"%ld %g\n", point, QS * NormFactor / (Cx * point));
    else
	{
	  /* Q: */
	  (void)fprintf(Output0, "%ld %g\n",
			point, QS * NormFactor / (point + outsidePoints));
	  /* eps: */
	  PO"%ld %g\n", point, CE * (C4 + (lambdaS - alfaS) / QS));
	  /* eps_vir: */
	  (void)fprintf(Output1, "%ld %g\n",
			point, CE *  eps_virS / (QS * N_fict));
	}
 }

FC Output);
if(integrationFlag != DENS) { FC Output1); FC Output0); }
FC Input);

}  /*  ------------- end of Final() ----------------  */
